//
// Created by 11067 on 2023/6/29.
//

#include "ACodeGen.h"

#include "../../utils/bitUtils.h"
#include "instructions/AInstrLS.h"
#include "instructions/AInstrMov.h"
#include "instructions/AInstrVFP.h"
#include <algorithm>
#include <string>

#include "instructions/AInstrALU.h"
#include "instructions/AInstrB.h"
#include "instructions/AInstrMul.h"
#include "pass/AfterCodegenPass.h"
#include "pass/ColorAllocator.h"
#include "pass/JTable.h"
#include "pass/MergeBlock.h"
#include "pass/MergeCondExec.h"
#include "pass/Peephole.h"

using std::literals::string_literals::operator""s;

namespace armv7 {
    const int PARAM_REG_CNT = 4;
    const int FLOAT_REG_CNT = 16;

    static int calculateStackParamOffset(int regCnt, int floatCnt) {
        int regOffset = regCnt - PARAM_REG_CNT > 0 ? regCnt - PARAM_REG_CNT : 0;
        int floatOffset = floatCnt - FLOAT_REG_CNT > 0 ? floatCnt - FLOAT_REG_CNT : 0;
        return (regOffset + floatOffset) * WORD_SIZE;
    }

    static int calculateVectorSize(const std::vector<int> &dims) {
        int s = 1;
        for (const auto &item: dims) {
            s *= item;
        }
        return s;
    }

    static bool isRhsMayBeImm(Ir::BinaryType binaryType) {
        return binaryType == Ir::ADD || binaryType == Ir::SUB ||
               //       binaryType == Ir::EQ || binaryType == Ir::NE ||
               //               binaryType == Ir::SLE || binaryType == Ir::SLT ||
               //               binaryType == Ir::SGE || binaryType == Ir::SGT ||
               binaryType == Ir::OR || binaryType == Ir::AND || binaryType == Ir::XOR;
    }

    static Cond irCmp2ACmp(Ir::BinaryType binaryType) {
        switch (binaryType) {
            case Ir::EQ:
                return Cond::Eq;
            case Ir::NE:
                return Cond::Ne;
            case Ir::SGT:
                return Cond::Gt;
            case Ir::SGE:
                return Cond::Ge;
            case Ir::SLE:
                return Cond::Le;
            case Ir::SLT:
                return Cond::Lt;
            default:
                panic("not a compare.");
        }
    }

    static Cond irCmp2RevACmp(Ir::BinaryType binaryType) {
        switch (binaryType) {
            case Ir::EQ:
                return Cond::Eq;
            case Ir::NE:
                return Cond::Ne;
            case Ir::SGT:
                return Cond::Lt;
            case Ir::SGE:
                return Cond::Le;
            case Ir::SLE:
                return Cond::Ge;
            case Ir::SLT:
                return Cond::Gt;
            default:
                panic("not a compare.");
        }
    }

    static int evaluateCmp(Ir::BinaryType binaryType, int a, int b) {
        switch (binaryType) {
            case Ir::EQ:
                return a == b ? 1 : 0;
            case Ir::NE:
                return a != b ? 1 : 0;
            case Ir::SGT:
                return a > b ? 1 : 0;
            case Ir::SGE:
                return a >= b ? 1 : 0;
            case Ir::SLE:
                return a <= b ? 1 : 0;
            case Ir::SLT:
                return a < b ? 1 : 0;
            default:
                panic("not a compare.");
        }
    }

    static int evaluateCmpF(Ir::BinaryType binaryType, float a, float b) {
        switch (binaryType) {
            case Ir::EQ:
                return a == b ? 1 : 0;
            case Ir::NE:
                return a != b ? 1 : 0;
            case Ir::SGT:
                return a > b ? 1 : 0;
            case Ir::SGE:
                return a >= b ? 1 : 0;
            case Ir::SLE:
                return a <= b ? 1 : 0;
            case Ir::SLT:
                return a < b ? 1 : 0;
            default:
                panic("not a compare.");
        }
    }

    static bool isCmp(Ir::BinaryType binaryType) {
        return binaryType == Ir::EQ || binaryType == Ir::NE ||
               binaryType == Ir::SGT || binaryType == Ir::SGE ||
               binaryType == Ir::SLT || binaryType == Ir::SLE;
    }

    void ACodeGen::run() {
        genData();
        genFunction();
        for (auto func: *mCompileUnit) {
            func->calcPrecSucc();
        }
        if (mergeBlock) {
            armv7::MergeBlock().run(*mCompileUnit);
        }
        armv7::ColorAllocator().run(*mCompileUnit);
        if (mergeBlock) {
            armv7::MergeBlock().run(*mCompileUnit);
        }
        info(mCompileUnit->toAssemblyString());
        if (mergeCondExec) {
            armv7::MergeCondExec().run(*mCompileUnit);
        }
        if (mergeBlock) {
            armv7::MergeBlock().run(*mCompileUnit);
        }
        if (jtable) {
            armv7::JTable().run(*mCompileUnit);
        }
        armv7::AfterCodegenPass().insertLTORG(*mCompileUnit);
        if (peepHole) {
            armv7::Peephole().run(*mCompileUnit);
        }
        armv7::AfterCodegenPass().run(*mCompileUnit);
    }

    // todo: 生成data有点问题
    void ACodeGen::genData() {
        for (const auto &glob_arr: compileUnit->get_global_int_arrays()) {
            if (!glob_arr.second.empty()) {
                AData *aData{makeADataInt(glob_arr.first, glob_arr.second)};
                mCompileUnit->data.emplace_back(aData);
                globVecNameMap.insert({glob_arr.first, aData});
            }
        }
        for (const auto &glob_arr: compileUnit->get_global_float_arrays()) {
            if (!glob_arr.second.empty()) {
                AData *aData{makeADataFloat(glob_arr.first, glob_arr.second)};
                mCompileUnit->data.emplace_back(aData);
                globVecNameMap.insert({glob_arr.first, aData});
            }
        }
        for (const auto &glob_val: compileUnit->get_globals()) {
            Instanceof(reg, Ir::Reg *, glob_val.get()) {
                if (reg->get_type() == Ir::Float) {
                    AData *aData{
                            makeADataFloat(reg->get_global_name(),
                                           {static_cast<float>(reg->get_num().float_num)})};
                    mCompileUnit->data.emplace_back(aData);
                    globVarMap.insert({glob_val.get(), aData});
                } else if (reg->get_type() == Ir::Int) {
                    AData *aData{
                            makeADataInt(reg->get_global_name(), {reg->get_num().int_num})};
                    mCompileUnit->data.emplace_back(aData);
                    globVarMap.insert({glob_val.get(), aData});
                }
            }
            Instanceof(vec, Ir::Vector *, glob_val.get()) {
                if (vec->get_type() == Ir::Float) {
                    auto &arrMap = compileUnit->get_global_float_arrays();
                    if (!arrMap.find(vec->get_name())->second.empty()) {
                        globVarMap.insert(
                                {glob_val.get(), globVecNameMap.at(vec->get_name())});
                    } else {
                        // 全局数组没有初始化列表
                        AData *pAData{makeADataFloatZeroed(
                                vec->get_name(), calculateVectorSize(vec->get_dims()))};
                        mCompileUnit->bss.emplace_back(pAData);
                        globVarMap.insert({glob_val.get(), pAData});
                    }
                } else if (vec->get_type() == Ir::Int) {
                    auto &arrMap = compileUnit->get_global_int_arrays();
                    if (!arrMap.find(vec->get_name())->second.empty()) {
                        globVarMap.insert(
                                {glob_val.get(), globVecNameMap.at(vec->get_name())});
                    } else {
                        // 全局数组没有初始化列表
                        AData *pAData{makeADataIntZeroed(
                                vec->get_name(), calculateVectorSize(vec->get_dims()))};
                        mCompileUnit->bss.emplace_back(pAData);
                        globVarMap.insert({glob_val.get(), pAData});
                    }
                }
            }
        }
    }

    void ACodeGen::genFunction() {
        for (auto &func: compileUnit->get_functions()) {
            irFunction = func.get();
            if (!irFunction->is_lib()) {
                curAFunction = new AFunction{irFunction->get_name()};
                funcMap.insert({irFunction, curAFunction});
                mCompileUnit->addLast(curAFunction);
                vRegAllocator.clear();
                irReg2AReg.clear();
                irStackVector2offset.clear();
                irParamVector2BaseReg.clear();
                irMemObject2AReg.clear();
                irStackVar2offset.clear();
                irAliasCvt.clear();
                savedRegMap.clear();
                indexHints.clear();
                loadStoreMemObj.clear();
                genEachFunction();
                curAFunction->vRegAllocated = vRegAllocator.getAllocCnt();
                curAFunction->vRegType = vRegAllocator.vRegType;
            } else {
                auto aFunc = new AFunction{irFunction->get_name()};
                aFunc->isLib = true;
                funcMap.insert({irFunction, aFunc});
                mCompileUnit->addLast(aFunc);
            }
        }
    }

    void ACodeGen::genEachFunction() {
        // 为每一个IR基本块分配后端IR基本块
        for (auto &pIrBlock: irFunction->get_bbs()) {
            auto *pABlock = new ABlock{".L"s + std::to_string(blockCnt++)};

            irBB2ABB.insert({pIrBlock.get(), pABlock});
            curAFunction->addLast(pABlock);
        }
        // 从IR block向MIR block复制信息
        for (auto &pIrBlock: irFunction->get_bbs()) {
            auto *pABlock = irBB2ABB[pIrBlock.get()];
            pABlock->loopDepth = pIrBlock->get_depth();
            for (auto &bb: pIrBlock->prev) {
                pABlock->prec.insert(irBB2ABB[bb]);
            }
            for (auto &bb: pIrBlock->succ) {
                pABlock->succ.insert(irBB2ABB[bb]);
            }
        }
        whenDebug for (auto &[pIrBlock, pABlock]: irBB2ABB) {
            for (auto prec: pABlock->prec) {
                atriAssert(prec->succ.count(pABlock));
            }
            for (auto succ: pABlock->succ) {
                atriAssert(succ->prec.count(pABlock));
            }
        }

        // 进入第一个块
        curABlock = curAFunction->getFirst();
        irBlock = irFunction->get_bbs().front().get();
        genProtectCtx();// 保护被调用者保存的寄存器
        resolveParam();
        for (auto &pIrBlock: irFunction->get_bbs()) {
            curABlock = irBB2ABB.at(pIrBlock.get());
            irBlock = pIrBlock.get();
            genBlock();
        }
        arrayIndexLastOffset();
        //        info(mCompileUnit->toAssemblyString());
        placeHolderReplace();

        whenDebug curAFunction->toAssemblyString(std::cerr);

        destructPhi();
        whenDebug curAFunction->toAssemblyString(std::cerr);
    }

    ACodeGen::ACodeGen(Ir::CompileUnit *_compileUnit) {
        compileUnit = _compileUnit;
    }

    void ACodeGen::genBlock() {
        std::unordered_map<AData *, AReg> cachedGlobalAddr{};
        if (curABlock->getSize() == 0) {
            curABlock->addLast(new AInstrComment{"head node for linked list."});
        }
        std::vector<Ir::Instruction *> instrs{};
        for (auto &instr: irBlock->get_instrs()) {
            instrs.push_back(instr.get());
        }
        for (size_t iinstr = 0; iinstr < instrs.size(); ++iinstr) {
            auto pInstruction = instrs[iinstr];
            Instanceof(binaryInstr, Ir::BinaryInstr *, pInstruction) {
                auto &lhs = binaryInstr->get_src_1();
                auto &rhs = binaryInstr->get_src_2();
                auto &d = binaryInstr->get_dst();
                AReg alhs = VReg{};
                AReg arhs = VReg{};
                int alhsi{};
                int arhsi{};// 可能被直接编码入指令的立即数
                            //                float alhsf{};
                            //                float arhsf{};
                            //                float arhsf{};
                // 如果是除法，可能有优化机会，先不加载寄存器
                // todo: 除常数优化
                if (isRhsMayBeImm(binaryInstr->get_binary_type())) {
                    if (binaryInstr->get_binary_type() == Ir::SUB ||
                        binaryInstr->get_binary_type() == Ir::ADD) {
                        if (lhs->is_constant() &&
                            (lhs->get_type() == Ir::Int || lhs->get_type() == Ir::Bool) &&
                            isImm8m(lhs->get_num().int_num)) {
                            alhsi = lhs->get_num().int_num;
                        } else {
                            alhs = genReg(lhs.get());
                        }
                    } else {
                        alhs = genReg(lhs.get());
                    }
                    if (rhs->is_constant() &&
                        (rhs->get_type() == Ir::Int || rhs->get_type() == Ir::Bool) &&
                        isImm8m(rhs->get_num().int_num)) {
                        arhsi = rhs->get_num().int_num;
                    } else {
                        arhs = genReg(rhs.get());
                    }
                } else if (isCmp(binaryInstr->get_binary_type())) {
                    if (lhs->is_constant() &&
                        (lhs->get_type() == Ir::Int || lhs->get_type() == Ir::Bool) &&
                        isImm8m(lhs->get_num().int_num)) {
                        alhsi = lhs->get_num().int_num;
                    } else if (lhs->is_constant() && lhs->get_type() == Ir::Float &&
                               lhs->get_num().float_num == 0) {
                        alhsi = 0;
                    } else {
                        alhs = genReg(lhs.get());
                    }
                    if (rhs->is_constant() &&
                        (rhs->get_type() == Ir::Int || rhs->get_type() == Ir::Bool) &&
                        isImm8m(rhs->get_num().int_num)) {
                        arhsi = rhs->get_num().int_num;
                    } else if (rhs->is_constant() && rhs->get_type() == Ir::Float &&
                               rhs->get_num().float_num == 0) {
                        arhsi = 0;
                    } else {
                        arhs = genReg(rhs.get());
                    }
                } else if (binaryInstr->get_binary_type() == Ir::SDIV ||
                           binaryInstr->get_binary_type() == Ir::SREM) {
                    alhs = genReg(lhs.get());
                    if ((rhs->get_type() == Ir::Int || rhs->get_type() == Ir::Bool) &&
                        rhs->is_constant()) {
                        arhsi = rhs->get_num().int_num;
                    } else {
                        arhs = genReg(rhs.get());
                    }
                } else if (binaryInstr->get_binary_type() == Ir::MUL &&
                           (rhs->get_type() == Ir::Int || rhs->get_type() == Ir::Bool)) {
                    if (rhs->is_constant() && isPow2(rhs->get_num().int_num)) {
                        arhsi = rhs->get_num().int_num;
                    } else {
                        arhs = genReg(rhs.get());
                    }
                    if (lhs->is_constant() && isPow2(lhs->get_num().int_num)) {
                        alhsi = lhs->get_num().int_num;
                    } else {
                        alhs = genReg(lhs.get());
                    }
                } else {
                    alhs = genReg(lhs.get());
                    arhs = genReg(rhs.get());
                }

                auto type = IRType2VRegType(d->get_type());
                auto operandType = IRType2VRegType(lhs->get_type());// 用于比较
                AReg dest;
                if (isCompare(binaryInstr->get_binary_type())) {
                    // 比较指令的结果只能是Int
                    // dest = vRegAllocator.alloc(VReg::Type::Int);
                    // 但先不分配
                } else {
                    dest = vRegAllocator.alloc(type);
                }

                AInstr *aInstr{};
                switch (binaryInstr->get_binary_type()) {
                    case Ir::ADD:
                        if (type == VReg::Type::Int) {
                            if (aRegNull(alhs) && aRegNull(arhs)) {
                                aInstr = new AInstrMov32(dest, alhsi + arhsi);
                            } else if (aRegNull(arhs)) {
                                aInstr = makeAInstrALURImm(AInstrALU::Type::Add, dest, alhs, arhsi);
                            } else if (aRegNull(alhs)) {
                                aInstr = makeAInstrALURImm(AInstrALU::Type::Add, dest, arhs, alhsi);
                            } else {
                                aInstr = makeAInstrALURR(AInstrALU::Type::Add, dest, alhs, arhs);
                            }
                        } else {
                            aInstr =
                                    makeFloatBinary(AInstrFloatBinary::Type::Vadd, dest, alhs, arhs);
                        }

                        break;
                    case Ir::SUB:
                        if (type == VReg::Type::Int) {
                            if (aRegNull(alhs) && aRegNull(arhs)) {
                                aInstr = new AInstrMov32(dest, alhsi - arhsi);
                            } else if (aRegNull(alhs)) {
                                aInstr = makeAInstrALURImm(AInstrALU::Type::Rsb, dest, arhs, alhsi);
                            } else if (aRegNull(arhs)) {
                                aInstr = makeAInstrALURImm(AInstrALU::Type::Sub, dest, alhs, arhsi);
                            } else {
                                aInstr = makeAInstrALURR(AInstrALU::Type::Sub, dest, alhs, arhs);
                            }

                        } else {
                            aInstr =
                                    makeFloatBinary(AInstrFloatBinary::Type::Vsub, dest, alhs, arhs);
                        }
                        break;
                    case Ir::SDIV:
                        if (type == VReg::Type::Int) {
                            if (aRegNull(arhs)) {
                                if (divOpt) {
                                    genOptimizedDiv(dest, alhs, arhsi);
                                } else {
                                    arhs = genReg(rhs.get());
                                    aInstr = makeSDiv(dest, alhs, arhs);
                                }
                            } else {
                                aInstr = makeSDiv(dest, alhs, arhs);
                            }
                        } else {
                            aInstr =
                                    makeFloatBinary(AInstrFloatBinary::Type::Vdiv, dest, alhs, arhs);
                        }
                        break;
                    case Ir::MUL:
                        if (type == VReg::Type::Int) {
                            if (aRegNull(alhs) && aRegNull(arhs)) {
                                aInstr = new AInstrMov32(dest, alhsi * arhsi);
                            } else if (aRegNull(alhs)) {
                                aInstr = makeMovBitRIMM(dest, arhs, ShiftType::LSL, calcCLZ(alhsi));
                            } else if (aRegNull(arhs)) {
                                aInstr = makeMovBitRIMM(dest, alhs, ShiftType::LSL, calcCLZ(arhsi));
                            } else {
                                aInstr = makeMul(dest, alhs, arhs);
                            }
                        } else {
                            aInstr =
                                    makeFloatBinary(AInstrFloatBinary::Type::Vmul, dest, alhs, arhs);
                        }
                        break;
                    case Ir::UDIV:
                        unimplemented();
                        break;
                    case Ir::UREM:
                        unimplemented();
                        break;
                    case Ir::SREM: {
                        if (aRegNull(arhs)) {
                            if (divOpt) {
                                genOptimizedMod(dest, alhs, arhsi);
                            } else {
                                arhs = genReg(rhs.get());
                                auto [instr0, instr1, instr2] = makeRem(dest, alhs, arhs);
                                curABlock->addLast(instr0);
                                curABlock->addLast(instr1);
                                curABlock->addLast(instr2);
                            }
                        } else {

                            //                            auto [instr0, instr1, instr2] =
                            //                            makeRem(dest, alhs, arhs);
                            //                            curABlock->addLast(instr0);
                            //                            curABlock->addLast(instr1);
                            //                            curABlock->addLast(instr2);
                            curABlock->addLast(makeSDiv(dest, alhs, arhs));
                            curABlock->addLast(makeMls(dest, dest, arhs, alhs));
                        }
                        break;
                    }
                    case Ir::SHL:
                        info("shl generates extra load immediate.");
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        aInstr = makeShlRR(dest, alhs, arhs);
                        break;
                    case Ir::AND:
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        if (aRegNull(alhs) && aRegNull(arhs)) {
                            aInstr = new AInstrMov32(dest, alhsi & arhsi);
                        } else if (aRegNull(alhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::And, dest, arhs, alhsi);
                        } else if (aRegNull(arhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::And, dest, alhs, arhsi);
                        } else {
                            aInstr = makeAInstrALURR(AInstrALU::Type::And, dest, alhs, arhs);
                        }
                        break;
                    case Ir::OR:
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        if (aRegNull(alhs) && aRegNull(arhs)) {
                            aInstr = new AInstrMov32(dest, alhsi & arhsi);
                        } else if (aRegNull(alhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::Orr, dest, arhs, alhsi);
                        } else if (aRegNull(arhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::Orr, dest, alhs, arhsi);
                        } else {
                            aInstr = makeAInstrALURR(AInstrALU::Type::Orr, dest, alhs, arhs);
                        }
                        break;
                    case Ir::XOR:
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        if (aRegNull(alhs) && aRegNull(arhs)) {
                            aInstr = new AInstrMov32(dest, alhsi & arhsi);
                        } else if (aRegNull(alhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::Eor, dest, arhs, alhsi);
                        } else if (aRegNull(arhs)) {
                            aInstr = makeAInstrALURImm(AInstrALU::Type::Eor, dest, alhs, arhsi);
                        } else {
                            aInstr = makeAInstrALURR(AInstrALU::Type::Eor, dest, alhs, arhs);
                        }
                        break;
                    case Ir::ASHR:
                        info("shl generates extra load immediate.");
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        aInstr = makeAsrRR(dest, alhs, arhs);
                        break;
                    case Ir::LSHR:
                        info("shl generates extra load immediate.");
                        if (type == VReg::Type::Float) {
                            warn("bitwise operation on float is meaningless");
                        }
                        aInstr = makeShrRR(dest, alhs, arhs);
                        break;
                    case Ir::FNEG:
                        unimplemented();
                        break;
                    case Ir::FSUB:
                        unimplemented();
                        break;
                    case Ir::FADD:
                        unimplemented();
                        break;
                    case Ir::FREM:
                        unimplemented();
                        break;
                    case Ir::FDIV:
                        unimplemented();
                        break;
                    case Ir::EQ:
                    case Ir::NE:
                    case Ir::SGT:
                    case Ir::SGE:
                    case Ir::SLT:
                    case Ir::SLE: {
                        bool optimized = false;
                        if (mergeCmpBr) {
                            if (iinstr + 1 < instrs.size()) {
                                auto *nxtInstr = instrs[iinstr + 1];
                                Instanceof(condBr, Ir::BranchInstr *, nxtInstr) {
                                    auto block1 = irBB2ABB.find(condBr->get_bb_1())->second;
                                    if (condBr->get_bb_2() != nullptr) {
                                        if (binaryInstr->get_dst() == condBr->get_reg()) {
                                            optimized = true;// 满足优化条件

                                            auto block2 = irBB2ABB.find(condBr->get_bb_2())->second;
                                            if (operandType == VReg::Type::Int) {
                                                if (aRegNull(alhs) && aRegNull(arhs)) {
                                                    int cnstCond = evaluateCmp(binaryInstr->get_binary_type(),
                                                                               alhsi, arhsi);
                                                    if (cnstCond) {
                                                        curABlock->addLast(makeJump(block1));
                                                        curABlock->succBranch = block1;
                                                    } else {
                                                        curABlock->addLast(makeJump(block2));
                                                        curABlock->succBranch = block2;
                                                    }
                                                } else if (aRegNull(alhs)) {
                                                    // 左操作数是立即数
                                                    curABlock->addLast(new AInstrCMP{false, arhs, alhsi});
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2RevACmp(binaryInstr->get_binary_type()),
                                                            block1));
                                                    curABlock->addLast(makeJump(block2));
                                                    curABlock->succBranch = block1;
                                                    curABlock->failBranch = block2;
                                                } else if (aRegNull(arhs)) {
                                                    // 右操作数是立即数
                                                    curABlock->addLast(new AInstrCMP{false, alhs, arhsi});
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2ACmp(binaryInstr->get_binary_type()), block1));
                                                    curABlock->addLast(makeJump(block2));
                                                    curABlock->succBranch = block1;
                                                    curABlock->failBranch = block2;
                                                } else {
                                                    curABlock->addLast(makeAInstrCMPRR(alhs, arhs));
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2ACmp(binaryInstr->get_binary_type()), block1));
                                                    curABlock->addLast(makeJump(block2));
                                                    curABlock->succBranch = block1;
                                                    curABlock->failBranch = block2;
                                                }
                                            } else {
                                                if (aRegNull(alhs) && aRegNull(arhs)) {
                                                    // panic("fcmp 0,0 ?");
                                                    int cnstCond =
                                                            evaluateCmpF(binaryInstr->get_binary_type(), 0., 0.);
                                                    if (cnstCond) {
                                                        curABlock->addLast(makeJump(block1));
                                                        curABlock->succBranch = block1;
                                                    } else {
                                                        curABlock->addLast(makeJump(block2));
                                                        curABlock->succBranch = block2;
                                                    }
                                                } else if (aRegNull(alhs)) {
                                                    curABlock->addLast(new AInstrFloatCmp{
                                                            Cond::None,
                                                            arhs,
                                                            alhsi,

                                                    });
                                                    curABlock->addLast(new AInstrVMRS{});
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2RevACmp(binaryInstr->get_binary_type()),
                                                            block1));
                                                    curABlock->addLast(makeJump(block2));
                                                } else if (aRegNull(arhs)) {
                                                    curABlock->addLast(new AInstrFloatCmp{
                                                            Cond::None,
                                                            alhs,
                                                            arhsi,

                                                    });
                                                    curABlock->addLast(new AInstrVMRS{});
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2ACmp(binaryInstr->get_binary_type()), block1));
                                                    curABlock->addLast(makeJump(block2));
                                                } else {
                                                    curABlock->addLast(makeAInstrFloatCmp(alhs, arhs));
                                                    curABlock->addLast(new AInstrVMRS{});
                                                    curABlock->addLast(makeBranch(
                                                            irCmp2ACmp(binaryInstr->get_binary_type()), block1));
                                                    curABlock->addLast(makeJump(block2));
                                                }
                                                curABlock->succBranch = block1;
                                                curABlock->failBranch = block2;
                                            }

                                            iinstr++;// 跳过下一条br
                                        }
                                    }
                                }
                            }
                        }
                        if (!optimized) {
                            dest = vRegAllocator.alloc(VReg::Type::Int);
                            if (operandType == VReg::Type::Int) {
                                if (aRegNull(alhs) && aRegNull(arhs)) {
                                    curABlock->addLast(new AInstrMov32{
                                            dest,
                                            evaluateCmp(binaryInstr->get_binary_type(), alhsi, arhsi)});
                                } else if (aRegNull(alhs)) {
                                    curABlock->addLast(new AInstrCMP{false, arhs, alhsi});
                                    auto [f, l] = makeMovCond(
                                            irCmp2RevACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                } else if (aRegNull(arhs)) {
                                    curABlock->addLast(new AInstrCMP{false, alhs, arhsi});
                                    auto [f, l] =
                                            makeMovCond(irCmp2ACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                } else {
                                    curABlock->addLast(makeAInstrCMPRR(alhs, arhs));
                                    auto [f, l] =
                                            makeMovCond(irCmp2ACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                }

                            } else {
                                if (aRegNull(alhs) && aRegNull(arhs)) {
                                    //                                    panic("fcmp 0,0?");
                                    auto cond = evaluateCmpF(binaryInstr->get_binary_type(), 0., 0.);
                                    curABlock->addLast(new AInstrMov32(dest, cond));
                                } else if (aRegNull(alhs)) {
                                    curABlock->addLast(new AInstrFloatCmp{Cond::None, arhs, alhsi});
                                    curABlock->addLast(new AInstrVMRS{});
                                    auto [f, l] = makeMovCond(
                                            irCmp2RevACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                } else if (aRegNull(arhs)) {
                                    curABlock->addLast(new AInstrFloatCmp{Cond::None, alhs, arhsi});
                                    curABlock->addLast(new AInstrVMRS{});
                                    auto [f, l] =
                                            makeMovCond(irCmp2ACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                } else {
                                    curABlock->addLast(makeAInstrCMPRR(alhs, arhs));
                                    curABlock->addLast(new AInstrVMRS{});
                                    auto [f, l] =
                                            makeMovCond(irCmp2ACmp(binaryInstr->get_binary_type()), dest);
                                    curABlock->addLast(f);
                                    curABlock->addLast(l);
                                }
                            }
                            break;
                        }
                    }
                }
                if (aInstr) {
                    curABlock->addLast(aInstr);
                }
                if (!aRegNull(dest)) {
                    irReg2AReg.insert({d.get(), dest});
                }
            }
            Instanceof(load, Ir::LoadArray *, pInstruction) {
                auto &irVec = load->get_addr();
                auto pIrVec = irVec.get();
                AReg reg;
                if (globVarMap.find(pIrVec) != globVarMap.end()) {
                    // 全局变量
                    auto aData = globVarMap.find(pIrVec)->second;
                    if (cachedGlobalAddr.find(aData) != cachedGlobalAddr.end()) {
                        reg = cachedGlobalAddr[aData];
                    } else {
                        reg = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(new AInstrMovAddr32(reg, aData));
                        cachedGlobalAddr.insert({aData, reg});
                    }
                } else if (irStackVector2offset.find(pIrVec) !=
                           irStackVector2offset.end()) {
                    reg = vRegAllocator.alloc(VReg::Type::Int);
                    // 栈上数组
                    auto offset = irStackVector2offset.find(pIrVec)->second;
                    auto *pPse = new AInstrAddRImmPse(reg, get_sp(), offset);
                    curABlock->addLast(pPse);
                    curAFunction->loadLocalArrayInstructions.push_back(pPse);
                } else if (irParamVector2BaseReg.find(pIrVec) !=
                           irParamVector2BaseReg.end()) {
                    // 参数数组
                    reg = irParamVector2BaseReg.find(pIrVec)->second;
                } else {
                    panic("load from unknown vector.");
                }

                // 生成gep，类似Pascal数组模板
                int elem_offset = 1;
                const auto &arrayDims = load->get_addr()->get_dims();
                for (auto &l: arrayDims) {
                    elem_offset *= l ? l : 1;
                }
                // 索引取数组不是这条指令，只能取元素
                atriAssert(arrayDims.size() == load->get_indexs().size());
                const auto arrayDim = load->get_addr()->get_dims().size();
                VReg temp{std::get<VReg>(reg)};
                for (int i = 0; i < arrayDim - 1; ++i) {
                    elem_offset /= arrayDims[i] ? arrayDims[i] : 1;
                    auto index = load->get_indexs()[i].get();
                    if (index->is_constant()) {
                        // 是常量的话可以算出来
                        int cindex;
                        if (index->get_type() == Ir::Float) {
                            warn("using float as index is weird:", index->get_num().float_num);
                            cindex = static_cast<int>(index->get_num().float_num);
                        }
                        // 可能存在优化机会，常量索引可以直接算出来，而不是生成一串add和加载立即数
                        cindex = index->get_num().int_num;
                        cindex = cindex * elem_offset * 4;// 4Bytes
                        VReg rindex = genLoadImm32(cindex);
                        VReg res = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(
                                makeAInstrALURR(AInstrALU::Type::Add, res, temp, rindex));
                        temp = res;
                    } else {
                        // 寄存器，
                        if (index->get_type() == Ir::Float) {
                            error("using float reg", index->get_id(), "as index");
                            panic("using float reg as index is not allowed");
                        }
                        VReg rindex = vRegAllocator.allocPlaceHolder(index);
                        VReg treg{};
                        if (elem_offset == 1) {
                            treg = rindex;
                        } else {
                            treg = vRegAllocator.alloc(VReg::Type::Int);
                            if (isPow2(elem_offset)) {
                                curABlock->addLast(makeMovBitRIMM(treg, rindex, ShiftType::LSL,
                                                                  calcCLZ(elem_offset)));
                            } else {
                                VReg relem_offset = genLoadImm32(elem_offset);
                                curABlock->addLast(makeMul(treg, rindex, relem_offset));
                            }
                        }

                        VReg res = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(
                                new AInstrALU{AInstrALU::Type::Add, Cond::None, false, res, temp,
                                              Operand2Reg{treg, ShiftType::LSL, 2}});
                        temp = res;
                    }
                }
                elem_offset /= arrayDims[arrayDim - 1] ? arrayDims[arrayDim - 1] : 1;
                auto index = load->get_indexs()[arrayDim - 1].get();
                atriAssert(elem_offset == 1);
                if (index->is_constant()) {
                    // 是常量的话可以算出来
                    int cindex;
                    if (index->get_type() == Ir::Float) {
                        warn("using float as index is weird:", index->get_num().float_num);
                        cindex = static_cast<int>(index->get_num().float_num);
                    }
                    // 可能存在优化机会，常量索引可以直接算出来，而不是生成一串add和加载立即数
                    cindex = index->get_num().int_num;
                    cindex = cindex * elem_offset * 4;// 4Bytes
                    indexHints.insert({load->get_dst().get(), IndexHint{temp, cindex}});
                } else {
                    // 寄存器，
                    if (index->get_type() == Ir::Float) {
                        error("using float reg", index->get_id(), "as index");
                        panic("using float reg as index is not allowed");
                    }
                    VReg rindex = vRegAllocator.allocPlaceHolder(index);
                    indexHints.insert({load->get_dst().get(), IndexHint{temp, rindex}});
                }
                irMemObject2AReg.insert({load->get_dst().get(), temp});
            }
            Instanceof(load, Ir::LoadInstr *, pInstruction) {
                auto &irDest = load->get_dst();
                if (irDest->get_type() == Ir::Int) {
                    if (irStackVar2offset.find(load->get_addr().get()) !=
                        irStackVar2offset.end()) {
                        VReg reg = vRegAllocator.alloc(VReg::Type::Int);
                        auto offset = irStackVar2offset.find(load->get_addr().get())->second;
                        auto *pPse = new AInstrAddRImmPse(reg, get_sp(), offset);
                        curABlock->addLast(pPse);
                        curAFunction->loadLocalArrayInstructions.push_back(pPse);
                        VReg dest = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(makeLdRImm(dest, reg, 0));
                        irReg2AReg.insert({irDest.get(), dest});
                    } else {
                        VReg dest = vRegAllocator.alloc(VReg::Type::Int);
                        const VReg &rn =
                                vRegAllocator.allocPlaceHolder(load->get_addr().get());
                        AInstrLS *pInstrLs = makeLdRImm(dest, rn, 0);
                        curABlock->addLast(pInstrLs);
                        irReg2AReg.insert({irDest.get(), dest});
                        loadStoreMemObj.push_back(pInstrLs);
                    }
                } else {
                    if (irStackVar2offset.find(load->get_addr().get()) !=
                        irStackVar2offset.end()) {
                        VReg reg = vRegAllocator.alloc(VReg::Type::Int);
                        auto offset = irStackVar2offset.find(load->get_addr().get())->second;
                        auto *pPse = new AInstrAddRImmPse(reg, get_sp(), offset);
                        curABlock->addLast(pPse);
                        curAFunction->loadLocalArrayInstructions.push_back(pPse);
                        VReg dest = vRegAllocator.alloc(VReg::Type::Float);
                        curABlock->addLast(makeLdRImmF(dest, reg, 0));
                        irReg2AReg.insert({irDest.get(), dest});
                    } else {
                        VReg dest = vRegAllocator.alloc(VReg::Type::Float);
                        const VReg &rn =
                                vRegAllocator.allocPlaceHolder(load->get_addr().get());
                        AInstrLS *pInstrLs = makeLdRImmF(dest, rn, 0);
                        curABlock->addLast(pInstrLs);
                        irReg2AReg.insert({irDest.get(), dest});
                        loadStoreMemObj.push_back(pInstrLs);
                    }
                }
            }
            Instanceof(load, Ir::LoadVector *, pInstruction) {
                auto &irVec = load->get_addr();
                auto pIrVec = irVec.get();
                AReg reg;
                if (globVarMap.find(pIrVec) != globVarMap.end()) {
                    // 全局变量
                    auto aData = globVarMap.find(pIrVec)->second;
                    if (cachedGlobalAddr.find(aData) != cachedGlobalAddr.end()) {
                        reg = cachedGlobalAddr[aData];
                    } else {
                        reg = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(new AInstrMovAddr32(reg, aData));
                        cachedGlobalAddr.insert({aData, reg});
                    }
                } else if (irStackVector2offset.find(pIrVec) !=
                           irStackVector2offset.end()) {
                    reg = vRegAllocator.alloc(VReg::Type::Int);
                    // 栈上数组
                    auto offset = irStackVector2offset.find(pIrVec)->second;
                    auto *pPse = new AInstrAddRImmPse(reg, get_sp(), offset);
                    curABlock->addLast(pPse);
                    curAFunction->loadLocalArrayInstructions.push_back(pPse);
                } else if (irParamVector2BaseReg.find(pIrVec) !=
                           irParamVector2BaseReg.end()) {
                    // 参数数组
                    reg = irParamVector2BaseReg.find(pIrVec)->second;
                } else {
                    panic("load from unknown vector.");
                }

                // 生成gep，类似Pascal数组模板
                int elem_offset = 1;
                const auto &arrayDims = load->get_addr()->get_dims();
                for (auto &l: arrayDims) {
                    elem_offset *= l ? l : 1;
                }
                // 索引取数组
                atriAssert(arrayDims.size() > load->get_indexs().size());
                const auto idxDim = load->get_indexs().size();
                VReg temp{std::get<VReg>(reg)};
                for (int i = 0; i < idxDim; ++i) {
                    elem_offset /= arrayDims[i] ? arrayDims[i] : 1;
                    auto index = load->get_indexs()[i].get();
                    if (index->is_constant()) {
                        // 是常量的话可以算出来
                        int cindex;
                        if (index->get_type() == Ir::Float) {
                            warn("using float as index is weird:", index->get_num().float_num);
                            cindex = index->get_num().float_num;
                        }
                        // 可能存在优化机会，常量索引可以直接算出来，而不是生成一串add和加载立即数
                        cindex = index->get_num().int_num;
                        cindex = cindex * elem_offset * 4;// 4Bytes
                        VReg rindex = genLoadImm32(cindex);
                        VReg res = vRegAllocator.alloc(VReg::Type::Int);
                        curABlock->addLast(
                                makeAInstrALURR(AInstrALU::Type::Add, res, temp, rindex));
                        temp = res;
                    } else {
                        // 寄存器，
                        if (index->get_type() == Ir::Float) {
                            error("using float reg", index->get_id(), "as index");
                            panic("using float reg as index is not allowed");
                        }
                        VReg rindex = vRegAllocator.allocPlaceHolder(index);
                        VReg treg{};
                        if (elem_offset == 1) {
                            treg = rindex;// * 1 没有作用
                        } else {
                            treg = vRegAllocator.alloc(VReg::Type::Int);
                            if (isPow2(elem_offset)) {
                                curABlock->addLast(makeMovBitRIMM(treg, rindex, ShiftType::LSL,
                                                                  calcCLZ(elem_offset)));
                            } else {
                                VReg relem_offset = genLoadImm32(elem_offset);
                                curABlock->addLast(makeMul(treg, rindex, relem_offset));
                            }
                        }
                        VReg res = vRegAllocator.alloc(VReg::Type::Int);

                        curABlock->addLast(
                                new AInstrALU{AInstrALU::Type::Add, Cond::None, false, res, temp,
                                              Operand2Reg{treg, ShiftType::LSL, 2}});
                        temp = res;
                    }
                }
                // 都是寄存器放地址，但是可能和变量名中“参数”不完全符合
                irParamVector2BaseReg.insert({load->get_dst().get(), temp});
            }
            Instanceof(load, Ir::LoadGlobal *, pInstruction) {
                auto &irDest = load->get_dst();
                auto &irAddr = load->get_addr();
                Ir::Reg *pIrAddr = irAddr.get();
                if (globVarMap.find(pIrAddr) == globVarMap.end()) {
                    panic("load from unknown global reg");
                }
                auto aData = globVarMap.find(pIrAddr)->second;
                AReg vReg;
                if (cachedGlobalAddr.find(aData) != cachedGlobalAddr.end()) {
                    vReg = cachedGlobalAddr[aData];
                } else {
                    vReg = vRegAllocator.alloc(VReg::Type::Int);
                    curABlock->addLast(new AInstrMovAddr32(vReg, aData));
                    cachedGlobalAddr.insert({aData, vReg});
                }
                AReg dest = VReg{};
                if (irDest->get_type() == Ir::Int || irDest->get_type() == Ir::Bool) {
                    dest = vRegAllocator.alloc(VReg::Type::Int);
                    curABlock->addLast(makeLdRImm(dest, vReg, 0));
                } else {
                    dest = vRegAllocator.alloc(VReg::Type::Float);
                    curABlock->addLast(makeLdRImmF(dest, vReg, 0));
                }

                irReg2AReg.insert({irDest.get(), dest});
            }
            Instanceof(store, Ir::StoreGlobal *, pInstruction) {
                auto &irVal = store->get_val();
                auto &irAddr = store->get_addr();
                Ir::Reg *pIrAddr = irAddr.get();
                AReg valReg = VReg{};
                if (globVarMap.find(pIrAddr) == globVarMap.end()) {
                    panic("store into unknown global reg");
                }
                auto aData = globVarMap.find(pIrAddr)->second;
                AReg vReg;
                if (cachedGlobalAddr.find(aData) != cachedGlobalAddr.end()) {
                    vReg = cachedGlobalAddr[aData];
                } else {
                    vReg = vRegAllocator.alloc(VReg::Type::Int);
                    curABlock->addLast(new AInstrMovAddr32(vReg, aData));
                    cachedGlobalAddr.insert({aData, vReg});
                }
                if (irVal->is_constant()) {
                    if (irVal->get_type() == Ir::Int || irVal->get_type() == Ir::Bool) {
                        int val = irVal->get_num().int_num;
                        valReg = genLoadImm32(val);
                    } else {
                        float val = irVal->get_num().float_num;
                        valReg = genLoadImm32F(val);
                    }
                } else {
                    valReg = vRegAllocator.allocPlaceHolder(irVal.get());
                }

                if (irVal->get_type() == Ir::Int || irVal->get_type() == Ir::Bool) {
                    curABlock->addLast(makeStrRImm(valReg, vReg, 0));
                } else {
                    curABlock->addLast(makeStrRImmF(valReg, vReg, 0));
                }
            }
            Instanceof(store, Ir::StoreInstr *, pInstruction) {
                auto &irVal = store->get_val();
                auto &irAddr = store->get_addr();
                auto *pIrAddr = irAddr.get();
                AReg destReg = VReg{};
                AReg valReg = VReg{};
                if (irStackVar2offset.find(pIrAddr) != irStackVar2offset.end()) {
                    auto offset = irStackVar2offset.find(store->get_addr().get())->second;
                    destReg = vRegAllocator.alloc(VReg::Type::Int);
                    auto *pPse = new AInstrAddRImmPse(destReg, get_sp(), offset);
                    curABlock->addLast(pPse);
                    curAFunction->loadLocalArrayInstructions.push_back(pPse);
                } else {
                    destReg = vRegAllocator.allocPlaceHolder(pIrAddr);
                }
                if (irVal->is_constant()) {
                    if (irVal->get_type() == Ir::Int || irVal->get_type() == Ir::Bool) {
                        int val = irVal->get_num().int_num;
                        valReg = genLoadImm32(val);
                    } else {
                        float val = irVal->get_num().float_num;
                        valReg = genLoadImm32F(val);
                    }
                } else {
                    valReg = vRegAllocator.allocPlaceHolder(irVal.get());
                }
                if (irVal->get_type() == Ir::Int || irVal->get_type() == Ir::Bool) {
                    AInstrLS *pLs = makeStrRImm(valReg, destReg, 0);
                    curABlock->addLast(pLs);
                    if (std::get<VReg>(destReg).isPlaceHolder()) {
                        loadStoreMemObj.push_back(pLs);
                    }
                } else {
                    AInstrLS *pLs = makeStrRImmF(valReg, destReg, 0);
                    curABlock->addLast(pLs);
                    if (std::get<VReg>(destReg).isPlaceHolder()) {
                        loadStoreMemObj.push_back(pLs);
                    }
                }
            }
            Instanceof(alloca, Ir::AllocaInstr *, pInstruction) {
                auto &memObj = alloca->get_addr();
                auto pMemObj1 = memObj.get();
                Instanceof(pVec, Ir::Vector *, pMemObj1) {
                    auto offset = curAFunction->stackArrayAllocated * 4;
                    irStackVector2offset.insert({pVec, offset});
                    curAFunction->stackArrayAllocated +=
                            calculateVectorSize(pVec->get_dims());
                }
                else Instanceof(pMemObj, Ir::MemObject *, pMemObj1) {
                    auto offset = curAFunction->stackArrayAllocated * 4;
                    irStackVar2offset.insert({pMemObj, offset});
                    curAFunction->stackArrayAllocated += 1;
                }
            }
            Instanceof(branch, Ir::BranchInstr *, pInstruction) {
                if (irBB2ABB.find(branch->get_bb_1()) == irBB2ABB.end()) {
                    panic("unknown BB to branch");
                }
                auto block1 = irBB2ABB.find(branch->get_bb_1())->second;
                if (branch->get_bb_2() == nullptr) {
                    // unconditional
                    curABlock->addLast(makeJump(block1));
                    curABlock->succBranch = block1;
                } else {
                    if (irBB2ABB.find(branch->get_bb_2()) == irBB2ABB.end()) {
                        panic("unknown BB to branch");
                    }
                    auto block2 = irBB2ABB.find(branch->get_bb_2())->second;
                    auto vReg = vRegAllocator.allocPlaceHolder(branch->get_reg().get());
                    curABlock->addLast(new AInstrCMP{false, vReg, 0});
                    curABlock->addLast(makeBranch(Cond::Ne, block1));
                    curABlock->addLast(makeJump(block2));
                    curABlock->succBranch = block1;
                    curABlock->failBranch = block2;
                }
            }
            Instanceof(call, Ir::CallInstr *, pInstruction) {
                curCallInstr = call;
                if (call->get_function()->is_lib() &&
                    mCompileUnit->atriBuiltinName.count(
                            call->get_function()->get_name())) {
                    mCompileUnit->genAtriBuiltin = true;
                }
                genCallInstr();
            }
            Instanceof(ret, Ir::Ret *, pInstruction) {

                if (!ret->is_void_flag()) {
                    const auto &retVal = ret->get_src().get();
                    if (retVal->is_constant()) {
                        if (retVal->get_type() == Ir::Int || retVal->get_type() == Ir::Bool) {
                            genLoadImm32R(get_reg_by_id(0), retVal->get_num().int_num);
                        } else {
                            genLoadImm32RF(get_reg_by_id(16), retVal->get_num().float_num);
                        }
                    } else {
                        if (retVal->get_type() == Ir::Int || retVal->get_type() == Ir::Bool) {
                            curABlock->addLast(makeMovR2R(
                                    get_reg_by_id(0), vRegAllocator.allocPlaceHolder(retVal)));
                        } else {
                            curABlock->addLast(makeVMovS2S(
                                    get_reg_by_id(16), vRegAllocator.allocPlaceHolder(retVal)));
                        }
                    }
                }
                genRecoverCtx();
                curABlock->addLast(makeRet());
            }
            Instanceof(cvt, Ir::Convert *, pInstruction) {
                auto pRegFrom = cvt->get_src().get();
                auto pRegTo = cvt->get_dst().get();
                if ((pRegTo->get_type() == Ir::Bool && pRegFrom->get_type() == Ir::Int) ||
                    (pRegFrom->get_type() == Ir::Bool && pRegTo->get_type() == Ir::Int)) {
                    if (pRegFrom->get_type() == Ir::Bool && pRegFrom->is_constant()) {
                        AReg vReg = genLoadImm32(pRegFrom->get_num().int_num);
                        irReg2AReg.insert({pRegTo, vReg});
                    } else {
                        irAliasCvt.insert({pRegTo, pRegFrom});
                    }
                } else {
                    if (pRegFrom->get_type() == Ir::Float &&
                        (pRegTo->get_type() == Ir::Int || pRegTo->get_type() == Ir::Bool)) {
                        // 一个常数的类型转换没有意义
                        if (pRegFrom->is_constant()) {
                            auto vReg = genLoadImm32((int) pRegFrom->get_num().float_num);
                            irReg2AReg.insert({pRegTo, vReg});
                        } else {
                            auto vReg = vRegAllocator.alloc(VReg::Type::Int);
                            auto vReg2 = vRegAllocator.alloc(VReg::Type::Float);
                            curABlock->addLast(new AInstrFloatUnary{
                                    AInstrFloatUnary::Type::CvtF2I, Cond::None, vReg2,
                                    vRegAllocator.allocPlaceHolder(pRegFrom)});
                            curABlock->addLast(
                                    new AInstrVMov{Cond::None, vReg, vReg2, AInstrVMov::Type::S2R});
                            irReg2AReg.insert({pRegTo, vReg});
                        }

                    } else if ((pRegFrom->get_type() == Ir::Int ||
                                pRegFrom->get_type() == Ir::Bool) &&
                               pRegTo->get_type() == Ir::Float) {
                        if (pRegFrom->is_constant()) {
                            auto vReg = genLoadImm32F((float) pRegFrom->get_num().int_num);
                            irReg2AReg.insert({pRegTo, vReg});
                        } else {
                            auto vReg = vRegAllocator.alloc(VReg::Type::Float);
                            auto vReg2 = vRegAllocator.alloc(VReg::Type::Float);
                            curABlock->addLast(new AInstrVMov{
                                    Cond::None, vReg2, vRegAllocator.allocPlaceHolder(pRegFrom),
                                    AInstrVMov::Type::R2S});
                            curABlock->addLast(new AInstrFloatUnary{
                                    AInstrFloatUnary::Type::CvtI2F, Cond::None, vReg, vReg2});
                            irReg2AReg.insert({pRegTo, vReg});
                        }

                    } else {
                        error("meaningless convert from ", pRegFrom->get_type(), " to ",
                              pRegTo->get_type());
                        panic("convert");
                    }
                }
            }
            Instanceof(phi, Ir::PhiInstr *, pInstruction) {
                auto aPhi = new AInstrPhi{};
                AReg dest =
                        vRegAllocator.alloc(IRType2VRegType(phi->get_dst()->get_type()));
                aPhi->dest = dest;
                irReg2AReg.insert({phi->get_dst().get(), dest});
                for (auto &[block, reg]: phi->get_bb_2_value()) {
                    //                    if (reg->is_constant()) {
                    //                        if (reg->get_type() == Ir::Float) {
                    //                            auto vReg =
                    //                            genLoadImm32F(reg->get_num().float_num);
                    //                            aPhi->pairs.emplace_back(vReg,
                    //                            irBB2ABB[block]);
                    //                        } else if (reg->get_type() == Ir::Int ||
                    //                        reg->get_type() == Ir::Bool) {
                    //                            auto vReg =
                    //                            genLoadImm32(reg->get_num().int_num);
                    //                            aPhi->pairs.emplace_back(vReg,
                    //                            irBB2ABB[block]);
                    //                        }
                    //                    } else {
                    auto vReg = vRegAllocator.allocPlaceHolder(reg.get());
                    aPhi->pairs.emplace_back(vReg, irBB2ABB[block]);
                    //                    }
                }
                curABlock->addLast(aPhi);
            }
        }
    }

    /**
 * @brief 处理传参
 */
    void ACodeGen::resolveParam() {
        int intCnt = 0, floatCnt = 0;// 超过寄存器数量的指令从栈中加载

        for (auto pparam = irFunction->get_params().begin();
             pparam != irFunction->get_params().end(); pparam++) {
            auto &param = *pparam;
            Instanceof(reg, Ir::Reg *, param.get()) {
                if (reg->get_type() == Ir::Int || reg->get_type() == Ir::Bool) {
                    if (intCnt < PARAM_REG_CNT) {
                        VReg vReg = vRegAllocator.alloc(VReg::Type::Int);
                        auto mov = makeMovR2R(vReg, get_reg_by_id(intCnt));
                        curABlock->addLast(mov);
                        intCnt++;
                        irReg2AReg.insert({reg, vReg});
                    } else {
                        AReg vReg = vRegAllocator.alloc(VReg::Type::Int);
                        AReg sp = get_ip();
                        // 这里没有考虑范围，且实际上是相对于caller的sp
                        // TODO: 在寄存器分配后重写这些指令
                        auto ldr =
                                makeLdRImm(vReg, sp, calculateStackParamOffset(intCnt, floatCnt));
                        curABlock->addLast(ldr);
                        curAFunction->loadParamInstructions.push_back(ldr);
                        intCnt++;
                        irReg2AReg.insert({reg, vReg});
                    }
                } else {
                    if (floatCnt < FLOAT_REG_CNT) {
                        AReg vReg = vRegAllocator.alloc(VReg::Type::Float);
                        AReg argReg = get_freg_by_id(floatCnt);
                        auto mov = makeVMovS2S(vReg, argReg);
                        curABlock->addLast(mov);
                        floatCnt++;
                        irReg2AReg.insert({reg, vReg});
                    } else {
                        AReg vReg = vRegAllocator.alloc(VReg::Type::Float);
                        AReg sp = get_ip();
                        // 这里没有考虑范围，且实际上是相对于caller的sp
                        // TODO: 在寄存器分配后重写这些指令
                        auto ldr = makeLdRImmF(vReg, sp,
                                               calculateStackParamOffset(intCnt, floatCnt));
                        curABlock->addLast(ldr);
                        curAFunction->loadParamInstructions.push_back(ldr);
                        floatCnt++;
                        irReg2AReg.insert({reg, vReg});
                    }
                }
            }
            else Instanceof(vec, Ir::Vector *, param.get()) {
                VReg vReg = vRegAllocator.alloc(VReg::Type::Int);
                if (intCnt < PARAM_REG_CNT) {
                    auto mov = makeMovR2R(vReg, get_reg_by_id(intCnt));
                    curABlock->addLast(mov);
                    intCnt++;
                } else {
                    Reg sp = get_ip();
                    // 这里没有考虑范围，且实际上是相对于caller的sp
                    // TODO: 在寄存器分配后重写这些指令
                    auto ldr =
                            makeLdRImm(vReg, sp, calculateStackParamOffset(intCnt, floatCnt));
                    curABlock->addLast(ldr);
                    curAFunction->loadParamInstructions.push_back(ldr);
                    intCnt++;
                }
                irParamVector2BaseReg.insert({vec, vReg});
            }
            else Instanceof(memobj, Ir::MemObject *, param.get()) {
                VReg vReg = vRegAllocator.alloc(VReg::Type::Int);
                if (intCnt < PARAM_REG_CNT) {
                    auto mov = makeMovR2R(vReg, get_reg_by_id(intCnt));
                    curABlock->addLast(mov);
                    intCnt++;
                } else {
                    Reg sp = get_ip();
                    // 这里没有考虑范围，
                    auto ldr =
                            makeLdRImm(vReg, sp, calculateStackParamOffset(intCnt, floatCnt));
                    curAFunction->loadParamInstructions.push_back(ldr);
                    curABlock->addLast(ldr);
                    intCnt++;
                }
                irMemObject2AReg.insert({memobj, vReg});
            }
        }
    }

    void ACodeGen::genProtectCtx() {
        auto savedRegs = get_saved_reg_wrapped();
        //        savedRegs.emplace_back(get_reg_by_id(14)); // LR
        auto savedFRegs = get_saved_freg_wrapped();
        std::vector<AReg> allSavedRegs{savedRegs};
        allSavedRegs.insert(allSavedRegs.cend(), savedFRegs.begin(),
                            savedFRegs.end());
        curABlock->addLast(new AInstrLiveInfoHint{{}, allSavedRegs});
        auto pushLR = makePush({get_lr()});
        curABlock->addLast(makeMovR2R(get_ip(), get_sp()));
        curABlock->addLast(pushLR);
        auto *pInstrSpAdd = new AInstrSpAdd{0};
        curABlock->addLast(pInstrSpAdd);

        for (auto &reg: savedRegs) {
            const VReg rd = vRegAllocator.alloc(VReg::Type::Int);
            savedRegMap.insert({reg, rd});
            curABlock->addLast(makeMovR2R(rd, reg));
        }
        for (auto &reg: savedFRegs) {
            const VReg rd = vRegAllocator.alloc(VReg::Type::Float);
            savedRegMap.insert({reg, rd});
            curABlock->addLast(makeVMovS2S(rd, reg));
        }
        //        auto pushInstr = makePush(savedRegs);
        //        curABlock->addLast(pushInstr);

        //        auto pushFInstr = makePushF(savedFRegs);
        //        curABlock->addLast(pushFInstr);

        curAFunction->reserveLocal = pInstrSpAdd;
    }

    VReg ACodeGen::genLoadImm32(int num) {
        if (isImm8m(num)) {
            warn("using mov,movt to load imm8m is suboptimal.");
        }
        VReg vReg = vRegAllocator.alloc(VReg::Type::Int);
        auto mov = new AInstrMov32(vReg, num);
        curABlock->addLast(mov);

        return vReg;
    }

    VReg ACodeGen::genLoadImm32F(float num) {
        VReg vReg = vRegAllocator.alloc(VReg::Type::Float);
        auto vldr = makeLoadF(vReg, num);
        curABlock->addLast(vldr);
        return vReg;
    }

    VReg ACodeGen::genReg(Ir::Reg *reg) {
        if (reg->is_constant()) {
            if (reg->get_type() == Ir::Int || reg->get_type() == Ir::Bool) {
                return genLoadImm32(reg->get_num().int_num);
            } else {
                return genLoadImm32F(reg->get_num().float_num);
            }
        } else {
            return vRegAllocator.allocPlaceHolder(reg);
        }
    }

    void ACodeGen::genCallInstr() {
        int intCnt = 0, floatCnt = 0;// 超过寄存器数量的指令存入栈中
        int ipOffset = 0;
        auto saveIp = curABlock->getLast();
        auto calleeFunction = curCallInstr->get_function();
        bool shouldAlign{false};// 需要执行8byte对齐
        std::vector<Ir::Type *> stackParams{};
        for (auto pparam = curCallInstr->get_params().begin();
             pparam != curCallInstr->get_params().end(); pparam++) {
            auto &param = *pparam;
            Instanceof(reg, Ir::Reg *, param.get()) {
                if (reg->get_type() == Ir::Int || reg->get_type() == Ir::Bool) {
                    if (intCnt < PARAM_REG_CNT) {
                        intCnt++;
                    } else {
                        stackParams.push_back(param.get());
                        intCnt++;
                    }
                } else {
                    if (floatCnt < FLOAT_REG_CNT) {
                        floatCnt++;
                    } else {
                        stackParams.push_back(param.get());
                        floatCnt++;
                    }
                }
            }
            else Instanceof(vec, Ir::Vector *, param.get()) {
                if (intCnt < PARAM_REG_CNT) {
                    intCnt++;
                } else {
                    stackParams.push_back(param.get());
                    intCnt++;
                }
            }
            else Instanceof(memobj, Ir::MemObject *, param.get()) {
                if (intCnt < PARAM_REG_CNT) {
                    intCnt++;
                } else {
                    stackParams.push_back(param.get());
                    intCnt++;
                }
            }
        }
        if (stackParams.size() & 1) {
            shouldAlign = true;
            //            curABlock->addLast(new AInstrALU{
            //                    AInstrALU::Type::Sub,
            //                    Cond::None,
            //                    false,
            //                    get_sp(),
            //                    get_sp(),
            //                    WORD_SIZE
            //            }); // 8byte align
        }
        intCnt = 0;
        floatCnt = 0;
        for (auto pparam = curCallInstr->get_params().begin();
             pparam != curCallInstr->get_params().end(); pparam++) {
            auto &param = *pparam;
            if (calculateStackParamOffset(intCnt, floatCnt) + WORD_SIZE - ipOffset >
                1020) {
                curABlock->addLast(makeAInstrALURImm(AInstrALU::Type::Add, AReg{get_ip()},
                                                     AReg{get_ip()}, 1020));
                ipOffset += 1020;
            }
            Instanceof(reg, Ir::Reg *, param.get()) {
                AReg vReg = VReg{};
                if (reg->is_constant()) {
                    if (reg->get_type() == Ir::Int) {
                        vReg = genLoadImm32(reg->get_num().int_num);
                    } else {
                        vReg = genLoadImm32F(reg->get_num().float_num);
                    }
                } else {
                    vReg = vRegAllocator.allocPlaceHolder(reg);
                }
                if (reg->get_type() == Ir::Int || reg->get_type() == Ir::Bool) {
                    if (intCnt < PARAM_REG_CNT) {
                        auto mov = makeMovR2R(get_reg_by_id(intCnt), vReg);
                        curABlock->addLast(mov);
                        intCnt++;
                    } else {
                        AReg sp = get_ip();
                        auto str = makeStrRImm(
                                vReg, sp, calculateStackParamOffset(intCnt, floatCnt) - ipOffset);
                        curABlock->addLast(str);
                        intCnt++;
                    }
                } else {
                    if (floatCnt < FLOAT_REG_CNT) {
                        AReg argReg = get_freg_by_id(floatCnt);
                        auto mov = makeVMovS2S(argReg, vReg);
                        curABlock->addLast(mov);
                        floatCnt++;
                    } else {
                        AReg sp = get_ip();
                        auto str = makeStrRImmF(
                                vReg, sp, calculateStackParamOffset(intCnt, floatCnt) - ipOffset);
                        curABlock->addLast(str);
                        floatCnt++;
                    }
                }
            }
            else Instanceof(vec, Ir::Vector *, param.get()) {
                VReg vReg = vRegAllocator.allocPlaceHolder(vec);
                if (intCnt < PARAM_REG_CNT) {
                    auto mov = makeMovR2R(get_reg_by_id(intCnt), vReg);
                    curABlock->addLast(mov);
                    intCnt++;
                } else {
                    Reg sp = get_ip();
                    auto str = makeStrRImm(
                            vReg, sp, calculateStackParamOffset(intCnt, floatCnt) - ipOffset);
                    curABlock->addLast(str);
                    intCnt++;
                }
            }
            else Instanceof(memobj, Ir::MemObject *, param.get()) {
                VReg vReg = vRegAllocator.allocPlaceHolder(memobj);
                if (intCnt < PARAM_REG_CNT) {
                    auto mov = makeMovR2R(get_reg_by_id(intCnt), vReg);
                    curABlock->addLast(mov);
                    intCnt++;
                } else {
                    Reg sp = get_ip();
                    auto str = makeStrRImm(
                            vReg, sp, calculateStackParamOffset(intCnt, floatCnt) - ipOffset);
                    curABlock->addLast(str);
                    intCnt++;
                }
            }
        }
        int stackUsed = calculateStackParamOffset(intCnt, floatCnt) / WORD_SIZE;
        if (shouldAlign) {
            stackUsed += 1;
        }
        if (stackUsed > 0 && stackUsed < 256) {
            saveIp->insertAfterMe(new AInstrALU{AInstrALU::Type::Sub, Cond::None, false,
                                                get_ip(), get_sp(),
                                                stackUsed * WORD_SIZE});
            curABlock->addLast(new AInstrALU{AInstrALU::Type::Sub, Cond::None, false,
                                             get_sp(), get_sp(),
                                             stackUsed * WORD_SIZE});
        } else if (stackUsed >= 256) {
            VReg vReg = vRegAllocator.alloc(VReg::Type::Int);
            AInstr *ld1 = new AInstrMov32(vReg, stackUsed * WORD_SIZE);
            saveIp->insertAfterMe(ld1);
            ld1->insertAfterMe(
                    makeAInstrALURR(AInstrALU::Type::Sub, get_ip(), get_sp(), vReg));
            VReg vReg1 = genLoadImm32(stackUsed * WORD_SIZE);
            curABlock->addLast(
                    makeAInstrALURR(AInstrALU::Type::Sub, get_sp(), get_sp(), vReg1));
        }
        if (funcMap.find(curCallInstr->get_function()) == funcMap.end()) {
            panic("calling unknown function");
        }
        AInstrCall::ReturnValType returnValType{AInstrCall::ReturnValType::None};
        if (!curCallInstr->is_void_flag()) {
            if (curCallInstr->get_dst()->get_type() == Ir::Int) {
                returnValType = AInstrCall::ReturnValType::Int;
            } else {
                returnValType = AInstrCall::ReturnValType::Float;
            }
        }
        std::vector<AReg> paramPassReg{};
        paramPassReg.reserve(intCnt + floatCnt);
        for (int i = 0; i < std::min(intCnt, PARAM_REG_CNT); ++i) {
            paramPassReg.emplace_back(get_reg_by_id(i));
        }
        for (int i = 0; i < std::min(floatCnt, FLOAT_REG_CNT); ++i) {
            paramPassReg.emplace_back(get_freg_by_id(i));
        }

        auto callingFunc = funcMap[curCallInstr->get_function()];
        curABlock->addLast(makeCall(funcMap[curCallInstr->get_function()],
                                    returnValType, std::move(paramPassReg)));

        // 恢复参数栈
        if (stackUsed > 0 && stackUsed < 256) {
            curABlock->addLast(new AInstrALU{AInstrALU::Type::Add, Cond::None, false,
                                             get_sp(), get_sp(),
                                             stackUsed * WORD_SIZE});
        } else if (stackUsed >= 256) {
            VReg vReg1 = genLoadImm32(stackUsed * WORD_SIZE);
            curABlock->addLast(
                    makeAInstrALURR(AInstrALU::Type::Add, get_sp(), get_sp(), vReg1));
        }
        // 加载返回值
        if (!curCallInstr->is_void_flag()) {
            if (curCallInstr->get_dst()->get_type() == Ir::Int) {
                AReg vReg = vRegAllocator.alloc(VReg::Type::Int);
                curABlock->addLast(makeMovR2R(vReg, get_reg_by_id(0)));
                irReg2AReg.insert({curCallInstr->get_dst().get(), vReg});
            } else {
                AReg vReg = vRegAllocator.alloc(VReg::Type::Float);
                AReg s0 = get_freg_by_id(0);
                curABlock->addLast(makeVMovS2S(vReg, s0));
                irReg2AReg.insert({curCallInstr->get_dst().get(), vReg});
            }
        }
    }

    void ACodeGen::genRecoverCtx() {

        auto savedRegs = get_saved_reg_wrapped();
        //        savedRegs.emplace_back(get_reg_by_id(14)); // LR
        //        auto popInstr = makePop(savedRegs);
        //        curABlock->addLast(popInstr);
        auto savedFRegs = get_saved_freg_wrapped();
        //        auto popFInstr = makePopF(savedFRegs);
        //        curABlock->addLast(popFInstr);
        std::vector<AReg> allSavedRegs{savedRegs};
        allSavedRegs.insert(allSavedRegs.cend(), savedFRegs.begin(),
                            savedFRegs.end());
        for (auto &reg: savedRegs) {
            curABlock->addLast(makeMovR2R(reg, savedRegMap[reg]));
        }
        for (auto &reg: savedFRegs) {
            curABlock->addLast(makeVMovS2S(reg, savedRegMap[reg]));
        }
        auto *pInstrSpAdd = new AInstrSpAdd{0};
        curABlock->addLast(pInstrSpAdd);
        curAFunction->cleanLocal = pInstrSpAdd;
        auto popLR = makePop({get_lr()});
        curABlock->addLast(popLR);
        curABlock->addLast(new AInstrLiveInfoHint{allSavedRegs, {}});
    }

    void ACodeGen::placeHolderReplace() {
        //        for (auto [_,hint]: indexHints) {
        //            if (std::holds_alternative<AReg>(hint.index)) {
        //                hint.index =
        //                tryReplacePlaceHolder(std::get<AReg>(hint.index));
        //            }
        //            hint.base = tryReplacePlaceHolder(hint.base);
        //        }
        for (auto block: *curAFunction) {
            for (auto instr: *block) {
                if (auto aluInstr = dynamic_cast<AInstrALU *>(instr)) {
                    aluInstr->rn = tryReplacePlaceHolder(aluInstr->rn);
                    aluInstr->operand2 = tryReplacePlaceHolderOp2(aluInstr->operand2);
                }
                if (auto lsInstr = dynamic_cast<AInstrLS *>(instr)) {
                    lsInstr->rn = tryReplacePlaceHolder(lsInstr->rn);
                    lsInstr->rd = tryReplacePlaceHolder(lsInstr->rd);
                    if (std::holds_alternative<RegLS>(lsInstr->operand2)) {
                        auto &regLs = std::get<RegLS>(lsInstr->operand2);
                        lsInstr->operand2 = RegLS{tryReplacePlaceHolder(regLs.rm),
                                                  regLs.shiftType, regLs.shift, regLs.minus};
                    }
                }
                if (auto movInstr = dynamic_cast<AInstrMov *>(instr)) {
                    movInstr->operand2 = tryReplacePlaceHolderOp2(movInstr->operand2);
                }
                if (auto vMovInstr = dynamic_cast<AInstrVMov *>(instr)) {
                    vMovInstr->rm = tryReplacePlaceHolder(vMovInstr->rm);
                }
                if (auto cmpInstr = dynamic_cast<AInstrCMP *>(instr)) {
                    cmpInstr->rm = tryReplacePlaceHolder(cmpInstr->rm);
                    cmpInstr->operand2 = tryReplacePlaceHolderOp2(cmpInstr->operand2);
                }
                if (auto vCmpInstr = dynamic_cast<AInstrFloatCmp *>(instr)) {
                    vCmpInstr->rm = tryReplacePlaceHolder(vCmpInstr->rm);
                    if (std::holds_alternative<AReg>(vCmpInstr->rn)) {
                        vCmpInstr->rn = tryReplacePlaceHolder(std::get<AReg>(vCmpInstr->rn));
                    }
                }
                if (auto mul = dynamic_cast<AInstrMul *>(instr)) {
                    mul->rm = tryReplacePlaceHolder(mul->rm);
                    mul->rs = tryReplacePlaceHolder(mul->rs);
                }
                if (auto mla = dynamic_cast<AInstrMla *>(instr)) {
                    mla->rm = tryReplacePlaceHolder(mla->rm);
                    mla->rn = tryReplacePlaceHolder(mla->rn);
                    mla->rs = tryReplacePlaceHolder(mla->rs);
                    mla->rd = tryReplacePlaceHolder(mla->rd);
                }
                if (auto smmul = dynamic_cast<AInstrSMMUL *>(instr)) {
                    smmul->rs = tryReplacePlaceHolder(smmul->rs);
                    smmul->rm = tryReplacePlaceHolder(smmul->rm);
                    smmul->rd = tryReplacePlaceHolder(smmul->rd);
                }

                if (auto div = dynamic_cast<AInstrSDiv *>(instr)) {
                    div->rm = tryReplacePlaceHolder(div->rm);
                    div->rn = tryReplacePlaceHolder(div->rn);
                }
                if (auto fbinary = dynamic_cast<AInstrFloatBinary *>(instr)) {
                    fbinary->rm = tryReplacePlaceHolder(fbinary->rm);
                    fbinary->rn = tryReplacePlaceHolder(fbinary->rn);
                }
                if (auto funary = dynamic_cast<AInstrFloatUnary *>(instr)) {
                    funary->rm = tryReplacePlaceHolder(funary->rm);
                }
                if (auto phi = dynamic_cast<AInstrPhi *>(instr)) {
                    for (auto iter = phi->pairs.begin(); iter != phi->pairs.end(); iter++) {
                        if (auto vReg = std::get_if<VReg>(&iter->first)) {
                            if (auto reg =
                                        dynamic_cast<Ir::Reg *>(vRegAllocator.getHolding(*vReg))) {
                                if (reg->is_constant()) {
                                    continue;
                                }
                            }
                        }
                        iter->first = tryReplacePlaceHolder(iter->first);
                    }
                }
            }
        }
    }

    AReg ACodeGen::tryReplacePlaceHolder(AReg original) {
        if (std::holds_alternative<Reg>(original)) {
            return original;
        }
        VReg &vReg = std::get<VReg>(original);
        if (vReg.id > 0) {
            return original;
        }
        auto holding = vRegAllocator.getHolding(vReg);

        Instanceof(reg, Ir::Reg *, holding) {
            // 处理convert bool
            while (irAliasCvt.count(reg)) {
                reg = irAliasCvt[reg];
            }
            if (irReg2AReg.count(reg)) {
                return irReg2AReg.find(reg)->second;
            } else {
                error("on replacing Ir::Reg:", reg->get_id());
                panic("unknown reg placeholder");
            }
        }
        Instanceof(vec, Ir::Vector *, holding) {
            if (irParamVector2BaseReg.count(vec)) {
                return irParamVector2BaseReg.find(vec)->second;
            } else {
                panic("unknown vec placeholder");
            }
        }
        Instanceof(memObj, Ir::MemObject *, holding) {
            if (irMemObject2AReg.count(memObj)) {
                return irMemObject2AReg.find(memObj)->second;
            } else {
                panic("unknown memObj placeholder");
            }
        }
        panic("unknown placeholder");
        // return original;
    }

    Operand2 ACodeGen::tryReplacePlaceHolderOp2(Operand2 operand2) {
        if (std::holds_alternative<int>(operand2)) {
            return operand2;
        }
        auto &operand2Reg = std::get<Operand2Reg>(operand2);
        auto base = tryReplacePlaceHolder(operand2Reg.base);
        if (std::holds_alternative<AReg>(operand2Reg.shift)) {
            auto shift = tryReplacePlaceHolder(std::get<AReg>(operand2Reg.shift));
            return Operand2Reg{
                    base,
                    operand2Reg.shiftType,
                    shift,
            };
        }
        return Operand2Reg{
                base,
                operand2Reg.shiftType,
                operand2Reg.shift,
        };
    }

    /**
 * @brief 关键边拆分实现的SSA destruction
 *
 * SSA book Page 37
 */
    void ACodeGen::destructPhi() {
        for (auto bb: curAFunction->toVector()) {
            std::unordered_map<ABlock *, AInstrParallelCopy *> bb2pc{};

            for (auto precBB:
                 std::vector<ABlock *>{bb->prec.begin(), bb->prec.end()}) {
                auto pc = new AInstrParallelCopy{};
                if (bb->prec.size() > 1 && precBB->succ.size() > 1) {
                    auto *newBB = new ABlock{".Lspl"s + std::to_string(blockCnt++)};
                    newBB->succBranch = bb;
                    precBB->succ.erase(bb);
                    precBB->succ.insert(newBB);
                    bb->prec.erase(precBB);
                    bb->prec.insert(newBB);
                    replaceJumpTarget(precBB, bb, newBB);
                    replacePhiBlock(bb, precBB, newBB);
                    auto j = makeJump(bb);
                    newBB->addLast(pc);
                    newBB->addLast(j);
                    newBB->prec.insert(precBB);
                    newBB->succ.insert(bb);
                    bb->insertAfterMe(newBB);
                    bb2pc.insert({newBB, pc});
                } else {
                    auto last = precBB->getLast();
                    while (last->getPrev() != nullptr &&
                           (dynamic_cast<AInstrB *>(last->getPrev()) ||
                            dynamic_cast<AInstrCMP *>(last->getPrev()))) {
                        last = last->getPrev();
                    }
                    last->insertBeforeMe(pc);
                    bb2pc.insert({precBB, pc});
                }
            }
            for (auto instr: bb->toVector()) {
                if (auto phi = dynamic_cast<AInstrPhi *>(instr)) {
                    for (auto &[reg, block]: phi->pairs) {
                        bb2pc[block]->copy.emplace_back(phi->dest, reg);
                    }
                    phi->removeMe();
                    delete phi;
                }
            }
        }
        // 并行复制串行化
        //        info(mCompileUnit->toAssemblyString());
        for (auto bb: *curAFunction) {
            for (auto instr: bb->toVector()) {
                if (auto pCopy = dynamic_cast<AInstrParallelCopy *>(instr)) {
                    std::vector<AInstr *> seq{};
                    while (std::any_of(
                            pCopy->copy.begin(), pCopy->copy.end(),
                            [&](auto &pair) { return pair.first != pair.second; })) {
                        auto s = std::find_if(
                                pCopy->copy.begin(), pCopy->copy.end(), [&](auto &p) {
                                    return p.first != p.second &&
                                           std::none_of(
                                                   pCopy->copy.begin(), pCopy->copy.end(),
                                                   [&](auto &p1) { return p1.second == p.first; });
                                });
                        if (s != pCopy->copy.end()) {
                            auto &pair = *s;
                            if (aReg2Type(pair.first) == VReg::Type::Int) {
                                auto cnst = isHoldingConstantInt(pair.second);
                                if (cnst.has_value()) {
                                    seq.push_back(new AInstrMov32(pair.first, cnst.value()));
                                } else {
                                    seq.push_back(makeMovR2R(pair.first, pair.second));
                                }

                            } else {
                                auto cnst = isHoldingConstantFloat(pair.second);
                                if (cnst.has_value()) {
                                    seq.push_back(makeLoadF(pair.first, cnst.value()));
                                } else {
                                    seq.push_back(makeVMovS2S(pair.first, pair.second));
                                }
                            }
                            pCopy->copy.erase(s);
                        } else {
                            auto pPair =
                                    std::find_if(pCopy->copy.begin(), pCopy->copy.end(),
                                                 [&](auto &p) { return p.first != p.second; });
                            atriAssert(pPair != pCopy->copy.end());
                            auto &pair = *pPair;
                            AReg aReg = vRegAllocator.alloc(aReg2Type(pair.second));
                            if (aReg2Type(pair.first) == VReg::Type::Int) {
                                auto cnst = isHoldingConstantInt(pair.second);
                                if (cnst.has_value()) {
                                    seq.push_back(new AInstrMov32(aReg, cnst.value()));
                                } else {
                                    seq.push_back(makeMovR2R(aReg, pair.second));
                                }

                            } else {
                                auto cnst = isHoldingConstantFloat(pair.second);
                                if (cnst.has_value()) {
                                    seq.push_back(makeLoadF(aReg, cnst.value()));
                                } else {
                                    seq.push_back(makeVMovS2S(aReg, pair.second));
                                }
                            }
                            pCopy->copy.erase(pPair);
                            pCopy->copy.emplace_back(pair.first, aReg);
                        }
                    }
                    for (auto &move: seq) {
                        instr->insertBeforeMe(move);
                    }
                    instr->removeMe();
                    delete instr;
                }
            }
        }
    }

    void ACodeGen::replaceJumpTarget(ABlock *inWhich, ABlock *old, ABlock *nnew) {
        for (auto instr: *inWhich) {
            if (auto b = dynamic_cast<AInstrB *>(instr)) {
                if (auto pt = std::get_if<ABlock *>(&b->target)) {
                    if ((*pt) == old) {
                        if (b->cond != Cond::None) {
                            inWhich->succBranch = nnew;
                        } else {
                            inWhich->failBranch = nnew;
                        }
                        *pt = nnew;
                    }
                }
            }
        }
    }

    void ACodeGen::replacePhiBlock(ABlock *inWhich, ABlock *old, ABlock *nnew) {
        for (auto instr: *inWhich) {
            if (auto phi = dynamic_cast<AInstrPhi *>(instr)) {
                decltype(phi->pairs) pairs{};
                for (auto [val, block]: phi->pairs) {
                    if (block == old) {
                        pairs.emplace_back(val, nnew);
                    } else {
                        pairs.emplace_back(val, block);
                    }
                }
                phi->pairs = pairs;
            }
        }
    }

    void ACodeGen::genLoadImm32R(const AReg &aReg, int num) {
        if (isImm8m(num)) {
            warn("using mov,movt to load imm8m is suboptimal.");
        }
        auto mov = new AInstrMov32(aReg, num);
        curABlock->addLast(mov);
    }

    void ACodeGen::genLoadImm32RF(const AReg &aReg, float num) {
        auto vldr = makeLoadF(aReg, num);
        curABlock->addLast(vldr);
    }

    std::optional<int> ACodeGen::isHoldingConstantInt(AReg &aReg) {
        if (auto pVReg = std::get_if<VReg>(&aReg)) {
            if (pVReg->isPlaceHolder()) {
                if (auto pIrReg =
                            dynamic_cast<Ir::Reg *>(vRegAllocator.getHolding(*pVReg))) {
                    if (pIrReg->is_constant() &&
                        (pIrReg->get_type() == Ir::Int || pIrReg->get_type() == Ir::Bool)) {
                        return pIrReg->get_num().int_num;
                    }
                }
            }
        }
        return std::nullopt;
    }

    std::optional<float> ACodeGen::isHoldingConstantFloat(AReg &aReg) {
        if (auto pVReg = std::get_if<VReg>(&aReg)) {
            if (pVReg->isPlaceHolder()) {
                if (auto pIrReg =
                            dynamic_cast<Ir::Reg *>(vRegAllocator.getHolding(*pVReg))) {
                    if (pIrReg->is_constant() && (pIrReg->get_type() == Ir::Float)) {
                        return pIrReg->get_num().float_num;
                    }
                }
            }
        }
        return std::nullopt;
    }

    void ACodeGen::genOptimizedDiv(const AReg &dest, const AReg &left, int imm) {
        int abs = imm > 0 ? imm : -imm;
        //        atriAssert(abs != 0 && abs != 1);
        atriAssert(abs != 0);
        if (imm == 1) {
            curABlock->addLast(makeMovR2R(dest, left));
            return;
        }
        if (imm == -1) {
            curABlock->addLast(makeAInstrALURImm(AInstrALU::Type::Rsb, dest, left, 0));
            return;
        }
        if ((abs & (abs - 1)) == 0) {
            int l = calcCLZ(abs);
            VReg v = vRegAllocator.alloc(VReg::Type::Int);
            AInstr *instr = makeMovBitRIMM(v, left, ShiftType::ASR, 31);
            curABlock->addLast(instr);

            VReg v1 = vRegAllocator.alloc(VReg::Type::Int);
            instr = new AInstrALU{AInstrALU::Type::Add,
                                  Cond::None,
                                  false,
                                  v1,
                                  left,
                                  Operand2Reg{v, ShiftType::LSR, 32 - l}};
            curABlock->addLast(instr);

            instr = makeMovBitRIMM(dest, v1, ShiftType::ASR, l);
            curABlock->addLast(instr);
        } else {
            using i64 = long long;
            i64 nc = (1ll << 31) - ((1ll << 31) % (i64) abs) - 1ll;
            i64 p = 32ll;
            while ((1ll << p) <= nc * ((i64) abs - (1ll << p) % (i64) abs)) {
                p++;
            }
            i64 m = (((1ll << p) + (i64) abs - (1ll << p) % (i64) abs) / (i64) abs);
            int n = (int) (m & 0xffffffffll);
            //            int n = (int) ((m << 32) >> 32);
            int shift = (int) (p - 32);

            auto v = vRegAllocator.alloc(VReg::Type::Int);
            AInstr *instr = new AInstrMov32(v, n);
            curABlock->addLast(instr);
            auto v2 = vRegAllocator.alloc(VReg::Type::Int);
            if (m >= 2147483648ll) {
                instr = makeSMMla(v2, left, v, left);
                curABlock->addLast(instr);
            } else {
                instr = new AInstrSMMUL{v2, left, v, Cond::None};
                curABlock->addLast(instr);
            }
            auto v3 = vRegAllocator.alloc(VReg::Type::Int);
            AInstr *mc3 = makeMovBitRIMM(v3, v2, ShiftType::ASR, shift);
            curABlock->addLast(mc3);

            AInstr *mc4 = new AInstrALU{AInstrALU::Type::Add,
                                        Cond::None,
                                        false,
                                        dest,
                                        v3,
                                        Operand2Reg{left, ShiftType::LSR, 31}};
            curABlock->addLast(mc4);
        }
        if (imm < 0) {
            auto *rsb = makeAInstrALURImm(AInstrALU::Type::Rsb, dest, dest, 0);
            curABlock->addLast(rsb);
        }
    }

    void ACodeGen::genOptimizedMod(const AReg &dest, const AReg &left, int imm) {
        genOptimizedDiv(dest, left, imm);
        auto immReg = vRegAllocator.alloc(VReg::Type::Int);
        curABlock->addLast(new AInstrMov32{immReg, imm});
        //        curABlock->addLast(makeMul(dest, dest, immReg));
        //        curABlock->addLast(makeAInstrALURR(AInstrALU::Type::Sub, dest, left,
        //        dest));
        curABlock->addLast(makeMls(dest, dest, immReg, left));
    }

    ACodeGen::ACodeGen(Ir::CompileUnit *_compileUnit,
                       const std::unordered_set<std::string> &optEnabled) {
        compileUnit = _compileUnit;
        for (auto &opt: optEnabled) {
            if (opt == "divopt") {
                divOpt = true;
            } else if (opt == "mergecmpbr") {
                mergeCmpBr = true;
            } else if (opt == "mergeblock") {
                mergeBlock = true;
            } else if (opt == "mergecond") {
                mergeCondExec = true;
            } else if (opt == "peephole") {
                peepHole = true;
            } else if (opt == "jtable") {
                jtable = true;
            }
        }
    }

    void ACodeGen::arrayIndexLastOffset() {
        for (auto &instr: loadStoreMemObj) {
            auto memObj = dynamic_cast<Ir::MemObject *>(
                    vRegAllocator.getHolding(std::get<VReg>(instr->rn)));
            auto indexIter = indexHints.find(memObj);
            if (indexIter == indexHints.end()) {
                panic("index into unknown memObj");
            }
            auto [_, index] = *indexIter;
            //            atriAssert(vRegAllocator.holdSame(index.base,instr->rn));
            if (std::holds_alternative<int>(index.index)) {
                int i = std::get<int>(index.index);
                if (instr->isFloat) {
                    if (abs(i) <= 1020) {
                        instr->rn = index.base;
                        instr->operand2 = i;
                    } else {
                        AReg t1 = vRegAllocator.alloc(VReg::Type::Int);
                        if (isImm8m(i)) {
                            auto addOffset =
                                    makeAInstrALURImm(AInstrALU::Type::Add, t1, index.base, i);
                            instr->insertBeforeMe(addOffset);
                        } else {
                            AReg t = vRegAllocator.alloc(VReg::Type::Int);
                            auto ldr = new AInstrMov32(t, i);
                            auto addOffset =
                                    makeAInstrALURR(AInstrALU::Type::Add, t1, index.base, t);
                            instr->insertBeforeMe(ldr);
                            instr->insertBeforeMe(addOffset);
                        }
                        instr->rn = t1;
                    }
                } else {
                    if (abs(i) <= 4095) {
                        instr->operand2 = i;
                    } else {
                        AReg t = vRegAllocator.alloc(VReg::Type::Int);
                        auto ldr = new AInstrMov32(t, i);
                        instr->insertBeforeMe(ldr);
                        instr->operand2 = RegLS{t, ShiftType::None, 0, false};
                    }
                }
            } else if (std::holds_alternative<AReg>(index.index)) {
                auto i = std::get<AReg>(index.index);
                if (instr->isFloat) {
                    AReg t1 = vRegAllocator.alloc(VReg::Type::Int);
                    auto addOffset = new AInstrALU{AInstrALU::Type::Add,
                                                   Cond::None,
                                                   false,
                                                   t1,
                                                   index.base,
                                                   Operand2Reg{i, ShiftType::LSL, 2}};
                    instr->insertBeforeMe(addOffset);
                    instr->rn = t1;
                } else {
                    instr->operand2 = RegLS{i, ShiftType::LSL, 2, false};
                }
            }
        }
    }

}// namespace armv7