//
// Created by 11067 on 2023/6/25.
//

#ifndef COMPILER_AARCH_H
#define COMPILER_AARCH_H

#include <string>
#include <variant>
#include "../../optimization/ir.h"
#include "../../utils/logger.h"
/**
 * 调用约定: 参数1~参数4 分别保存到 R0~R3 寄存器中 ，r0,s0是返回值。
 * s0-s15都可用于传参
 * 剩下的参数从右往左依次入栈，被调用者实现栈平衡，返回值存放在 R0 中。
 * https://learn.microsoft.com/en-us/cpp/build/overview-of-arm-abi-conventions?view=msvc-170
 * (中文翻译有点抽象，看英文)
 */
namespace armv7 {
    constexpr int WORD_SIZE = 4;
    enum class RegAttribute {
        CallerSaved,
        CalleeSaved,
        NA,
    };

    /**
     * @brief 寄存器
     * 约定0-15是基本寄存器，16-47是32bit浮点寄存器。
     */
    struct Reg {
        int id;
        const char *abiName;
        RegAttribute regAttribute;

        bool operator==(const Reg &rhs) const;

        bool operator!=(const Reg &rhs) const;

        std::string toAssemblyString() const;
    };

    struct SIMDReg {
        int id;
        const char *abiName;
    };
    struct VReg {
        int id{0};
        enum class Type {
            Float,
            Int,
        };
        Type type{Type::Float};

        bool nullReg() const;

        bool operator==(const VReg &rhs) const;

        bool operator!=(const VReg &rhs) const;

        std::string toAssemblyString() const;

        bool isPlaceHolder() const;

    };

//    /**
//     * @brief 占位寄存器
//     *
//     * 因为分析到某条指令时，其操作数可能尚未被翻译为后端寄存器，占位
//     */
//    struct PlaceHolderReg {
//        Ir::Reg reg;
//    };
    class VRegAllocator {
        int allocCnt = 1;
        int placeHolderCnt = -1;
        std::vector<Ir::Type *> placeHolderMap;

    public:
        std::vector<VReg::Type> vRegType{};

        int getAllocCnt() const {
            return allocCnt - 1;
        }

        VReg alloc(VReg::Type type) {
            vRegType.push_back(type);
            allocCnt++;
            return VReg{
                    allocCnt - 1,
                    type,
            };
        }

        void delloc() {
            vRegType.pop_back();
            allocCnt--;
        }

        /**
         * @brief 生成占位符（使用负数id的VReg）
         * @param irObj 对应IR的Type
         * @return 分配的虚拟寄存器
         */
        VReg allocPlaceHolder(Ir::Type *irObj) {
            placeHolderMap.push_back(irObj);
            placeHolderCnt--;
            return VReg{
                    placeHolderCnt + 1,
                    VReg::Type::Int, // 这里没有
            };
        }

        Ir::Type *getHolding(VReg &vReg) {
            return placeHolderMap[-vReg.id - 1];
        }

        bool holdSame(const VReg &vReg1, const VReg &vReg2) {
            return placeHolderMap[-vReg1.id - 1] == placeHolderMap[-vReg2.id - 1];
        }
        void clear() {
            allocCnt = 1;
            placeHolderCnt = -1;
            placeHolderMap.clear();
            vRegType.clear();
        }
    };

    inline VReg::Type IRType2VRegType(Ir::ScalaType type) {
        switch (type) {
            case Ir::Int:
            case Ir::Bool:
                return VReg::Type::Int;
            case Ir::Float:
                return VReg::Type::Float;
            case Ir::Void:
                unreachable();
        }
    }

    typedef std::variant<
            VReg,
            Reg
    > AReg;

    inline bool matchRegAlloc(const AReg &aReg, const VReg &vReg) {
        if (std::holds_alternative<Reg>(aReg)) {
            return false;
        }
        return std::get<VReg>(aReg) == vReg;
    }

    inline std::string aReg2String(const AReg &aReg) {
        if (std::holds_alternative<Reg>(aReg)) {
            return std::get<Reg>(aReg).abiName;
        } else {
            return std::get<VReg>(aReg).toAssemblyString();
        }
    }

    inline std::ostream &operator<<(std::ostream &out, const AReg &reg) {
        out << aReg2String(reg);
        return out;
    }

    inline VReg::Type aReg2Type(AReg &aReg) {
        if (auto vReg = std::get_if<VReg>(&aReg)) {
            return vReg->type;
        } else {
            auto &reg = std::get<Reg>(aReg);
            if (reg.id < 16) {
                return VReg::Type::Int;
            } else {
                return VReg::Type::Float;
            }
        }
    }

    inline bool aRegNull(const AReg &aReg) {
        if (auto vReg = std::get_if<VReg>(&aReg)) {
            return vReg->nullReg();
        }
        return false;
    }


    Reg &get_reg_by_id(int i);
    SIMDReg &get_simd_reg_by_id(int i);
    SIMDReg &to_alias(Reg freg);
    /**
     * @brief 按id获取浮点寄存器
     * @param i f32寄存器id
     * @return 寄存器单例
     */
    Reg &get_freg_by_id(int i);

    Reg &get_sp();

    Reg &get_lr();

    Reg &get_ip();

    Reg &get_pc();

    std::vector<Reg> get_saved_reg();

    std::vector<Reg> get_saved_freg();

    std::vector<AReg> get_saved_reg_wrapped();

    std::vector<AReg> get_saved_freg_wrapped();

    std::vector<Reg> get_temp_reg();

    constexpr int saved_reg_cnt = 8;
    constexpr int saved_freg_cnt = 16;

    /**
     * func reg unique id，前面固定为PReg id，vReg id加上PReg的数量
     * @param aReg VReg | Reg
     * @return uid
     */
    int fruid(const AReg &aReg);

    constexpr int getPRegCnt() {
        return 48;
    }

}

template<>
struct std::hash<armv7::VReg> {
    std::size_t operator()(armv7::VReg const &vReg) const noexcept {
        std::size_t h1 = std::hash<int>{}(vReg.id);
        std::size_t h2 = std::hash<int>{}(static_cast<int>(vReg.type));
        return h1 ^ (h2 << 1); // or use boost::hash_combine
    }
};

template<>
struct std::hash<armv7::Reg> {
    std::size_t operator()(armv7::Reg const &reg) const noexcept {
        std::size_t h1 = std::hash<int>{}(reg.id);
        return h1; // or use boost::hash_combine
    }
};

#endif //COMPILER_AARCH_H
