//
// Created by 11067 on 2023/6/28.
//

#ifndef COMPILER_AINSTRLS_H
#define COMPILER_AINSTRLS_H

#include <variant>
#include "../aarch.h"
#include "../AData.h"
#include "../AInstr.h"
#include "common.h"

namespace armv7 {
    /**
     * @brief 用于加载label高位[16:31]到寄存器的伪指令
     *
     * 汇编为movt #:upper:label
     * 但是gcc-9的行为一般是相对pc加载，不知道有没有性能区别
     * 之前参赛队都是用的这条伪指令
     */
    struct AInstrLaHigh : AInstr {
        AReg rd;
        const AData *label;
        AInstrLaHigh(const AReg &rd, const AData *label);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief 用于加载label低位到寄存器的伪指令
     */
    struct AInstrLaLow : AInstr {
        AReg rd;
        AData *label;

        AInstrLaLow(AReg &rd, AData *label);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    struct AInstrMovAddr32 : AInstr {
        AReg rd;
        AData *label;
        AInstrMovAddr32(const AReg &rd, AData *label);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief STMDB SP!, <reglist> 的规范格式
     *
     *  1. STM: 保存多个，D(decrease)B(before)是在执行前减少SP ，SP是保存位置，栈顶
     *  2. !:在传输完毕后更新基址寄存器(SP)
     *  3. 效果就是PUSH
     *  浮点形式完全一致，就是前面加V，可以复用
     */
    struct AInstrPush : AInstr {
        std::vector<AReg> reglist;
        bool isFloat;

        AInstrPush(const std::vector<AReg> &reglist, bool isFloat);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };


    /**
     * @brief LDM{IA} SP!, <reglist> 的规范格式
     *
     * IA(increase after)被省略，同上
     *
     * @see AInstrPush
     */
    struct AInstrPop : AInstr {
        std::vector<AReg> reglist;
        bool isFloat;

        AInstrPop(const std::vector<AReg> &reglist, bool isFloat);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief Rm <opsh> 带移位和符号的寄存器
     *
     * 用于Load Store
     *
     * @see AInstrLS
     */
    struct RegLS {
        AReg rm;
        ShiftType shiftType;
        int shift;

        bool operator==(const RegLS &rhs) const;

        bool operator!=(const RegLS &rhs) const;

        bool minus;
    };

    /**
     * @brief str和ldr指令，包含了其各种形式
     *
     * 注意第二个参数偏移的范围,浮点数也会复用这个类，但是浮点版本只能是
     * 立即数offset形式，且范围是[-1020,1020]
     */
    struct AInstrLS : AInstr {
        enum class Type {
            Load,
            Store
        };
        enum class SetType {
            None,
            PostIndex,
            PreIndex,
        };
        bool isFloat;
        Type type;
        SetType setType;
        AReg rd;
        AReg rd2;
        AReg rn;
        std::variant<int, RegLS> operand2; //offset [-4095,4095]

        AInstrLS(bool isFloat, Cond cond, Type type, const AReg &rd, const AReg &rn,
                 const std::variant<int, RegLS> &operand2, const AReg &rd2, SetType setType);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };
    bool isSameRegLS(std::variant<int, RegLS> addr1, std::variant<int, RegLS> addr2);
    bool isSameAddr(AInstrLS* ls1, AInstrLS* ls2);
    /**
     * @brief 寄存器相对寻址加载
     * @param rd 目标寄存器
     * @param rn 基址寄存器
     * @param offset 符合范围的立即数，对于GPR是[-4095,4095]，对于VFP是[-1020,1020]
     * @return 构造的指令
     */
    AInstrLS *makeLdRImm(const AReg &rd, const AReg &rn, int offset);

    AInstrLS *makeLdRImmF(const AReg &rd, const AReg &rn, int offset);

    /**
     * @brief 基址变址寻址加载
     *
     * 如果立即数编码不下
     *
     * @param rd 目标寄存器
     * @param rn 基址寄存器
     * @param rm 变址寄存器
     * @return 构造的指令
     */
    AInstrLS *makeLdRR(const AReg &rd, const AReg &rn, const AReg &rm);

    AInstrLS *makeLdRRF(const AReg &rd, const AReg &rn, const AReg &rm);

    AInstrLS *makeStrRImm(const AReg &rd, const AReg &rn, int offset);

    AInstrLS *makeStrRImmF(const AReg &rd, const AReg &rn, int offset);

    AInstrLS *makeStrRR(const AReg &rd, const AReg &rn, const AReg &rm);

    AInstrLS *makeStrRRF(const AReg &rd, const AReg &rn, const AReg &rm);

    AInstrPush *makePush(const std::vector<AReg> &regList);

    AInstrPush *makePushF(std::vector<AReg> &regList);

    AInstrPop *makePop(const std::vector<AReg> &regList);

    AInstrPop *makePopF(std::vector<AReg> &regList);

} // armv7

#endif //COMPILER_AINSTRLS_H
