//
// Created by 11067 on 2023/6/26.
//

#include "AInstrMov.h"
#include "../../../utils/bitUtils.h"

using namespace std::literals;
namespace armv7 {
    std::vector<AReg> AInstrMov::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMov::getUse() {
        std::vector<AReg> ret{};
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto operand2Reg = std::get<Operand2Reg>(operand2);
            ret.push_back(operand2Reg.base);
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                ret.push_back(std::get<AReg>(operand2Reg.shift));
            }
        }
        return ret;
    }

    void AInstrMov::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto &operand2Reg = std::get<Operand2Reg>(operand2);
            if (matchRegAlloc(operand2Reg.base, vReg)) {
                operand2Reg.base = pReg;
            }
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                auto &shift = std::get<AReg>(operand2Reg.shift);
                if (matchRegAlloc(shift, vReg)) {
                    shift = pReg;
                }
            }
        }
    }

    std::string AInstrMov::toAssemblyString() {
        std::string ret{};
        if (neg) {
            ret += "mvn";
        } else {
            ret += "mov";
        }
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        ret += ' ';
        ret += aReg2String(rd);
        ret += ", ";
        ret += operand2ToString(operand2);
        return ret;
    }

    AInstrMov::AInstrMov(const AReg &rd, const Operand2 &operand2, bool setFlag, bool neg, Cond cond) : AInstr{cond},
                                                                                                        rd(rd),
                                                                                                        operand2(
                                                                                                                operand2),
                                                                                                        setFlag(setFlag),
                                                                                                        neg(neg) {}

    bool AInstrMov::canMerge() {
        auto reg = std::get_if<Operand2Reg>(&operand2);
        return cond == Cond::None && !setFlag && !neg && reg != nullptr && reg->shiftType == ShiftType::None;
    }

    std::vector<AReg> AInstrMovLi::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMovLi::getUse() {
        return {};
    }

    void AInstrMovLi::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrMovLi::toAssemblyString() {
        auto ret = "movw"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", " + std::to_string(imm16);
    }

    AInstrMovLi::AInstrMovLi(const AReg &rd, int imm16) : rd(rd), imm16(imm16) {}

    std::vector<AReg> AInstrMovt::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMovt::getUse() {
        return {};
    }

    void AInstrMovt::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrMovt::toAssemblyString() {
        auto ret = "movt"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", " + std::to_string(imm16);
    }

    AInstrMovt::AInstrMovt(const AReg &rd, int imm16) : rd(rd), imm16(imm16) {}

    AInstrMov *makeMovR2R(const AReg &rd, const AReg &rs) {
        return new AInstrMov{
                rd,
                Operand2Reg{
                        rs,
                        ShiftType::None,
                        0
                },
                false,
                false,
                Cond::None,
        };
    }

    AInstrPseLdr *makeLoadF(AReg rd, float constant) {
        return new AInstrPseLdr{rd, constant, true};
    }

    AInstrMov *makeShlRR(AReg &rd, AReg &base, AReg &shift) {
        return new AInstrMov{
                rd,
                Operand2Reg{
                        base,
                        ShiftType::LSL,
                        shift
                },
                false,
                false,
                Cond::None
        };
    }

    AInstrMov *makeShrRR(AReg &rd, AReg &base, AReg &shift) {
        return new AInstrMov{
                rd,
                Operand2Reg{
                        base,
                        ShiftType::LSR,
                        shift
                },
                false,
                false,
                Cond::None
        };
    }

    AInstrMov *makeAsrRR(AReg &rd, AReg &base, AReg &shift) {
        return new AInstrMov{
                rd,
                Operand2Reg{
                        base,
                        ShiftType::ASR,
                        shift
                },
                false,
                false,
                Cond::None
        };
    }

    std::pair<AInstrMov *, AInstrMov *> makeMovCond(Cond cond, AReg &rd) {
        return {
                new AInstrMov{
                        rd,
                        0,
                        false,
                        false,
                        Cond::None,
                },
                new AInstrMov{
                        rd,
                        1,
                        false,
                        false,
                        cond,
                },
        };
    }

    AInstrMov *makeMovBitRIMM(const AReg &rd, const AReg &rs, ShiftType shiftType, int imm) {
        return new AInstrMov{
                rd,
                Operand2Reg{
                        rs,
                        shiftType,
                        imm
                },
                false,
                false,
                Cond::None
        };
    }

    std::vector<AReg> AInstrPseLdr::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrPseLdr::getUse() {
        return {};
    }

    void AInstrPseLdr::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrPseLdr::toAssemblyString() {
        std::string ret{};
        if (isFloat) {
//            ret += "vldr.f32 ";
            ret += "vldr";
        } else {
            ret += "ldr";
        }
        ret += cond2StringView(cond);
        ret += ' ';
        ret += aReg2String(rd);

        if (auto imm = std::get_if<int>(&constant)) {
            ret += ", =";
            ret += std::to_string(toUI32(*imm));
        } else {
            ret += ", =";
            auto immf = std::get<float>(constant);
            ret += std::to_string(toUI32(immf));
        }
        return ret;
    }

    AInstrPseLdr::AInstrPseLdr(const AReg &rd, const std::variant<int, float> &constant, bool isFloat) : rd(rd),
                                                                                                         constant(
                                                                                                                 constant),
                                                                                                         isFloat(isFloat) {}

    unsigned int AInstrPseLdr::getConst() {
        if (isFloat) {
            return toUI32(std::get<float>(constant));
        } else {
            return toUI32(std::get<int>(constant));
        }
    }

    AInstrMov32::AInstrMov32(const AReg &rd, int imm32) : rd(rd), imm32(imm32) {}

    std::vector<AReg> AInstrMov32::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMov32::getUse() {
        return {};
    }

    void AInstrMov32::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrMov32::toAssemblyString() {
        auto ret = "mov32"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", " + std::to_string(imm32);
    }
} // armv7