//
// Created by 11067 on 2023/6/26.
//

#ifndef COMPILER_AINSTRMOV_H
#define COMPILER_AINSTRMOV_H

#include "../AInstr.h"
#include "common.h"

namespace armv7 {
    /**
     * @brief mov,通过operand2提供了移位的功能
     * neg实际使用mvn指令
     */
    struct AInstrMov : public AInstr {
        AReg rd;
        Operand2 operand2;
        bool setFlag;
        bool neg; //变为使用mvn指令

        AInstrMov(const AReg &rd, const Operand2 &operand2, bool setFlag, bool neg, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        bool canMerge();
    };

    AInstrMov *makeShlRR(AReg &rd, AReg &base, AReg &shift);

    AInstrMov *makeShrRR(AReg &rd, AReg &base, AReg &shift);

    AInstrMov *makeAsrRR(AReg &rd, AReg &base, AReg &shift);

    /**
     * @brief 生成movgt,movlt序列
     * @param rd 目标寄存器
     * @return 指令序列
     */
    std::pair<AInstrMov *, AInstrMov *> makeMovCond(Cond cond, AReg &rd);

    /**
     * @brief 构造mov rd,rs
     * @param rd 目标寄存器
     * @param rs 源寄存器
     * @return 指令
     */
    AInstrMov *makeMovR2R(const AReg &rd, const AReg &rs);
    /**
     * 产生和位运算指令等价的mov指令
     * @param rd
     * @param rs
     * @param shiftType
     * @param imm
     * @return
     */
    AInstrMov *makeMovBitRIMM(const AReg &rd, const AReg &rs, ShiftType shiftType, int imm);

    /**
     * @brief 加载立即数到低16位
     * 注意顺序，会把高位清零
     */
    struct AInstrMovLi : public AInstr {
        AReg rd;
        int imm16;

        AInstrMovLi(const AReg &rd, int imm16);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief 加载立即数到高16位
     */
    struct AInstrMovt : public AInstr {
        AReg rd{};
        int imm16{};

        AInstrMovt(const AReg &rd, int imm16);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    struct AInstrMov32 : public AInstr {
        AReg rd{};
        int imm32{};

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        AInstrMov32(const AReg &rd, int imm32);
    };

    /**
     * @brief 从常量池加载立即数的伪指令
     *
     * 对应ldr，vldr伪指令
     *
     * @note 使用.ltorg指令保证在范围内存在可用常量池
     */
    struct AInstrPseLdr : public AInstr {
        AReg rd;
        std::variant<int, float> constant;
        bool isFloat;

        AInstrPseLdr(const AReg &rd, const std::variant<int, float> &constant, bool isFloat);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        unsigned int getConst();
    };

    AInstrPseLdr *makeLoadF(AReg rd, float constant);
} // armv7

#endif //COMPILER_AINSTRMOV_H
