//
// Created by 11067 on 2023/6/26.
//

#include "AInstrMul.h"

namespace armv7 {

    std::vector<AReg> AInstrMul::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMul::getUse() {
        return {rm, rs};
    }

    void AInstrMul::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (matchRegAlloc(rs, vReg)) {
            rs = pReg;
        }
    }

    std::string AInstrMul::toAssemblyString() {
        std::string op{"mul"};
        if (setFlag) {
            op += 's';
        }
        if (cond != Cond::None) {
            op += cond2StringView(cond);
        }
        op += ' ';
        op += aReg2String(rd) + ", ";
        op += aReg2String(rm) + ", ";
        op += aReg2String(rs);
        return op;
    }

    AInstrMul::AInstrMul(const AReg &rd, const AReg &rm, const AReg &rs, bool setFlag, Cond cond) : AInstr{cond},
                                                                                                    rd(rd), rm(rm),
                                                                                                    rs(rs),
                                                                                                    setFlag(setFlag) {}

    std::vector<AReg> AInstrMla::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMla::getUse() {
        return {rm, rn, rs};
    }

    void AInstrMla::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (matchRegAlloc(rn, vReg)) {
            rn = pReg;
        }
        if (matchRegAlloc(rs, vReg)) {
            rs = pReg;
        }
    }

    std::string AInstrMla::toAssemblyString() {
        std::string ret{signFlag ? "sm" : ""};
        ret += addFlag ? "mla" : "mls";
        ret += cond2StringView(cond);
        ret += setFlag ? "s" : "";
        ret += " ";
        ret += aReg2String(rd);
        ret += ", ";
        ret += aReg2String(rm);
        ret += ", ";
        ret += aReg2String(rs);
        ret += ", ";
        ret += aReg2String(rn);
        return ret;
    }

    AInstrMla::AInstrMla(const AReg &rd, const AReg &rm, const AReg &rs, const AReg &rn, bool setFlag, bool addFlag,
                         bool signFlag, Cond cond) : AInstr{cond}, rd(rd), rm(rm), rs(rs), rn(rn), setFlag(setFlag),
                                                     addFlag(addFlag),
                                                     signFlag(signFlag) {}

    std::vector<AReg> AInstrSDiv::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrSDiv::getUse() {
        return {rm, rn};
    }

    void AInstrSDiv::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (matchRegAlloc(rn, vReg)) {
            rn = pReg;
        }
    }

    std::string AInstrSDiv::toAssemblyString() {
        std::string op{"sdiv"};
        if (setFlag) {
            op += 's';
        }
        if (cond != Cond::None) {
            op += cond2StringView(cond);
        }
        op += ' ';
        op += aReg2String(rd) + ", ";
        op += aReg2String(rm) + ", ";
        op += aReg2String(rn);
        return op;
    }

    AInstrSDiv::AInstrSDiv(const AReg &rd, const AReg &rn, const AReg &rm, bool setFlag, Cond cond) : AInstr{cond},
                                                                                                      rd(rd), rm(rm),
                                                                                                      rn(rn),
                                                                                                      setFlag(setFlag) {}

    AInstrSDiv *makeSDiv(AReg rd, AReg rm, AReg rn) {
        return new AInstrSDiv{
                rd,
                rn,
                rm,
                false,
                Cond::None
        };
    }

    AInstrMul *makeMul(AReg rd, AReg rm, AReg rn) {
        return new AInstrMul{
                rd,
                rm,
                rn,
                false,
                Cond::None
        };
    }

    std::tuple<AInstrSDiv *, AInstrMul *, AInstrALU *> makeRem(AReg rd, AReg rm, AReg rn) {
        return {
                makeSDiv(rd, rm, rn),
                new AInstrMul{
                        rd,
                        rd,
                        rn,
                        false,
                        Cond::None
                },
                makeAInstrALURR(
                        AInstrALU::Type::Sub,
                        rd,
                        rm,
                        rd
                )
        };
    }

    AInstrMla *makeSMMla(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc) {
        return new AInstrMla{
                rd,
                r1,
                r2,
                acc,
                false,
                true,
                true,
                Cond::None
        };
    }

    AInstrMla *makeMls(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc) {
        return new AInstrMla{
                rd,
                r1,
                r2,
                acc,
                false,
                false,
                false,
                Cond::None
        };
    }

    AInstrMla *makeMla(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc) {
        return new AInstrMla{
                rd,
                r1,
                r2,
                acc,
                false,
                true,
                false,
                Cond::None
        };
    }

    std::vector<AReg> AInstrSMMUL::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrSMMUL::getUse() {
        return {rm, rs};
    }

    void AInstrSMMUL::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (matchRegAlloc(rs, vReg)) {
            rs = pReg;
        }
    }

    std::string AInstrSMMUL::toAssemblyString() {
        std::string ret{"smmul"};
        ret += cond2StringView(cond);
        ret += ' ';
        ret += aReg2String(rd);
        ret += ", ";
        ret += aReg2String(rm);
        ret += ", ";
        ret += aReg2String(rs);
        return ret;
    }

    AInstrSMMUL::AInstrSMMUL(const AReg &rd, const AReg &rm, const AReg &rs, Cond cond) : AInstr{cond}, rd(rd), rm(rm),
                                                                                          rs(rs) {}
}