//
// Created by 11067 on 2023/7/10.
//

#include "ColorAllocator.h"
#include "LiveInfo.h"
#include "../instructions/AInstrMov.h"
#include "../instructions/AInstrVFP.h"
#include "../../../utils/logger.h"
#include "../instructions/AInstrALU.h"

#include "../instructions/AInstrB.h"
#include <algorithm>
#include <cmath>

namespace armv7 {
    void ColorAllocator::run(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {

            if (!func->isLib) {
                info("color allocating:", func->label);
                auto allocator = ColorAllocatorExecutor{func};
                allocator.run();
            }
        }
    }

    void ColorAllocatorExecutor::run() {
        while (true) {
            dbgvar(initial.size());
            dbgvar(function->vRegAllocated);
            dbgvar(spiltedReg.size());
//            for (int i = 1; i < function->vRegAllocated + 1; ++i) {
//                if (std::none_of(initial.begin(), initial.end(),
//                                 [&i](auto &t) {return fruid(t) == i + 47;})) {
//                    info("here!!!");
//                }
//            }
//            atriAssert(initial.size() == function->vRegAllocated - spiltedReg.size());

            auto liveInfo = LiveInfo();
            liveInfo.runOnFunction(function);
            whenDebug function->toAssemblyString(std::cerr);
            info("color allocator running.");
            calculateSpillPriority();
            clear();
            build();
            makeWorkList();
            do {
                if (!simplifyWorkList.empty()) {
                    simplify();
                } else if (!workListMoves.empty()) {
                    coalesce();
                } else if (!freezeWorkList.empty()) {
                    freeze();
                } else if (!spillWorkList.empty()) {
                    selectSpill();
                }
            } while (!(simplifyWorkList.empty() && workListMoves.empty()
                       && freezeWorkList.empty() && spillWorkList.empty()));
            assignColors();
            if (!spilledNodes.empty()) {
                rewriteProgram();
            } else {
                break;
            }
        }
        finish();
    }

    ColorAllocatorExecutor::ColorAllocatorExecutor(AFunction *function) : function(function) {
        for (auto bb: *function) {
            for (auto instr: *bb) {
                for (auto reg: instr->getDef()) {
                    if (std::holds_alternative<VReg>(reg)) {
                        initial.insert(reg);
                    }

                }
                for (auto reg: instr->getUse()) {
                    if (std::holds_alternative<VReg>(reg)) {
                        initial.insert(reg);
                    }
                }
            }
        }
    }

    void ColorAllocatorExecutor::calculateSpillPriority() {
        spillPriority.clear();
        spillPriority.resize(getAllRegCnt(), 0.);
        for (auto bb: *function) {
            for (auto instr: *bb) {
                for (auto reg: instr->getDef()) {
                    spillPriority[fruid(reg)] += pow(SpillParam, bb->loopDepth);
                }
                for (auto reg: instr->getUse()) {
                    spillPriority[fruid(reg)] += pow(SpillParam, bb->loopDepth);
                }
            }
        }
    }

    void ColorAllocatorExecutor::clear() {
        simplifyWorkList.clear();
        freezeWorkList.clear();
        spillWorkList.clear();
        spilledNodes.clear();
        coalescedNodes.clear();
        coloredNodes.clear();
        selectStack.clear();
        selectSet.clear();
        coalescesMoves.clear();
        constrainedMoves.clear();
        frozenMoves.clear();
        workListMoves.clear();
        workListMoves.clear();
        activeMoves.clear();
        adjList.clear();
        adjList.resize(getAllRegCnt());
        adjSet.clear();
        degree.clear();
        degree.resize(getAllRegCnt());
        alias.clear();
        alias.resize(getAllRegCnt());
        moveList.clear();
        moveList.resize(getAllRegCnt());
        colors.clear();
        colors.resize(getAllRegCnt());
        for (int i = 0; i < getPRegCnt(); ++i) {
            colors[i] = get_reg_by_id(i);
            degree[i] = INT32_MAX;
            alias[i] = VReg{};
            moveList[i].clear();
        }
        for (int i = getPRegCnt(); i < getAllRegCnt(); ++i) {
            colors[i] = VReg{};
            degree[i] = 0;
            adjList[i].clear();
            alias[i] = VReg{};
            moveList[i].clear();
        }
    }

    void ColorAllocatorExecutor::build() {

        for (int i = 0; i < function->vRegType.size(); ++i) {
            if (spiltedReg.count(i + 1)) {
                continue;
            }
            // float vReg 和所有通用寄存器冲突
            if (function->vRegType[i] == VReg::Type::Float) {
                for (int j = 0; j < 16; ++j) {
                    adjList[i + getPRegCnt()].insert(get_reg_by_id(j));
                    degree[i + getPRegCnt()]++;
                }
            }
            // int vReg 和所有VFP寄存器冲突
            if (function->vRegType[i] == VReg::Type::Int) {
                for (int j = 16; j < 48; ++j) {
                    adjList[i + getPRegCnt()].insert(get_reg_by_id(j));
                    degree[i + getPRegCnt()]++;
                }
            }
            // sp,lr和pc与所有虚拟寄存器冲突
            adjList[i + getPRegCnt()].insert(get_sp());
            adjList[i + getPRegCnt()].insert(get_pc());
            adjList[i + getPRegCnt()].insert(get_lr());
            degree[i + getPRegCnt()] += 3;
        }
        for (auto bb: *function) {
            unordered_set<AReg> live{bb->liveOut.begin(), bb->liveOut.end()};
            auto instrList = bb->toVector();
            for (int i = instrList.size() - 1; i >= 0; --i) {
                auto &instr = instrList[i];

                //两种move可以合并，VFP之间的和GPR之间的，跨越的不能合并
                if (auto move = dynamic_cast<AInstrMov *>(instr)) {
                    if (move->canMerge()) {
                        for (auto elem: instr->getUse()) {
                            live.erase(elem);
                        }
                        for (auto reg: instr->getUse()) {
                            moveList[fruid(reg)]
                                    .insert({std::get<Operand2Reg>(move->operand2).base, move->rd});
                        }
                        for (auto reg: instr->getDef()) {
                            moveList[fruid(reg)]
                                    .insert({std::get<Operand2Reg>(move->operand2).base, move->rd});
                        }
                        workListMoves.insert({std::get<Operand2Reg>(move->operand2).base, move->rd});
                    }
                }
                if (auto move = dynamic_cast<AInstrVMov *>(instr)) {
                    if (move->canMerge()) {
                        for (auto elem: instr->getUse()) {
                            live.erase(elem);
                        }
                        for (auto reg: instr->getUse()) {
                            moveList[fruid(reg)]
                                    .insert({move->rm, move->rd});
                        }
                        for (auto reg: instr->getDef()) {
                            moveList[fruid(reg)]
                                    .insert({move->rm, move->rd});
                        }
                        workListMoves.insert({move->rm, move->rd});
                    }
                }
                for (auto elem: instr->getDef()) {
                    live.insert(elem);
                }
                for (auto reg: instr->getDef()) {
                    for (auto reg1: live) {
                        addEdge(reg, reg1);
                        // 在这里维护adjSet？但是我的课设删了这一部分
                    }
                }
                //所有跨过程调用活跃的变量和临时寄存器冲突
                if (auto call = dynamic_cast<AInstrCall *>(instr)) {
                    for (auto &reg: live) {
                        if (std::holds_alternative<VReg>(reg)) {
                            for (auto &reg1: get_temp_reg()) {
                                addEdge(reg, reg1);
                            }
                        }
                    }
                }
                for (auto reg: instr->getDef()) {
                    live.erase(reg);
                }
                for (auto reg: instr->getUse()) {
                    live.insert(reg);
                }

            }
        }
    }

    void ColorAllocatorExecutor::makeWorkList() {
        for (auto reg: initial) {
            if (degree[fruid(reg)] >= getPRegCnt()) {
                spillWorkList.insert(reg);
            } else if (moveRelated(reg)) {
                freezeWorkList.insert(reg);
            } else {
                simplifyWorkList.push_back(reg);
            }
        }
    }

    int ColorAllocatorExecutor::getAllRegCnt() {
        return getPRegCnt() + function->vRegAllocated;
    }

    void ColorAllocatorExecutor::simplify() {
        AReg &aReg = simplifyWorkList.back();
        atriAssert(std::holds_alternative<VReg>(aReg))
//        dbgvar(aReg2String(aReg));
//        dbgvar(degree[fruid(aReg)]);
        //atriAssert(degree[fruid(aReg)] < getPRegCnt());
        simplifyWorkList.pop_back();
        selectStack.push_back(aReg);
        selectSet.insert(aReg);
        for (auto reg: adjList[fruid(aReg)]) {
            if (selectSet.count(reg) == 0 && coalescedNodes.count(reg) == 0) {
                decrementDegree(reg);
            }
        }
    }

    void ColorAllocatorExecutor::coalesce() {
        auto m = *workListMoves.begin();
        auto x = getAlias(m.second); //to
        auto y = getAlias(m.first); //from
        AReg u, v;
        if (std::holds_alternative<Reg>(y)) {
            u = y;
            v = x;
        } else {
            u = x;
            v = y;
        }
        workListMoves.erase(m);
        auto adjU = adjacent(u);
        auto adjV = adjacent(v);
        auto unionAdjUV = unordered_set{adjU};
        unionAdjUV.insert(adjV.begin(), adjV.end());
        if (u == v) {
            coalescesMoves.insert(m);
            addWorkList(u);
        } else if (std::holds_alternative<Reg>(v) || adjList[fruid(u)].count(v) || adjList[fruid(v)].count(u)) {
            constrainedMoves.insert(m);
            addWorkList(u);
            addWorkList(v);
        } else if (
                (std::holds_alternative<Reg>(u) && std::all_of(adjV.begin(), adjV.end(),
                                                               [&u, this](auto t) { return ok(t, u); })) ||
                (std::holds_alternative<VReg>(u) && conservative(unionAdjUV))
                ) {
            coalescesMoves.insert(m);
            combine(u, v);
            addWorkList(u);
        } else {
            activeMoves.insert(m);
        }
    }

    bool ColorAllocatorExecutor::ok(const AReg &t, const AReg &r) {
        return degree[fruid(t)] < getPRegCnt() || std::holds_alternative<Reg>(t)
               || adjList[fruid(t)].count(r) || adjList[fruid(r)].count(t);
    }

    bool ColorAllocatorExecutor::conservative(const unordered_set<AReg> &nodes) {
        int k = 0;
        for (auto &n: nodes) {
            if (degree[fruid(n)] >= getPRegCnt()) {
                k++;
            }
        }
        return k < getPRegCnt();
    }

    void ColorAllocatorExecutor::combine(const AReg &u, const AReg &v) {
        if (freezeWorkList.count(v)) {
            freezeWorkList.erase(v);
        } else {
            spillWorkList.erase(v);
        }
        coalescedNodes.insert(v);
        alias[fruid(v)] = u;
        moveList[fruid(u)].insert(moveList[fruid(v)].begin(), moveList[fruid(v)].end());
        unordered_set<AReg> t1{};
        t1.insert(v);
        enableMoves(t1);
        for (auto t: adjacent(v)) {
            addEdge(t, u);
            decrementDegree(t);
        }
        if (degree[fruid(u)] >= getPRegCnt() && freezeWorkList.count(u)) {
            freezeWorkList.erase(u);
            spillWorkList.insert(u);
        }

    }


    void ColorAllocatorExecutor::freeze() {
        auto u = *freezeWorkList.begin();
        freezeWorkList.erase(u);
        simplifyWorkList.push_back(u);
        freezeMoves(u);
    }

    void ColorAllocatorExecutor::selectSpill() {
        vector<AReg> spillList{spillWorkList.begin(), spillWorkList.end()};
        auto end = std::remove_if(spillList.begin(), spillList.end(), [this](const auto &elem) {
            return spillCreated.count(elem) > 0;
        });
        spillList.erase(end, spillList.cend());
        std::sort(spillList.begin(), spillList.end(), [&](const auto &l, const auto &r) {
            return spillPriority[fruid(l)] / degree[fruid(l)] < spillPriority[fruid(r)] / degree[fruid(r)];
        });
        auto regToSpill = *spillList.begin();
        atriAssert(spillCreated.count(regToSpill) == 0);
        spillWorkList.erase(regToSpill);
        simplifyWorkList.push_back(regToSpill);
        freezeMoves(regToSpill);
    }

    void ColorAllocatorExecutor::assignColors() {
        while (!selectStack.empty()) {
            auto reg = selectStack.back();
            selectStack.pop_back();
            unordered_set<AReg> okColors{};
            for (int i = 0; i < getPRegCnt(); ++i) {
                okColors.insert(get_reg_by_id(i));
            }
            for (auto &reg1: adjList[fruid(reg)]) {
                auto regAlias = getAlias(reg1);
                if (coloredNodes.count(regAlias) || std::holds_alternative<Reg>(regAlias)) {
                    okColors.erase(colors[fruid(regAlias)]);
                }
            }
            if (okColors.empty()) {
                spilledNodes.insert(reg);
            } else {
                coloredNodes.insert(reg);
                auto c = *okColors.begin();
                colors[fruid(reg)] = c;
            }


        }
        for (auto &reg: coalescedNodes) {
            colors[fruid(reg)] = colors[fruid(getAlias(reg))];
        }
    }

    void ColorAllocatorExecutor::rewriteProgram() {
        unordered_set<AReg> newTemp{};
        for (auto &r: spilledNodes) {

            atriAssert(std::holds_alternative<VReg>(r))
            auto vReg = std::get<VReg>(r);
            spiltedReg.insert(vReg.id);
            auto slotOffset = spilledVarCnt * WORD_SIZE;
            spilledVarCnt++;
            for (auto bb: *function) {
                for (auto instr: bb->toVector()) {
                    auto use = instr->getUse();
                    auto def = instr->getDef();
                    auto usedHere = std::count(use.begin(), use.end(), r);
                    auto defHere = std::count(def.begin(), def.end(), r);
                    if (usedHere || defHere) {
                        AReg tReg = allocSplitVReg(vReg.type);
                        newTemp.insert(tReg);
                        spillCreated.insert(tReg);
                        if (usedHere) {
                            if (vReg.type == VReg::Type::Int) {
                                if (slotOffset > 4095) {
                                    auto movImm = new AInstrMov32{get_reg_by_id(10), slotOffset};
                                    auto ldr = makeLdRR(tReg, get_sp(), get_reg_by_id(10));
                                    instr->insertBeforeMe(movImm);
                                    instr->insertBeforeMe(ldr);
                                } else {
                                    auto ldr = makeLdRImm(tReg, get_sp(), slotOffset);
                                    instr->insertBeforeMe(ldr);
                                }
                            } else {
                                if (slotOffset > 1020) {
                                    auto movImm = new AInstrMov32{get_reg_by_id(10), slotOffset};
                                    auto ldr = makeLdRRF(tReg, get_sp(), get_reg_by_id(10));
                                    instr->insertBeforeMe(movImm);
                                    instr->insertBeforeMe(ldr);
                                } else {
                                    auto ldr = makeLdRImmF(tReg, get_sp(), slotOffset);
                                    instr->insertBeforeMe(ldr);
                                }

                            }
                        }
                        if (defHere) {
                            if (vReg.type == VReg::Type::Int) {
                                if (slotOffset > 4095) {
                                    auto movImm = new AInstrMov32{get_reg_by_id(10), slotOffset};
                                    auto str = makeStrRR(tReg, get_sp(), get_reg_by_id(10));
                                    instr->insertAfterMe(str);
                                    instr->insertAfterMe(movImm);
                                } else {
                                    auto str = makeStrRImm(tReg, get_sp(), slotOffset);
                                    instr->insertAfterMe(str);
                                }
                            } else {
                                if (slotOffset > 1020) {
                                    auto movImm = new AInstrMov32{get_reg_by_id(10), slotOffset};
                                    auto str = makeStrRRF(tReg, get_sp(), get_reg_by_id(10));
                                    instr->insertAfterMe(str);
                                    instr->insertAfterMe(movImm);
                                } else {
                                    auto str = makeStrRImmF(tReg, get_sp(), slotOffset);
                                    instr->insertAfterMe(str);
                                }
                            }
                        }
                        instr->allocate(vReg, tReg);
                    }

                }
            }
        }
        spilledNodes.clear();
        initial.clear();
        initial.insert(coloredNodes.begin(), coloredNodes.end());
        initial.insert(coalescedNodes.begin(), coalescedNodes.end());
        initial.insert(newTemp.begin(), newTemp.end());

        coloredNodes.clear();
        coalescedNodes.clear();
    }

    void ColorAllocatorExecutor::finish() {
        whenDebug for (int i = 0; i < function->vRegAllocated - 1; ++i) {
//                atriAssert(spiltedReg.count(i+1) > 0 || colors[48+i] != AReg{VReg{}})
                if (!(spiltedReg.count(i + 1) > 0 || colors[48 + i] != AReg{VReg{}})) {
                    error("vreg ", i + 1, " not allocated");
                }
            }
        auto stackRequired = spilledVarCnt + function->stackArrayAllocated;
        if ((stackRequired + 1) & 1) { // + LR
            stackRequired += 1;
        }
        auto localOffset = stackRequired * 4;
        function->reserveLocal->offset = -localOffset;
        function->cleanLocal->offset = localOffset;
        if (localOffset > 1020) {
            warn("too many local, offset:", localOffset);
        }
        //替换伪指令，ip被用作固定临时寄存器
        for (auto instr: function->loadLocalArrayInstructions) {
            instr->offset += spilledVarCnt * 4;
            if (isImm8m(instr->offset)) {
                instr->insertBeforeMe(new AInstrALU{
                        AInstrALU::Type::Add,
                        Cond::None,
                        false,
                        instr->rd,
                        instr->rm, // should be sp
                        instr->offset
                });

            } else {
                instr->insertBeforeMe(new AInstrMov32{get_ip(), instr->offset});
                instr->insertBeforeMe(makeAInstrALURR(AInstrALU::Type::Add, instr->rd, instr->rm, get_ip()));
            }
            instr->removeMe();
            delete instr;
        }
        int ipOffset = 0;
        for (auto instr: function->loadParamInstructions) {
            auto imm = std::get<int>(instr->operand2);
            if ((instr->isFloat && imm - ipOffset > 1020) || imm - ipOffset > 4095) {
                ipOffset += 1020;
                instr->insertBeforeMe(makeAInstrALURImm(
                        AInstrALU::Type::Add,
                        AReg{get_ip()},
                        AReg{get_ip()},
                        1020));
            }
            instr->operand2 = imm - ipOffset;
        }
        for (auto bb: *function) {
            for (auto instr: bb->toVector()) {
                for (auto reg: instr->getDef()) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
                        instr->allocate(*vReg, colors[fruid(*vReg)]);
                    }
                }
                for (auto reg: instr->getUse()) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
                        instr->allocate(*vReg, colors[fruid(*vReg)]);
                    }
                }
                if (auto move = dynamic_cast<AInstrMov *>(instr)) {
                    if (move->canMerge()) {
                        if (move->rd == std::get<Operand2Reg>(move->operand2).base) {
                            move->removeMe();
                        }
                    }
                }
                if (auto move = dynamic_cast<AInstrVMov *>(instr)) {
                    if (move->canMerge()) {
                        if (move->rd == move->rm) {
                            move->removeMe();
                        }
                    }
                }
            }
        }
    }

    bool ColorAllocatorExecutor::moveRelated(const AReg &reg) {
        return !nodeMoves(reg).empty();
    }

    unordered_set<Move> ColorAllocatorExecutor::nodeMoves(const AReg &reg) {
        auto ret = unordered_set<Move>{};
        for (auto move: moveList[fruid(reg)]) {
            if (activeMoves.count(move) > 0 || workListMoves.count(move) > 0) {
                ret.insert(move);
            }
        }
        return ret;
    }

    void ColorAllocatorExecutor::decrementDegree(AReg &reg) {
        auto d = degree[fruid(reg)];
        degree[fruid(reg)] = d - 1;
        if (d == getPRegCnt()) {
//            auto t = adjacent(reg);
//            t.insert(reg);
//            enableMoves(t);
            fusionEnableMove(reg);
            spillWorkList.erase(reg);
            if (moveRelated(reg)) {
                freezeWorkList.insert(reg);
            } else {
                simplifyWorkList.push_back(reg);
            }
        }
    }

    unordered_set<AReg> ColorAllocatorExecutor::adjacent(const AReg &reg) {
        auto ret = unordered_set<AReg>{adjList[fruid(reg)]};
        for (auto &e: selectSet) {
            ret.erase(e);
        }
        for (auto &e: coalescedNodes) {
            ret.erase(e);
        }
        return ret;
    }

    void ColorAllocatorExecutor::enableMoves(unordered_set<AReg> &nodes) {
        for (auto &reg: nodes) {
            for (auto &move1: nodeMoves(reg)) {
                if (activeMoves.count(move1)) {
                    activeMoves.erase(move1);
                    workListMoves.insert(move1);
                }
            }
        }
    }

    AReg ColorAllocatorExecutor::getAlias(const AReg &reg) {
        if (coalescedNodes.count(reg)) {
            return getAlias(alias[fruid(reg)]);
        } else {
            return reg;
        }
    }

    void ColorAllocatorExecutor::addWorkList(const AReg &u) {
        if (std::holds_alternative<VReg>(u) && !moveRelated(u) && degree[fruid(u)] < getPRegCnt()) {
            freezeWorkList.erase(u);
            simplifyWorkList.push_back(u);
        }
    }

    void ColorAllocatorExecutor::freezeMoves(const AReg &u) {
        for (auto m: nodeMoves(u)) {
            auto x = m.second; // to
            auto y = m.first; // from
            AReg v;
            if (getAlias(y) == getAlias(u)) {
                v = getAlias(x);
            } else {
                v = getAlias(y);
            }
            activeMoves.erase(m);
            frozenMoves.insert(m);
            if (nodeMoves(v).empty() && degree[fruid(v)] < getPRegCnt()) {
                freezeWorkList.erase(v);
                simplifyWorkList.push_back(v);
            }
        }
    }

    VReg ColorAllocatorExecutor::allocSplitVReg(VReg::Type type1) {
        function->vRegType.push_back(type1);
        VReg ret{function->vRegAllocated + 1, type1};
        function->vRegAllocated++;
        return ret;
    }

    void ColorAllocatorExecutor::addEdge(const AReg &u, const AReg &v) {
        if (u == v) {
            return;
        }
        if (auto vReg1 = std::get_if<VReg>(&u)) {
            if (adjList[fruid(u)].count(v) == 0) {
                adjList[fruid(u)].insert(v);
                degree[fruid(u)]++;
            }

        }
        if (auto vReg = std::get_if<VReg>(&v)) {
            if (adjList[fruid(v)].count(u) == 0) {
                adjList[fruid(v)].insert(u);
                degree[fruid(v)]++;
            }
        }
    }

    void ColorAllocatorExecutor::fusionEnableMove(const AReg &reg) {
        for (const auto &reg1: adjList[fruid(reg)]) {
            if (!selectSet.count(reg1) && !coalescedNodes.count(reg1)) {
                for (auto &move1: nodeMoves(reg1)) {
                    if (activeMoves.count(move1)) {
                        activeMoves.erase(move1);
                        workListMoves.insert(move1);
                    }
                }
            }
        }
        for (auto &move1: nodeMoves(reg)) {
            if (activeMoves.count(move1)) {
                activeMoves.erase(move1);
                workListMoves.insert(move1);
            }
        }
    }
} // armv7