//
// Created by 11067 on 2023/7/31.
//

#include "MergeCondExec.h"
#include "../../../utils/config.h"
#include "../instructions/AInstrALU.h"
#include "../instructions/AInstrB.h"
#include "../instructions/AInstrMov.h"
#include "../instructions/AInstrVFP.h"
#include <algorithm>

namespace armv7 {
    /**
     * 使用条件执行替换跳转，基于前端会将短路运算生成基本块序列的假设，ARMv7可以使用条件执行生成更简洁的
     * 指令序列
     * @param function
     */
    void MergeCondExec::mergeCondExec(AFunction *function) {
        auto blocks = function->toVector();
        std::unordered_set<ABlock *> toRemove{};
        auto condExecNoHazard = [](ABlock *block) {
            auto isLowLatency = [](AInstr *instr) {
                if (dynamic_cast<AInstrALU *>(instr) != nullptr ||
                    dynamic_cast<AInstrCMP *>(instr) != nullptr ||
                    dynamic_cast<AInstrLS *>(instr) != nullptr ||
                    dynamic_cast<AInstrMov32 *>(instr) != nullptr ||
                    dynamic_cast<AInstrMov *>(instr) != nullptr ||
                    dynamic_cast<AInstrComment*>(instr) != nullptr
//                    dynamic_cast<AInstrVMov *>(instr) != nullptr ||
//                    dynamic_cast<AInstrPseLdr *>(instr) != nullptr
                    ) {
                    return true;
                } else if (auto b = dynamic_cast<AInstrB *>(instr)) {
                    return !b->isCall() && b->cond == Cond::None;
                } else {
                    return false;
                }
            };
            for (auto instr: *block) {
                if (!isLowLatency(instr)) {
                    return false;
                }
            }
            return true;
        };
        for (auto bb: blocks) {
            if (toRemove.count(bb)) {
                continue;
            }
            if (bb->succBranch == nullptr || bb->failBranch == nullptr) {
                continue;
            }
            auto nxtBlock = bb->succBranch;
            if (nxtBlock->failBranch == nullptr &&
                (nxtBlock->prec.size() == 1 && *(nxtBlock->prec.begin()) == bb) &&
                (nxtBlock->succBranch == bb->failBranch) &&
                nxtBlock->getInstrCntNoComment() < CondExecThreshold &&
                condExecNoHazard(nxtBlock)) {
                if (bb->getSize() + nxtBlock->getSize() > 1024) {
                    goto revMergeCond; // too large for .ltorg
                }
                auto lastInstr = bb->getLast();
                auto last2Instr = lastInstr->getPrev();
                auto cond = last2Instr->cond;
                auto nxtBlockInstrs = nxtBlock->toVector();
                bool cantCond{false};
                for (auto instr: nxtBlockInstrs) {
                    if (!instr->permitCond()) {
                        cantCond = true;
                    }
                }
                if (cantCond) { goto revMergeCond; }
                info("mergeCondExec...");
                lastInstr->removeMe();
                last2Instr->removeMe();
                delete lastInstr;
                delete last2Instr;
                for (auto instr: nxtBlockInstrs) {
                    if (instr != nxtBlockInstrs.back()) {
                        instr->setCond(cond); // 最后一条跳转或者返回不加条件
                        // 大部分指令都可以设条件
                    }
                    bb->addLast(instr);
                    bb->succBranch = nxtBlock->succBranch;
                    bb->failBranch = nullptr;
                    bb->succ = nxtBlock->succ;
                    for (auto nxtBlockSucc: nxtBlock->succ) {
                        nxtBlockSucc->prec.erase(nxtBlock);
                        nxtBlockSucc->prec.insert(bb);
                    }
                    toRemove.insert(nxtBlock);
                }

            }
            revMergeCond:
            // 反转条件，看能否合并
            nxtBlock = bb->failBranch;
            if (nxtBlock == nullptr) { continue; }
            if (nxtBlock->failBranch == nullptr &&
                (nxtBlock->prec.size() == 1 && *(nxtBlock->prec.begin()) == bb) &&
                (nxtBlock->succBranch == bb->succBranch) &&
                nxtBlock->getInstrCntNoComment() < CondExecThreshold &&
                condExecNoHazard(nxtBlock)) {
                if (bb->getSize() + nxtBlock->getSize() > 1024) {
                    continue; // too large for .ltorg
                }
                auto lastInstr = bb->getLast();
                auto last2Instr = lastInstr->getPrev();
                auto cond = revCond(last2Instr->cond);
                auto nxtBlockInstrs = nxtBlock->toVector();
                bool cantCond{false};
                for (auto instr: nxtBlockInstrs) {
                    if (!instr->permitCond()) {
                        cantCond = true;
                    }
                }
                if (cantCond) { continue; }
                info("mergeCondExecRev...");
                lastInstr->removeMe();
                last2Instr->removeMe();
                delete lastInstr;
                delete last2Instr;
                for (auto instr: nxtBlockInstrs) {
                    if (instr != nxtBlockInstrs.back()) {
                        instr->setCond(cond); // 最后一条跳转或者返回不加条件
                        // 大部分指令都可以设条件
                    }
                    bb->addLast(instr);
                    bb->succBranch = nxtBlock->succBranch;
                    bb->failBranch = nullptr;
                    bb->succ = nxtBlock->succ;
                    for (auto nxtBlockSucc: nxtBlock->succ) {
                        nxtBlockSucc->prec.erase(nxtBlock);
                        nxtBlockSucc->prec.insert(bb);
                    }
                    toRemove.insert(nxtBlock);
                }

            }

        }
        for (auto item: toRemove) {
            item->removeMe();
            delete item;
        }

    }

    void MergeCondExec::run(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {
            if (!func->isLib) {
                func->calcPrecSucc();
                mergeCondExec(func);
            }
        }
    }
} // armv7