//
// Created by 11067 on 2023/8/2.
//

#include "Peephole.h"
#include "../instructions/AInstrB.h"
#include "../instructions/AInstrMov.h"
#include "../instructions/AInstrALU.h"

namespace armv7 {
    void Peephole::run(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {
            if (!func->isLib) {

                runOnFunction(func);
            }
        }
    }

    void Peephole::runOnFunction(AFunction *function) {
        cleanupComment(function);
        trivPeephole(function);
        redundantJump(function);
    }

    void Peephole::redundantJump(AFunction *function) {
        bool withJTable{false};
        for (auto bb: *function) {
            bb->aligned = true;
            if (bb->ajTable) {
                withJTable = true;
            }
        }
//        if (withJTable) {
//            return;
//        }
        std::unordered_set<ABlock *> moved{};
        for (auto bb: function->toVector()) {
            if (bb->ajTable) {
                continue;
            }
            auto lastInstr = bb->getLast();
            if (auto b = dynamic_cast<AInstrB *>(lastInstr)) {
                if (b->cond == Cond::None && !b->isCall()) {
                    auto target = std::get<ABlock *>(b->target);
                    auto nxtBlock = bb->getNext();
                    if (nxtBlock != nullptr && nxtBlock == target) {
                        if (!bb->trailingLTORG) {
                            info("peepHole: redundantJump.");
                            b->removeMe();
                            delete b;
                            moved.insert(nxtBlock);
                            moved.insert(bb);
                            nxtBlock->aligned = false;
                            continue;
                        } else if (nxtBlock->getInstrCntNoComment() + bb->getInstrCntNoComment() +
                                   nxtBlock->trailingLTORG + bb->trailingLTORG < 1024) {
                            nxtBlock->trailingLTORG += bb->trailingLTORG;
                            bb->trailingLTORG = 0; // 使用后一个块的常量池
                            info("peepHole: redundantJump.");
                            b->removeMe();
                            delete b;
                            moved.insert(nxtBlock);
                            moved.insert(bb);
                            nxtBlock->aligned = false;
                            continue;
                        }
                    }
                    // 尝试移动bb
                    auto pred = [](ABlock *block) {
                        if (block != nullptr) {
                            if (auto b = dynamic_cast<AInstrB *>(block->getLast())) {
                                if (b->cond == Cond::None && !b->isCall()) {
                                    auto target = std::get<ABlock *>(b->target);
                                    auto nxtBlock = block->getNext();
                                    if (nxtBlock != nullptr && nxtBlock == target) {
                                        return true;// target在原来的位置就有机会合并，换位不赚
                                    }
                                }
                            }
                        }
                        return false;
                    };
                    auto targetPrev = target->getPrev();
                    if (!moved.count(target) && !pred(targetPrev) && bb != target) {
                        target->removeMe();
                        bb->insertAfterMe(target);
                        moved.insert(target);
                        moved.insert(bb);
                        if (!bb->trailingLTORG) {
                            info("peepHole: redundantJump.");
                            b->removeMe();
                            delete b;
                            target->aligned = false;
                            continue;
                        } else if (target->getInstrCntNoComment() + bb->getInstrCntNoComment() +
                                   target->trailingLTORG + bb->trailingLTORG < 1024) {
                            target->trailingLTORG += bb->trailingLTORG;
                            bb->trailingLTORG = 0; // 使用后一个块的常量池
                            info("peepHole: redundantJump.");
                            b->removeMe();
                            delete b;
                            target->aligned = false;
                            continue;
                        }
                    }
                }
            }
        }
    }

    void Peephole::cleanupComment(AFunction *function) {
        for (auto bb: *function) {
            for (auto instr: bb->toVector()) {
                if (auto comment = dynamic_cast<AInstrComment *>(instr)) {
                    comment->removeMe();
                    delete comment;
                }
            }
        }
    }

    void Peephole::trivPeephole(AFunction *function) {
        auto loadAfterStore = [](AInstr *instr1, AInstr *instr2) {
            if (auto str = dynamic_cast<AInstrLS *>(instr1)) {
                if (auto ldr = dynamic_cast<AInstrLS *>(instr2)) {
                    if (str->type != AInstrLS::Type::Store) {
                        return false;
                    }
                    if (ldr->type != AInstrLS::Type::Load) {
                        return false;
                    }
                    if (isSameAddr(str, ldr) && str->isFloat == ldr->isFloat) {
                        info("peepHole: loadAfterStore.");
                        return true;
                    }
                }
            }
            return false;
        };

        auto mulAdd = []() {};
        for (auto bb: *function) {
            auto instrVec = bb->toVector();
            int iinstr = 0;
            while (iinstr + 1 < instrVec.size()) {
                auto instr1 = instrVec[iinstr];
                auto instr2 = instrVec[iinstr + 1];
                if (loadAfterStore(instr1, instr2)) {
                    auto str = dynamic_cast<AInstrLS *>(instr1);
                    auto ldr = dynamic_cast<AInstrLS *>(instr2);
                    if (ldr->rd != str->rd) {
                        auto mov = makeMovR2R(ldr->rd, str->rd);
                        ldr->insertBeforeMe(mov);
                    }
                    ldr->removeMe();
                    delete ldr;
                    iinstr += 2;
                } else {
                    iinstr += 1;
                }
            }
        }
    }

    void Peephole::blockReorder(AFunction *function) {

    }
} // armv7