//
// Created by 11067 on 2023/5/22.
//

#include "MCBlock.h"

#include <utility>

namespace riscv {

    MCBlock::MCBlock(MCFunction *_function, std::string label)
        : label(std::move(label)), function(_function) {}

    // Ir的MemObject转换为后端的Operand
    Reg Mapping::get_operand(const Ir::MemObject &memObject) {
        int mem_id = memObject.get_id();
        auto it = operand_map.find(mem_id);
        if (it == operand_map.end()) {
            Reg op = gen_var_operand();
            operand_map.insert({mem_id, op});
            return op;
        } else {
            return it->second;
        }
    }

    Reg Mapping::get_global_addr(const std::string &name) {
        if (curBlock->globalName2Reg.count(name)) {
            return curBlock->globalName2Reg[name];
        } else {
            Reg addr = gen_var_operand();
            curBlock->globalName2Reg[name] = addr;
            curBlock->push_back(new LoadLocalAddrInstr(addr, name));
            return addr;
        }
    }

    // Ir的Reg转换为后端的Operand
    Reg Mapping::get_operand(Ir::Reg reg) {
        // 常量则生成一个RegConst
        if (reg.is_constant()) {
            if (reg.get_type() == Ir::ScalaType::Int) {
                return gen_const_operand(reg.get_num().int_num);
            } else if (reg.get_type() == Ir::ScalaType::Float) {
                return gen_const_operand(reg.get_num().float_num);
            } else {
                return gen_const_operand(reg.get_num().int_num);
            }
        } else {
            int reg_id = reg.get_id();
            auto it = operand_map.find(reg_id);
            if (it == operand_map.end()) {
                if (irCvt.count(reg_id)) {
                    return operand_map.find(irCvt[reg_id])->second;
                }
                bool isFloat = reg.get_type() == Ir::ScalaType::Float;
                Reg op = gen_var_operand(isFloat);
                operand_map.insert({reg_id, op});
                return op;
            } else {
                return it->second;
            }
        }
        panic("wrong vir reg alloc");
    }

    // 生成RegConst
    Reg Mapping::gen_const_operand(int val) { return {val}; }
    Reg Mapping::gen_const_operand(double val) {
        Reg var = gen_var_operand(true);
        auto bitVal = toUI32(static_cast<float>(val));

        Reg temp = gen_var_operand();
        curBlock->push_back(new MoveU32Instr(temp, bitVal));
        curBlock->push_back(new Bit2FInstr(var, temp));
        return var;
    }

    Reg Mapping::gen_var_operand(bool isFloat) {
        auto type = isFloat ? Ir::Float : Ir::Int;
        vRegType.push_back(type);
        return {reg_num++, isFloat};
    }

    void MCBlock::construct(Ir::BasicBlock *ir_block, Mapping *mapping) {
        mapping->curBlock = this;
        for (auto &ir_instr: ir_block->get_instrs()) {
            Ir::Instruction *instr = ir_instr.get();
            // DefInst
            Instanceof(binary_inst, Ir::BinaryInstr *, instr) {
                Ir::BinaryType type = binary_inst->get_binary_type();
                Reg dest = mapping->get_operand(*binary_inst->get_dst()),
                    src_1 = mapping->get_operand(*binary_inst->get_src_1()),
                    src_2 = mapping->get_operand(*binary_inst->get_src_2());
                switch (type) {
                    case Ir::FADD:
                    case Ir::ADD:
                        push_back(new AddInstr(dest, src_1, src_2));
                        break;
                    case Ir::FSUB:
                    case Ir::SUB:
                        if (src_1.isConst && !src_2.isConst && src_1.int_num == 0) {
                            push_back(new NegInstr(dest, src_2));
                        } else {
                            push_back(new SubInstr(dest, src_1, src_2));
                        }
                        break;
                    case Ir::MUL: {
                        gen_mul_opt(dest, src_1, src_2);
                        break;
                    }
                    case Ir::FDIV:
                    case Ir::SDIV:
                    case Ir::UDIV: {
                        if (!src_1.isConst && src_2.isConst) {
                            int iniValue = src_2.int_num;
                            bool isNeg = iniValue < 0;
                            int value = std::abs(iniValue);
                            if (value == 1) {
                                if (isNeg) {
                                    push_back(new NegInstr(dest, src_1));
                                } else {
                                    push_back(new MoveInstr(dest, src_1));
                                }
                            } else if ((value & (value - 1)) == 0) {
                                int shift = static_cast<int>(std::log2(static_cast<double>(value)));
                                if (value == 2) {
                                    push_back(
                                            new LShrInstr(dest, src_1, mapping->gen_const_operand(31)));
                                    push_back(new AddInstr(dest, dest, src_1));
                                } else {
                                    push_back(
                                            new AShrInstr(dest, src_1, mapping->gen_const_operand(31)));
                                    push_back(new LShrInstr(dest, dest,
                                                            mapping->gen_const_operand(32 - shift)));
                                    push_back(new AddInstr(dest, src_1, dest));
                                }
                                push_back(
                                        new AShrInstr(dest, dest, mapping->gen_const_operand(shift)));
                                if (isNeg) {
                                    push_back(new NegInstr(dest, dest));
                                }
                            } else {
                                int le = std::max(static_cast<int>(std::log2(value)), 1);
                                long long ini = (1ll + ((1ll << (31 + le))) / value) - (1ll << 32);
                                int shift = le - 1;
                                push_back(new Move64Instr(dest, ini));
                                push_back(new MulHi64Instr(dest, dest, src_1));
                                push_back(new AddInstr(dest, dest, src_1));
                                push_back(
                                        new AShrInstr(dest, dest, mapping->gen_const_operand(shift)));
                                push_back(
                                        new SltInstr(get_temp(), src_1, mapping->gen_const_operand(0)));
                                push_back(new AddInstr(dest, dest, get_temp()));
                                if (isNeg) {
                                    push_back(new NegInstr(dest, dest));
                                }
                            }
                        } else {
                            push_back(new DivInstr(dest, src_1, src_2));
                        }
                        break;
                    }
                    case Ir::FREM:
                    case Ir::SREM:
                    case Ir::UREM:
                        // todo 怪怪的
                        if (dest.type == Ir::Float) {
                            Reg temp1 = mapping->gen_var_operand(),
                                temp2 = mapping->gen_var_operand(),
                                dest_temp = mapping->gen_var_operand();
                            push_back(new CvtF2IInstr(temp1, src_1));
                            push_back(new CvtF2IInstr(temp2, src_2));
                            push_back(new RemInstr(dest_temp, temp1, temp2));
                            push_back(new CvtI2FInstr(dest, dest_temp));
                        } else {
                            if (!src_1.isConst && src_2.isConst) {
                                int iniValue = src_2.int_num;
                                int value = std::abs(iniValue);
                                if (value == 1) {
                                    push_back(new MoveInstr(dest, src_1));
                                } else if ((value & (value - 1)) == 0) {
                                    int shift =
                                            static_cast<int>(std::log2(static_cast<double>(value)));
                                    if (value == 2) {
                                        push_back(
                                                new LShrInstr(dest, src_1, mapping->gen_const_operand(31)));
                                        push_back(new AddInstr(dest, dest, src_1));
                                    } else {
                                        push_back(
                                                new AShrInstr(dest, src_1, mapping->gen_const_operand(31)));
                                        push_back(new LShrInstr(
                                                dest, dest, mapping->gen_const_operand(32 - shift)));
                                        push_back(new AddInstr(dest, dest, src_1));
                                    }
                                    if (isImm12(value - 1)) {
                                        push_back(new BICInstr(dest, dest,
                                                               mapping->gen_const_operand(value - 1)));
                                        push_back(new SubInstr(dest, src_1, dest));
                                    } else {
                                        push_back(new LShrInstr(dest, dest, shift));
                                        push_back(new ShlInstr(dest, dest,
                                                               mapping->gen_const_operand(shift)));
                                        push_back(new SubInstr(dest, src_1, dest));
                                    }
                                } else {
                                    int le = std::max(static_cast<int>(std::log2(value)), 1);
                                    long long ini =
                                            (1ll + ((1ll << (31 + le))) / value) - (1ll << 32);
                                    int shift = le - 1;
                                    push_back(new Move64Instr(dest, ini));
                                    push_back(new MulHi64Instr(dest, dest, src_1));
                                    push_back(new AddInstr(dest, dest, src_1));
                                    push_back(
                                            new AShrInstr(dest, dest, mapping->gen_const_operand(shift)));
                                    push_back(new SltInstr(get_temp(), src_1,
                                                           mapping->gen_const_operand(0)));
                                    push_back(new AddInstr(dest, dest, get_temp()));
                                    gen_mul_opt(dest, dest, mapping->gen_const_operand(value));
                                    push_back(new SubInstr(dest, src_1, dest));
                                }
                            } else {
                                push_back(new RemInstr(dest, src_1, src_2));
                            }
                        }
                        break;
                    case Ir::AND:
                        push_back(new AndInstr(dest, src_1, src_2));
                        break;
                    case Ir::OR:
                        push_back(new OrInstr(dest, src_1, src_2));
                        break;
                    case Ir::XOR:
                        push_back(new XorInstr(dest, src_1, src_2));
                        break;
                    case Ir::SHL:
                        push_back(new ShlInstr(dest, src_1, src_2));
                        break;
                    case Ir::LSHR:
                        push_back(new LShrInstr(dest, src_1, src_2));
                        break;
                    case Ir::ASHR:
                        push_back(new AShrInstr(dest, src_1, src_2));
                        break;
                    case Ir::FNEG:
                        warn("fneg appears");
                        push_back(new NegInstr(dest, src_1));
                        break;
                        //            case Ir::FNEG: push_back(new OrInstr(dest, src_1, src_2));
                        //            break;
                    case Ir::EQ:
                        push_back(new EqInstr(dest, src_1, src_2));
                        break;
                    case Ir::NE:
                        push_back(new NeInstr(dest, src_1, src_2));
                        break;
                    case Ir::SGT:
                        push_back(new GreInstr(dest, src_1, src_2));
                        break;
                    case Ir::SGE:
                        push_back(new GeqInstr(dest, src_1, src_2));
                        break;
                    case Ir::SLE:
                        push_back(new SleInstr(dest, src_1, src_2));
                        break;
                    case Ir::SLT:
                        push_back(new SltInstr(dest, src_1, src_2));
                        break;
                    default:
                        panic("what's this instr???");
                        break;
                }
            }
            else Instanceof(branch_inst, Ir::BranchInstr *, instr) {
                if (mapping->irBB2MCBB.find(branch_inst->get_bb_1()) ==
                    mapping->irBB2MCBB.end()) {
                    panic("unknown BB to branch");
                } else {
                    // 无条件跳转
                    if (branch_inst->get_bb_2() == nullptr) {
                        MCBlock *jump_target = mapping->irBB2MCBB[branch_inst->get_bb_1()];
                        auto next_block = this->getNext();
                        push_back(new GotoInstr(mapping->irBB2MCBB[branch_inst->get_bb_1()]));
                    } else {
                        MCBlock *true_target = mapping->irBB2MCBB[branch_inst->get_bb_1()],
                                *false_target = mapping->irBB2MCBB[branch_inst->get_bb_2()];
                        if (mapping->irBB2MCBB.find(branch_inst->get_bb_2()) ==
                            mapping->irBB2MCBB.end()) {
                            panic("unknown BB to branch");
                        } else {
                            auto cond = mapping->get_operand(*branch_inst->get_reg());
                            push_back(new IfConInstr(
                                    mapping->irBB2MCBB[branch_inst->get_bb_1()], cond));
                            push_back(
                                    new GotoInstr(mapping->irBB2MCBB[branch_inst->get_bb_2()]));
                        }
                    }
                }
            }
            else Instanceof(store_inst, Ir::StoreInstr *, instr) {
                Reg val = mapping->get_operand(*store_inst->get_val());
                //            Reg addr = mapping->get_operand(*store_inst->get_addr());
                if (function->obj_stack_arr.find(store_inst->get_addr().get()) !=
                    function->obj_stack_arr.end()) {
                    auto offset = mapping->gen_const_operand(
                            function->obj_stack_arr.find(store_inst->get_addr().get())->second);
                    if (val.type == Ir::Int) {
                        if (val.isConst) {
                            if (val.int_num == 0) {
                                push_back(new StoreInstr(get_zero(), get_sp(), offset));
                            } else {
                                Reg temp = mapping->gen_var_operand();
                                push_back(new MoveInstr(temp, val));
                                push_back(new StoreInstr(temp, get_sp(), offset));
                            }
                        } else {
                            push_back(new StoreInstr(val, get_sp(), offset));
                        }
                    } else if (val.type == Ir::Float) {
                        push_back(new StoreFInstr(val, get_sp(), offset));
                    }
                } else {
                    Reg addr = mapping->get_operand(*store_inst->get_addr());
                    if (function->ls_opt_map.count(addr)) {
                        //          push_back(new DebugInstr());
                        auto pair = function->ls_opt_map[addr];
                        if (val.type == Ir::Int) {
                            if (val.isConst) {
                                if (val.int_num == 0) {
                                    push_back(
                                            new StoreInstr(get_zero(), pair.first,
                                                           mapping->gen_const_operand(pair.second)));
                                } else {
                                    Reg temp = mapping->gen_var_operand();
                                    push_back(new MoveInstr(temp, val));
                                    push_back(
                                            new StoreInstr(temp, pair.first,
                                                           mapping->gen_const_operand(pair.second)));
                                }
                            } else {
                                push_back(new StoreInstr(
                                        val, pair.first, mapping->gen_const_operand(pair.second)));
                            }
                        } else if (val.type == Ir::Float) {
                            push_back(new StoreFInstr(val, pair.first,
                                                      mapping->gen_const_operand(pair.second)));
                        }
                    } else {
                        if (val.type == Ir::Int) {
                            if (val.isConst) {
                                if (val.int_num == 0) {
                                    push_back(new StoreInstr(get_zero(), addr));
                                } else {
                                    Reg temp = mapping->gen_var_operand();
                                    push_back(new MoveInstr(temp, val));
                                    push_back(new StoreInstr(temp, addr));
                                }
                            } else {
                                push_back(new StoreInstr(val, addr));
                            }
                        } else if (val.type == Ir::Float) {
                            push_back(new StoreFInstr(val, addr));
                        }
                    }
                }
            }
            else Instanceof(storeg_inst, Ir::StoreGlobal *, instr) {
                std::string var_label = storeg_inst->get_addr()->get_global_name();
                Reg val = mapping->get_operand(*storeg_inst->get_val());
                //      Reg addr = mapping->gen_var_operand();
                Reg addr = mapping->get_global_addr(var_label);
                //      push_back(new LoadLocalAddrInstr(addr, var_label));
                if (val.type == Ir::Int) {
                    if (val.isConst) {
                        if (val.int_num == 0) {
                            push_back(new StoreInstr(get_zero(), addr));
                        } else {
                            Reg temp = mapping->gen_var_operand();
                            push_back(new MoveInstr(temp, val));
                            push_back(new StoreInstr(temp, addr));
                        }
                    } else {
                        push_back(new StoreInstr(val, addr));
                    }
                } else if (val.type == Ir::Float) {
                    push_back(new StoreFInstr(val, addr));
                }
            }
            else Instanceof(alloc_inst, Ir::AllocaInstr *, instr) {
                auto pMemObj = alloc_inst->get_addr().get();
                Instanceof(pVec, Ir::Vector *, pMemObj) {
                    auto offset = function->stackTop * INT_SIZE;
                    function->vec_stack_arr.insert({pVec, offset});
                    function->stackTop += get_vec_size(pVec->get_dims());
                }
                else Instanceof(pMem, Ir::MemObject *, pMemObj) {
                    auto offset = function->stackTop * INT_SIZE;
                    function->obj_stack_arr.insert({pMem, offset});
                    function->stackTop += 1;
                }
            }
            else Instanceof(load_inst, Ir::LoadInstr *, instr) {
                Reg dst = mapping->get_operand(*load_inst->get_dst());
                if (function->obj_stack_arr.find(load_inst->get_addr().get()) !=
                    function->obj_stack_arr.end()) {
                    auto offset = mapping->gen_const_operand(
                            function->obj_stack_arr.find(load_inst->get_addr().get())->second);
                    if (dst.type == Ir::Int) {
                        push_back(new LoadInstr(dst, get_sp(), offset));
                    } else {
                        push_back(new LoadFInstr(dst, get_sp(), offset));
                    }
                } else {
                    Reg addr = mapping->get_operand(*load_inst->get_addr());
                    if (function->ls_opt_map.count(addr)) {
                        //                    push_back(new DebugInstr());
                        auto pair = function->ls_opt_map[addr];
                        if (dst.type == Ir::Float) {
                            push_back(new LoadFInstr(dst, pair.first,
                                                     mapping->gen_const_operand(pair.second)));
                        } else {
                            push_back(new LoadInstr(dst, pair.first,
                                                    mapping->gen_const_operand(pair.second)));
                        }
                    } else {
                        if (dst.type == Ir::Float) {
                            push_back(new LoadFInstr(dst, addr));
                        } else {
                            push_back(new LoadInstr(dst, addr));
                        }
                    }
                }
            }
            else Instanceof(loadg_inst, Ir::LoadGlobal *, instr) {
                Reg dst = mapping->get_operand(*loadg_inst->get_dst());
                std::string var_label = loadg_inst->get_addr()->get_global_name();
                //      Reg addr = mapping->gen_var_operand();
                Reg addr = mapping->get_global_addr(var_label);
                //      push_back(new LoadLocalAddrInstr(addr, var_label));
                if (dst.type == Ir::Int) {
                    push_back(new LoadInstr(dst, addr));
                } else if (dst.type == Ir::Float) {
                    push_back(new LoadFInstr(dst, addr));
                }
            }
            // 从数组中加载元素
            else Instanceof(loada_inst, Ir::LoadArray *, instr) {
                // 加载基地址
                auto irVec = loada_inst->get_addr().get();
                // 全局变量
                auto global_map = function->compileUnit->globalVarMap;
                // 函数栈上
                auto vec_stack_addr = function->vec_stack_arr;
                // 参数数组
                auto para_vec_map = function->para_vec_base;
                Reg addr, op_offset;
                if (global_map.find(irVec) != global_map.end()) {
                    //        addr = mapping->gen_var_operand();
                    auto data = global_map.find(irVec)->second->label;
                    const std::string &data_label = data;
                    addr = mapping->get_global_addr(data_label);
                    //        push_back(new LoadLocalAddrInstr(addr, data_label));
                } else if (vec_stack_addr.find(irVec) != vec_stack_addr.end()) {
                    addr = mapping->gen_var_operand();
                    auto offset = vec_stack_addr.find(irVec)->second;
                    push_back(
                            new Add64Instr(addr, get_sp(), mapping->gen_const_operand(offset)));
                } else if (para_vec_map.find(irVec) != para_vec_map.end()) {
                    addr = para_vec_map.find(irVec)->second;
                } else {
                    addr = mapping->get_operand(*irVec);
                    //                panic("load from unknown vector");
                }
                // 计算偏移量
                int ele_offset = 1;
                auto &arrayDims = loada_inst->get_addr()->get_dims();
                for (auto &l: arrayDims) {
                    ele_offset *= l ? l : 1;
                }
                if (arrayDims.size() > loada_inst->get_indexs().size()) {
                    panic("loadArray trying to get subArray");
                } else if (arrayDims.size() < loada_inst->get_indexs().size()) {
                    panic("loadArray out of range");
                }
                op_offset = mapping->gen_var_operand();
                vector<bool> is_const_indexs{};
                for (auto &i: loada_inst->get_indexs()) {
                    auto index = i.get();
                    if (index->is_constant()) {
                        is_const_indexs.push_back(true);
                    } else {
                        is_const_indexs.push_back(false);
                    }
                }
                bool all_const = true, initial = false;
                int offset_temp = 0;
                for (auto i = 0; i < loada_inst->get_indexs().size(); i++) {
                    ele_offset /= arrayDims[i] ? arrayDims[i] : 1;
                    auto index = loada_inst->get_indexs()[i].get();
                    if (is_const_indexs[i]) {
                        offset_temp += index->get_num().int_num * ele_offset * INT_SIZE;
                    } else if (!is_const_indexs[i]) {
                        all_const = false;
                        Reg op_index = mapping->get_operand(*index);
                        Reg op_temp = mapping->gen_var_operand();
                        gen_mul_opt(op_temp, op_index,
                                    mapping->gen_const_operand(ele_offset * INT_SIZE));
                        if (!initial) {
                            push_back(new MoveInstr(op_offset, op_temp));
                            initial = true;
                        } else {
                            push_back(new AddInstr(op_offset, op_offset, op_temp));
                        }
                    }
                }
                Reg dst = mapping->get_operand(*loada_inst->get_dst());
                if (all_const) {
                    // todo 可以优化 少一条加法指令
                    function->ls_opt_map[dst] = std::make_pair(addr, offset_temp);
                    //                push_back(new DebugInstr("add"));
                    //        push_back(
                    //            new Add64Instr(dst, addr,
                    //            mapping->gen_const_operand(offset_temp)));
                } else {
                    if (offset_temp != 0) {
                        push_back(new AddInstr(op_offset, op_offset,
                                               mapping->gen_const_operand(offset_temp)));
                    }
                    push_back(new Add64Instr(dst, addr, op_offset));
                }

                //            op_offset = mapping->gen_var_operand();
                //            push_back(new MoveInstr(op_offset,
                //            mapping->gen_const_operand(0))); for (auto i = 0; i <
                //            loada_inst->get_indexs().size(); i++) {
                //                ele_offset /= arrayDims[i] ? arrayDims[i] : 1;
                //                auto index = loada_inst->get_indexs()[i].get();
                //                if (index->is_constant()) {
                //                    Reg op_index =
                //                    mapping->gen_const_operand(index->get_num().int_num
                //                    * ele_offset * INT_SIZE); push_back(new
                //                    AddInstr(op_offset, op_offset, op_index));
                //                } else {
                //                    Reg op_index = mapping->get_operand(*index);
                //                    Reg op_temp = mapping->gen_var_operand();
                //                    gen_mul_opt(op_temp, op_index,
                //                    mapping->gen_const_operand(ele_offset * INT_SIZE));
                //                    push_back(new AddInstr(op_offset, op_offset,
                //                    op_temp));
                //                }
                //            }
                //            Reg dst = mapping->get_operand(*loada_inst->get_dst());
                //            push_back(new Add64Instr(dst, addr, op_offset));
            }
            // 从数组中加载子数组 只用于函数参数
            else Instanceof(loadv_inst, Ir::LoadVector *, instr) {
                // 加载基地址
                auto irVec = loadv_inst->get_addr().get();
                // 全局变量
                auto global_map = function->compileUnit->globalVarMap;
                // 函数栈上
                auto vec_stack_addr = function->vec_stack_arr;
                // 参数数组
                auto para_vec_map = function->para_vec_base;
                Reg addr, op_offset;
                if (global_map.find(irVec) != global_map.end()) {
                    //        addr = mapping->gen_var_operand();
                    auto data = global_map.find(irVec)->second->label;
                    const std::string &data_label = data;
                    addr = mapping->get_global_addr(data_label);
                    //        push_back(new LoadLocalAddrInstr(addr, data_label));
                } else if (vec_stack_addr.find(irVec) != vec_stack_addr.end()) {
                    addr = mapping->gen_var_operand();
                    auto offset = vec_stack_addr.find(irVec)->second;
                    push_back(
                            new Add64Instr(addr, get_sp(), mapping->gen_const_operand(offset)));
                } else if (para_vec_map.find(irVec) != para_vec_map.end()) {
                    addr = para_vec_map.find(irVec)->second;
                } else {
                    addr = mapping->get_operand(*irVec);
                    //                panic("load from unknown vector");
                }
                // 计算偏移量
                int ele_offset = 1;
                auto &arrayDims = loadv_inst->get_addr()->get_dims();
                for (auto &l: arrayDims) {
                    //        ele_offset *= l;
                    ele_offset *= l ? l : 1;
                }
                if (arrayDims.size() == loadv_inst->get_indexs().size()) {
                    panic("loadVector trying to get single element");
                } else if (arrayDims.size() < loadv_inst->get_indexs().size()) {
                    panic("loadVector out of range");
                }

                op_offset = mapping->gen_var_operand();
                vector<bool> is_const_indexs{};
                for (auto &i: loadv_inst->get_indexs()) {
                    auto index = i.get();
                    if (index->is_constant()) {
                        is_const_indexs.push_back(true);
                    } else {
                        is_const_indexs.push_back(false);
                    }
                }
                bool all_const = true, initial = false;
                int offset_temp = 0;
                for (auto i = 0; i < loadv_inst->get_indexs().size(); i++) {
                    //        ele_offset /= arrayDims[i];
                    ele_offset /= arrayDims[i] ? arrayDims[i] : 1;
                    auto index = loadv_inst->get_indexs()[i].get();
                    if (is_const_indexs[i]) {
                        offset_temp += index->get_num().int_num * ele_offset * INT_SIZE;
                    } else if (!is_const_indexs[i]) {
                        all_const = false;
                        Reg op_index = mapping->get_operand(*index);
                        Reg op_temp = mapping->gen_var_operand();
                        gen_mul_opt(op_temp, op_index,
                                    mapping->gen_const_operand(ele_offset * INT_SIZE));
                        if (!initial) {
                            push_back(new MoveInstr(op_offset, op_temp));
                            initial = true;
                        } else {
                            push_back(new AddInstr(op_offset, op_offset, op_temp));
                        }
                    }
                }
                Reg dst = mapping->get_operand(*loadv_inst->get_dst());
                if (all_const) {
                    push_back(
                            new Add64Instr(dst, addr, mapping->gen_const_operand(offset_temp)));
                } else {
                    if (offset_temp != 0) {
                        push_back(new AddInstr(op_offset, op_offset,
                                               mapping->gen_const_operand(offset_temp)));
                    }
                    push_back(new Add64Instr(dst, addr, op_offset));
                }

                //            op_offset = mapping->gen_var_operand();
                //            push_back(new MoveInstr(op_offset,
                //            mapping->gen_const_operand(0))); for (auto i = 0; i <
                //            loadv_inst->get_indexs().size(); i++) {
                //                ele_offset /= arrayDims[i];
                //                auto index = loadv_inst->get_indexs()[i].get();
                //                if (index->is_constant()) {
                //                    Reg op_index =
                //                    mapping->gen_const_operand(index->get_num().int_num
                //                    * ele_offset * INT_SIZE); push_back(new
                //                    AddInstr(op_offset, op_offset, op_index));
                //                } else {
                //                    Reg op_index = mapping->get_operand(*index);
                //                    Reg op_temp = mapping->gen_var_operand();
                //                    gen_mul_opt(op_temp, op_index,
                //                    mapping->gen_const_operand(ele_offset * INT_SIZE));
                //                    push_back(new AddInstr(op_offset, op_offset,
                //                    op_temp));
                //                }
                //            }
                //            Reg dst = mapping->get_operand(*loadv_inst->get_dst());
                //            push_back(new Add64Instr(dst, addr, op_offset));
            }
            else Instanceof(call_inst, Ir::CallInstr *, instr) {
                //
                if (call_inst->get_function()->is_lib() &&
                    function->compileUnit->atriBuiltinName.count(
                            call_inst->get_function()->get_name())) {
                    function->compileUnit->genAtriBuiltin = true;
                }
                // 取参数
                int int_arg_cnt = 0, float_arg_cnt = 0;
                auto cur_func = call_inst->get_function();
                auto params = call_inst->get_params();
                for (const auto &arg: params) {
                    Instanceof(reg, Ir::Reg *, arg.get()) {
                        Reg src = mapping->get_operand(*reg);
                        if (reg->get_type() == Ir::Int) {
                            if (int_arg_cnt < ARG_REG_NUM) {
                                push_back(
                                        new MoveInstr(get_reg_by_id(ARG_REGS[int_arg_cnt++]), src));
                            } else {
                                int offset = -(int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
                                if (src.isConst) {
                                    if(src.int_num==0) {
                                        push_back(new StoreInstr(get_zero(), get_sp(),
                                                                 mapping->gen_const_operand(offset)));
                                    } else {
                                        Reg temp = mapping->gen_var_operand();
                                        push_back(new MoveInstr(temp, src));
                                        push_back(new StoreInstr(temp, get_sp(),
                                                                 mapping->gen_const_operand(offset)));
                                    }
                                } else {
                                    push_back(new StoreInstr(src, get_sp(),
                                                             mapping->gen_const_operand(offset)));
                                }
                                int_arg_cnt++;
                            }
                        } else if (arg->get_type() == Ir::Float) {
                            if (float_arg_cnt < ARG_REG_NUM) {
                                push_back(new MoveInstr(get_freg_by_id(ARG_REGS[float_arg_cnt++]),
                                                        src));
                            } else {
                                int offset = -(int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
                                push_back(new StoreFInstr(src, get_sp(),
                                                          mapping->gen_const_operand(offset)));
                                float_arg_cnt++;
                            }
                        } else {
                            assert(false);
                        }
                    }
                    else Instanceof(vec, Ir::Vector *, arg.get()) {
                        Reg src = mapping->get_operand(*vec);
                        if (int_arg_cnt < ARG_REG_NUM) {
                            push_back(
                                    new MoveInstr(get_reg_by_id(ARG_REGS[int_arg_cnt++]), src));
                        } else {
                            int offset = -(int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
                            push_back(new StoreInstr(src, get_sp(),
                                                     mapping->gen_const_operand(offset)));
                            int_arg_cnt++;
                        }
                    }
                }
                bool is_float = false;
                if (!call_inst->is_void_flag() &&
                    call_inst->get_dst()->get_type() == Ir::Float) {
                    is_float = true;
                }
                push_back(new CallInstr(cur_func->get_name(), call_inst->is_void_flag(),
                                        is_float, int_arg_cnt, float_arg_cnt));
                // 加载返回值
                if (!call_inst->is_void_flag()) {
                    Reg dest = mapping->get_operand(*call_inst->get_dst());
                    if (call_inst->get_dst()->get_type() == Ir::Int) {
                        push_back(new MoveInstr(dest, get_ret()));
                    } else if (call_inst->get_dst()->get_type() == Ir::Float) {
                        push_back(new MoveInstr(dest, get_ret(true)));
                    } else {
                        assert(false);
                    }
                }
            }
            else Instanceof(ret_inst, Ir::Ret *, instr) {
                if (!ret_inst->is_void_flag()) {
                    Reg ret = mapping->get_operand(*ret_inst->get_src());
                    push_back(new RetureInstr(ret));
                    this->function->exitBlock.push_back(this);
                } else {
                    push_back(new RetureInstr());
                    this->function->exitBlock.push_back(this);
                }
            }
            else Instanceof(phi_inst, Ir::PhiInstr *, instr) {
                Reg dest = mapping->get_operand(*phi_inst->get_dst());
                auto r_phi = new PhiInstr(dest);
                for (auto &[block, reg]: phi_inst->get_bb_2_value()) {
                    if (reg->is_constant()) {
                        if (reg->get_type() == Ir::Float) {
                            auto holder = Reg();
                            holder.type = Ir::Float;
                            holder.isConst = true;
                            holder.float_num = reg->get_num().float_num;
                            r_phi->pairs.emplace_back(holder, mapping->irBB2MCBB[block]);
                        } else if (reg->get_type() == Ir::Int ||
                                   reg->get_type() == Ir::Bool) {
                            r_phi->pairs.emplace_back(
                                    mapping->gen_const_operand(reg->get_num().int_num),
                                    mapping->irBB2MCBB[block]);
                        }
                    } else {
                        r_phi->pairs.emplace_back(mapping->get_operand(*reg),
                                                  mapping->irBB2MCBB[block]);
                    }
                }
                push_back(r_phi);
            }
            else Instanceof(cvt_inst, Ir::Convert *, instr) {
                auto src = cvt_inst->get_src().get();
                auto dst = cvt_inst->get_dst().get();
                if (src->get_type() == Ir::Bool && dst->get_type() == Ir::Int) {
                    if (src->is_constant()) {
                        auto src_op = mapping->get_operand(*src);
                        auto dst_op = mapping->get_operand(*dst);
                        push_back(new MoveInstr(dst_op, src_op));
                    } else {
                        mapping->irCvt.insert({dst->get_id(), src->get_id()});
                    }
                } else {
                    auto src_op = mapping->get_operand(*src);
                    auto dst_op = mapping->get_operand(*dst);
                    // float -> int
                    if (src->get_type() == Ir::Float && dst->get_type() == Ir::Int) {
                        push_back(new CvtF2IInstr(dst_op, src_op));
                    }
                    // int -> float
                    else if (src->get_type() == Ir::Int && dst->get_type() == Ir::Float) {
                        if (src_op.isConst) {
                            if(src_op.int_num==0){
                                push_back(new CvtI2FInstr(dst_op, get_zero()));
                            } else {
                                auto temp = mapping->gen_var_operand();
                                push_back(new MoveInstr(temp, src_op));
                                push_back(new CvtI2FInstr(dst_op, temp));
                            }
                        } else {
                            push_back(new CvtI2FInstr(dst_op, src_op));
                        }
                    }
                }
            }
        }
    }

    void MCBlock::gen_mul_opt(const Reg &op1, const Reg &op2, const Reg &op3) {
        auto mapping = this->function->map;
        if (op2.isConst && !op3.isConst) {
            int iniValue = op2.int_num;
            bool isNeg = iniValue < 0;
            int value = std::abs(iniValue);
            if (value == 0) {
                push_back(new MoveInstr(op1, mapping.gen_const_operand(0)));
            } else if ((value & (value - 1)) == 0) {
                int shift = static_cast<int>(std::log2(static_cast<double>(value)));
                push_back(new ShlInstr(op1, op3, mapping.gen_const_operand(shift)));
                if (isNeg) {
                    push_back(new NegInstr(op1, op1));
                }
            } else if (value == 3) {
                push_back(new AddInstr(op1, op3, op3));
                push_back(new AddInstr(op1, op1, op3));
                if (isNeg) {
                    push_back(new NegInstr(op1, op1));
                }
            } else {
                push_back(new MulInstr(op1, op2, op3));
            }
        } else if (!op2.isConst && op3.isConst) {
            int iniValue = op3.int_num;
            bool isNeg = iniValue < 0;
            int value = std::abs(iniValue);
            if (value == 0) {
                push_back(new MoveInstr(op1, mapping.gen_const_operand(0)));
            } else if ((value & (value - 1)) == 0) {
                int shift = static_cast<int>(std::log2(static_cast<double>(value)));
                push_back(new ShlInstr(op1, op2, mapping.gen_const_operand(shift)));
                if (isNeg) {
                    push_back(new NegInstr(op1, op1));
                }
            } else if (value == 3) {
                push_back(new AddInstr(op1, op2, op2));
                push_back(new AddInstr(op1, op1, op2));
                if (isNeg) {
                    push_back(new NegInstr(op1, op1));
                }
            } else {
                push_back(new MulInstr(op1, op2, op3));
            }
        } else {
            push_back(new MulInstr(op1, op2, op3));
        }
    }

    void MCBlock::clear_active() {
        activeIn.clear();
        activeOut.clear();
        activeDef.clear();
        activeUse.clear();
    }

    void MCBlock::push_back(MRInstr *instr) { addLast(instr); }

    void MCBlock::push_back(const std::vector<MRInstr *> &instrs) {
        for (auto instr: instrs) {
            addLast(instr);
        }
    }

    void MCBlock::gen_asm(std::ostream &out) {
        out << label << ":\n";
        for (const auto &instr: *this) {
            out << '\t';
            instr->gen_asm(out);
        }
    }

    void MCBlock::clear() {
        auto instrList = this->toVector();
        for (auto &instr: instrList) {
            delete instr;
        }
    }

}// namespace riscv