//
// Created by 11067 on 2023/6/25.
//

#include "MCCompileUnit.h"

namespace riscv {
    MCCompileUnit::MCCompileUnit(Ir::CompileUnit &ir) : block_num(0) {
        // 自定义函数
        atriBuiltinName = {"_atri_memset", "_atri_memcpy"};
        // _atri_memcpy(dest,src,len) len以字节为单位，要求4byte对齐,
        // _atri_memset(dest, len) len以字节为单位，要求4byte对齐
        supportFunctions = {
                R"(.global _atri_memset
_atri_memset:
    # 参数：a0 - 目标地址，a1 - 长度
    beqz a1, _atri_end_memset  # 如果长度为0，直接返回
    andi t1, a0, 7
    beqz t1, _atri_loop_memset # 如果地址对齐，进入循环
    sw t0, 0(a0)         # 否则，先将一个字写入目标地址
    addi a0, a0, 4       # 目标地址 += 4
    addi a1, a1, -4      # 长度 -= 4
_atri_loop_memset:
    sd t0, 0(a0)         # 在循环中，将一个字写入目标地址
    addi a0, a0, 8       # 目标地址 += 8
    addi a1, a1, -8      # 长度 -= 8
    bnez a1, _atri_loop_memset # 如果长度不为0，继续循环
_atri_end_memset:
    ret)",
                R"(.global _atri_memcpy
_atri_memcpy:
    # 参数：a0 - 目标地址，a1 - 源地址，a2 - 长度
    beqz a2, _atri_end_memcpy  # 如果长度为0，直接返回
    andi t1, a0, 7
    beqz t1, _atri_loop_memcpy # 如果地址对齐，进入循环
    lw t0, 0(a1)         # 否则，先从源地址读取一个字
    sw t0, 0(a0)         # 然后将这个字写入目标地址
    addi a0, a0, 4       # 目标地址 += 4
    addi a1, a1, 4       # 源地址 += 4
    addi a2, a2, -4      # 长度 -= 4
_atri_loop_memcpy:
    ld t0, 0(a1)         # 在循环中，从源地址读取一个字
    addi a1, a1, 8       # 源地址 += 8
    sd t0, 0(a0)         # 然后将这个字写入目标地址
    addi a0, a0, 8       # 目标地址 += 8

    addi a2, a2, -8      # 长度 -= 8
    bnez a2, _atri_loop_memcpy # 如果长度不为0，继续循环
_atri_end_memcpy:
    ret)"

        };

        // 存储全局数据
        for (const auto &glob_arr: ir.get_global_int_arrays()) {
            bool isZeroed{true};
            for (auto &num: glob_arr.second) {
                if (num != 0) {
                    isZeroed = false;
                    break;
                }
            }
            if (!isZeroed) {
                MCData *mcData{makeMCDataInt(glob_arr.first, glob_arr.second)};
                //            data.emplace_back(mcData);
                globalVecNameMap.insert({glob_arr.first, mcData});
            }
        }
        for (const auto &glob_arr: ir.get_global_float_arrays()) {
            bool isZeroed{true};
            for (auto &num: glob_arr.second) {
                if (num != 0) {
                    isZeroed = false;
                    break;
                }
            }
            if (!isZeroed) {
                MCData *mcData{makeMCDataFloat(glob_arr.first, glob_arr.second)};
                //            data.emplace_back(mcData);
                globalVecNameMap.insert({glob_arr.first, mcData});
            }
        }
        for (const auto &glob_val: ir.get_globals()) {
            Instanceof(reg, Ir::Reg *, glob_val.get()) {
                if (reg->get_type() == Ir::Float) {
                    MCData *mcData{
                            makeMCDataFloat(reg->get_global_name(),
                                            {static_cast<float>(reg->get_num().float_num)})};
                    globalVarMap.insert({glob_val.get(), mcData});
                } else if (reg->get_type() == Ir::Int) {
                    MCData *mcData{
                            makeMCDataInt(reg->get_global_name(), {reg->get_num().int_num})};
                    globalVarMap.insert({glob_val.get(), mcData});
                }
            }
            Instanceof(vec, Ir::Vector *, glob_val.get()) {
                if (vec->get_type() == Ir::Float) {
                    auto nameMap = globalVecNameMap;
                    if (nameMap.find(vec->get_name()) != nameMap.end()) {
                        globalVarMap.insert(
                                {glob_val.get(), globalVecNameMap.at(vec->get_name())});
                    } else {
                        // 全局数组没有初始化列表
                        MCData *pMcData{makeMCDataFloatZeroed(vec->get_name(),
                                                              get_vec_size(vec->get_dims()))};
                        //                    data.emplace_back(pMcData);
                        globalVarMap.insert({glob_val.get(), pMcData});
                    }
                } else if (vec->get_type() == Ir::Int) {
                    auto nameMap = globalVecNameMap;
                    if (nameMap.find(vec->get_name()) != nameMap.end()) {
                        globalVarMap.insert(
                                {glob_val.get(), globalVecNameMap.at(vec->get_name())});
                    } else {
                        // 全局数组没有初始化列表
                        MCData *pMcData{makeMCDataIntZeroed(vec->get_name(),
                                                            get_vec_size(vec->get_dims()))};
                        //                    data.emplace_back(pMcData);
                        globalVarMap.insert({glob_val.get(), pMcData});
                    }
                }
            }
        }
        // 存储函数 遍历ir中的函数
        for (auto &ir_func: ir.get_functions()) {
            if (!ir_func->is_lib())
                // 为每一个函数分配后端函数
                addLast(new MCFunction(this, ir_func ? ir_func.get() : nullptr));
        }
    }

    void MCCompileUnit::gen_global_var(std::ostream &out) {
        for (const auto &g_var: globalVarMap) {
            if (auto ivec = std::get_if<std::vector<IntOrZero>>(&g_var.second->data)) {
                for (const auto &intOrZero: *ivec) {
                    //        out << ".globl\t" + g_var.second->label + "\n";
                    if (auto zero_num = std::get_if<int>(&intOrZero)) {
                        out << ".bss\n";
                        out << ".align\t3\n";
                        out << g_var.second->label + ":\n";
                        out << ".space\t" << *zero_num * INT_SIZE << "\n";
                    } else if (auto intVec = std::get_if<std::vector<int>>(&intOrZero)) {
                        out << ".data\n";
                        if (intVec->size() == 1) {
                            out << ".align\t3\n";
                            out << g_var.second->label + ":\n";
                            for (const auto &intElm: *intVec)
                                out << ".word\t" << intElm << "\n";
                        } else {
                            out << ".align\t3\n";
                            out << g_var.second->label + ":\n";
                            for (const auto &intElm: *intVec)
                                out << ".word\t" << intElm << "\n";
                        }
                    }
                }
            } else if (auto fvec =
                               std::get_if<std::vector<FloatOrZero>>(&g_var.second->data)) {
                for (const auto &floatOrZero: *fvec) {
                    //        out << ".globl\t" + g_var.second->label + "\n";
                    if (auto zero_num = std::get_if<int>(&floatOrZero)) {
                        out << ".bss\n";
                        out << ".align\t3\n";
                        out << g_var.second->label + ":\n";
                        out << ".space\t" << *zero_num * INT_SIZE << "\n";
                    } else if (auto floatVec =
                                       std::get_if<std::vector<float>>(&floatOrZero)) {
                        out << ".data\n";
                        if (floatVec->size() == 1) {
                            out << ".align\t3\n";
                            out << g_var.second->label + ":\n";
                            for (const auto &floatElm: *floatVec)
                                out << ".word\t" << toUI32(floatElm) << "\n";
                        } else {
                            out << ".align\t3\n";
                            out << g_var.second->label + ":\n";
                            for (const auto &floatElm: *floatVec)
                                out << ".word\t" << toUI32(floatElm) << "\n";
                        }
                    }
                }
            }
        }
    }

    void MCCompileUnit::gen_ro_data(std::ostream &out) {
        out << ".section\t"
            << ".rodata\n";
        for (const auto &item: float_nums) {
            out << ".align\t3\n";
            out << ".LC" << toUI32(static_cast<float>(item)) << ":\n";
            out << ".word\t" << toUI32(static_cast<float>(item)) << "\n";
        }
    }

    void MCCompileUnit::gen_asm(std::ostream &out) {
        gen_global_var(out);
        out << ".globl\t"
            << "main"
            << "\n";
        out << ".type\t"
            << "main"
            << ", @function\n";
        out << ".text\n";
        // 输出自定义函数
        if (genAtriBuiltin) {
            out << supportFunctions[0] << "\n";
            out << supportFunctions[1] << "\n";
        }
        for (const auto &func: *this) {
            func->gen_asm(out);
            //        out<<"stackTop:"<<func->stackTop<<",
            //        stackBottom:"<<func->stackBottom<<"\n";
        }
        //    gen_ro_data(out);
    }

    void MCCompileUnit::clear() {
        for (const auto &func: *this)
            func->clear();
        auto funcList = this->toVector();
        for (auto &func: funcList) {
            delete func;
        }
    }

    void MCCompileUnit::reg_alloc() {
        for (const auto &func: *this)
            func->reg_alloc();
    }

    void MCCompileUnit::foreOPT() {
        for (const auto &func: *this)
            func->foreOPT();
    }

    void MCCompileUnit::afterOPT() {
        for (const auto &func: *this)
            func->afterOPT();
    }

}// namespace riscv