//
// Created by 11067 on 2023/5/22.
//

#include "MCFunction.h"

namespace riscv {
MCFunction::MCFunction(MCCompileUnit *compileUnit, Ir::Function *ir_func) {
  this->compileUnit = compileUnit;
  label = ir_func->get_name();
  auto *b = new MCBlock(this, ".entry_" + label);
  this->entry = b;
  addLast(b);
  // 为每一个IR基本块分配后端IR基本块
  for (auto &irBlock : ir_func->get_bbs()) {
    auto pBlock =
        new MCBlock(this, ".L" + std::to_string(compileUnit->block_num++));
    map.irBB2MCBB[irBlock.get()] = pBlock;
    map.mcBB2IRBB[pBlock] = irBlock.get();
    addLast(pBlock);
    pBlock->loopDepth = irBlock->get_depth();
  }
  b->succ.push_back(map.irBB2MCBB[ir_func->get_first_bb()]);
  map.irBB2MCBB[ir_func->get_first_bb()]->prev.push_back(b);
  for (auto &irBlock : ir_func->get_bbs()) {
    auto *pBlock = map.irBB2MCBB[irBlock.get()];
    for (const auto &bb : irBlock->prev) {
      pBlock->prev.push_back(map.irBB2MCBB[bb]);
    }
    for (const auto &bb : irBlock->succ) {
      pBlock->succ.push_back(map.irBB2MCBB[bb]);
    }
  }
  // 移栈
  b->push_back(new PushStackInstr(this));
  // 保存参数
  int int_arg_cnt = 0, float_arg_cnt = 0;
  for (auto &arg : ir_func->get_params()) {
    Instanceof(reg, Ir::Reg *, arg.get()) {
      Reg dest = map.get_operand(*reg);
      if (reg->get_type() == Ir::Int) {
        if (int_arg_cnt < ARG_REG_NUM) {
          b->push_back(
              new MoveInstr(dest, get_reg_by_id(ARG_REGS[int_arg_cnt++])));
        } else {
          int offset = (int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
          b->push_back(
              new LoadStackInstr(dest, map.gen_const_operand(offset), this));
          int_arg_cnt++;
        }
      } else if (arg->get_type() == Ir::Float) {
        if (float_arg_cnt < ARG_REG_NUM) {
          b->push_back(
              new MoveInstr(dest, get_freg_by_id(ARG_REGS[float_arg_cnt++])));
        } else {
          int offset = (int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
          b->push_back(
              new LoadStackInstr(dest, map.gen_const_operand(offset), this));
          float_arg_cnt++;
        }
      } else {
        assert(false);
      }
    }
    Instanceof(vec, Ir::Vector *, arg.get()) {
      Reg dest = map.get_operand(*vec);
      if (int_arg_cnt < ARG_REG_NUM) {
        b->push_back(
            new MoveInstr(dest, get_reg_by_id(ARG_REGS[int_arg_cnt++])));
      } else {
        int offset = (int_arg_cnt + float_arg_cnt + 1) * INT_SIZE;
        b->push_back(
            new LoadStackInstr(dest, map.gen_const_operand(offset), this));
        int_arg_cnt++;
      }
      para_vec_base.insert({vec, dest});
    }
  }
  b->push_back(new GotoInstr(b->succ[0]));
  stackBottom += int_arg_cnt + float_arg_cnt;
  ra_stack_arr = stackBottom; // 栈中存放ra的位置
  stackBottom += 2;
  // 对每个块进行构造
  for (const auto &block : *this) {
    if (block != b) {
      block->construct(map.mcBB2IRBB[block], &map);
    }
  }
  destructPhi();
}

template <typename T> bool MCFunction::is_var(const T &op) {
  Instanceof(res, Reg *, op) { return !res->isConst; }
  return false;
}

void MCFunction::active_analyse() {
  // 获取use和def
  for (const auto &block : *this) {
    block->clear_active();
    for (const auto &code : *block) {
      Instanceof(call_inst, CallInstr *, code) {
        for (auto reg : call_inst->get_use()) {
          if (!block->activeDef.count(reg)) {
            block->activeUse.insert(reg);
          }
        }
        for (auto reg : call_inst->get_def()) {
          if (!block->activeUse.count(reg)) {
            block->activeDef.insert(reg);
          }
        }
      }
      else Instanceof(fm_inst, FMInstr *, code) {
        for (auto reg : fm_inst->get_use()) {
          if (!block->activeDef.count(reg)) {
            block->activeUse.insert(reg);
          }
        }
        for (auto reg : fm_inst->get_def()) {
          if (!block->activeUse.count(reg)) {
            block->activeDef.insert(reg);
          }
        }
      }
      else {
        if (is_var(code->get_op2())) {
          if (!block->activeDef.count(code->op2))
            block->activeUse.insert(code->op2);
        }
        if (is_var(code->get_op3())) {
          if (!block->activeDef.count(code->op3))
            block->activeUse.insert(code->op3);
        }
        if (is_var(code->get_op1())) {
          if (!block->activeUse.count(code->op1))
            block->activeDef.insert(code->op1);
        }
      }
    }
  }
  // 计算in和out
  bool inChanged = true;
  auto blocks = this->toVector();
  while (inChanged) {
    inChanged = false;
    for (int i = blocks.size() - 1; i >= 0; i--) {
      auto block = blocks[i];
      for (const auto &after : block->succ) {
        block->activeOut = set_union(block->activeOut, after->activeIn);
      }
      std::set<Reg> initialIn = block->activeIn;
      block->activeIn = set_union(
          block->activeUse, set_difference(block->activeOut, block->activeDef));
      if (!set_equal(initialIn, block->activeIn)) {
        inChanged = true;
      }
    }
  }
}

std::set<Reg> MCFunction::set_union(const std::set<Reg> &setA,
                                    const std::set<Reg> &setB) {
  std::set<Reg> res;
  std::set_union(setA.begin(), setA.end(), setB.begin(), setB.end(),
                 std::inserter(res, res.begin()));
  return res;
}

std::set<Reg> MCFunction::set_difference(const std::set<Reg> &setA,
                                         const std::set<Reg> &setB) {
  std::set<Reg> res;
  std::set_difference(setA.begin(), setA.end(), setB.begin(), setB.end(),
                      std::inserter(res, res.begin()));
  return res;
}

bool MCFunction::set_equal(const std::set<Reg> &setA,
                           const std::set<Reg> &setB) {
  return (setA.size() == setB.size()) &&
         (std::equal(setA.begin(), setA.end(), setB.begin(), setB.end()));
}

void MCFunction::gen_asm(std::ostream &out) {
  out << ".align 2\n";
//  out << ".globl\t" << label << "\n";
//  out << ".type\t" << label << ", @function\n";
  out << label << ":\n";
  for (const auto &block : *this)
    block->gen_asm(out);
}

void MCFunction::destructPhi() {
  for (auto &bb : this->toVector()) {
    std::unordered_map<MCBlock *, ParaCopyInstr *> bb2pc{};
    for (auto &prevBB : std::vector<MCBlock *>{bb->prev.begin(), bb->prev.end()}) {
      //        for (auto prevBB : bb->prev) {
      auto pc = new ParaCopyInstr();
      // 分裂
      if (bb->prev.size() > 1 && prevBB->succ.size() > 1) {
        auto *newBB = new MCBlock(
            this, ".Lspl" + std::to_string(compileUnit->block_num++));
        prevBB->succ.erase(
            std::remove(prevBB->succ.begin(), prevBB->succ.end(), bb),
            prevBB->succ.end());
        prevBB->succ.push_back(newBB);
        bb->prev.erase(std::remove(bb->prev.begin(), bb->prev.end(), prevBB),
                       bb->prev.end());
        bb->prev.push_back(newBB);
        replaceJumpTarget(prevBB, bb, newBB);
        replacePhiBlock(bb, prevBB, newBB);
        auto j = new GotoInstr(bb);
        newBB->push_back(pc);
        newBB->push_back(j);

        newBB->prev.push_back(prevBB);
        newBB->succ.push_back(bb);
        bb->insertBeforeMe(newBB);
        bb2pc.insert({newBB, pc});
      } else {
        auto last = prevBB->getLast();
        if (last == nullptr) {
          prevBB->push_back(pc);
        } else {
          while (last->getPrev() != nullptr &&
                 (dynamic_cast<IfConInstr *>(last->getPrev()) ||
                  dynamic_cast<GotoInstr *>(last->getPrev()) ||
                  dynamic_cast<RetureInstr *>(last->getPrev()) ||
                  dynamic_cast<CompareInstr *>(last->getPrev()))) {
            last = last->getPrev();
          }
          last->insertBeforeMe(pc);
        }
        bb2pc.insert({prevBB, pc});
      }
    }
    for (auto instr : bb->toVector()) {
      Instanceof(phi_instr, PhiInstr *, instr) {
        for (auto &[reg, block] : phi_instr->pairs) {
          bb2pc[block]->copy.emplace_back(phi_instr->op1, reg);
        }
        phi_instr->removeMe();
        delete phi_instr;
      }
    }
  }
  // 并行复制串行化
  for (auto bb : *this) {
    for (auto instr : bb->toVector()) {
      Instanceof(pCopy, ParaCopyInstr *, instr) {
        std::vector<MRInstr *> seq{};
        while (std::any_of(
            pCopy->copy.begin(), pCopy->copy.end(),
            [&](auto &pair) { return pair.first != pair.second; })) {
          // 没有其他second等于当前first
          auto s = std::find_if(
              pCopy->copy.begin(), pCopy->copy.end(), [&](auto &p) {
                return p.first != p.second &&
                       std::none_of(
                           pCopy->copy.begin(), pCopy->copy.end(),
                           [&](auto &p1) { return p1.second == p.first; });
              });
          if (s != pCopy->copy.end()) {
            auto &pair = *s;
            if (pair.second.isConst && pair.second.type == Ir::Float) {
              Reg var = map.gen_var_operand(true);
              Reg temp = map.gen_var_operand();
              double val = pair.second.float_num;
              auto bitVal = toUI32(static_cast<float>(val));

              seq.push_back(new MoveU32Instr(temp, bitVal));
              seq.push_back(new Bit2FInstr(var, temp));

              seq.push_back(new MoveInstr(pair.first, var));
            } else {
              seq.push_back(new MoveInstr(pair.first, pair.second));
            }
            pCopy->copy.erase(s);
          } else {
            auto pPair =
                std::find_if(pCopy->copy.begin(), pCopy->copy.end(),
                             [&](auto &p) { return p.first != p.second; });
            assert(pPair != pCopy->copy.end());
            auto &pair = *pPair;
            Reg op = map.gen_var_operand(pair.first.type == Ir::Float);
            if (pair.second.isConst && pair.second.type == Ir::Float) {
              Reg var = map.gen_var_operand(true);
              Reg temp = map.gen_var_operand();
              double val = pair.second.float_num;
              auto bitVal = toUI32(static_cast<float>(val));

              seq.push_back(new MoveU32Instr(temp, bitVal));
              seq.push_back(new Bit2FInstr(var, temp));

              seq.push_back(new MoveInstr(pair.first, var));
            } else {
              seq.push_back(new MoveInstr(op, pair.second));
            }
            pCopy->copy.erase(pPair);
            pCopy->copy.emplace_back(pair.first, op);
          }
        }
        for (auto &code : seq) {
          instr->insertBeforeMe(code);
        }
        instr->removeMe();
        delete instr;
      }
    }
    for (auto instr : bb->toVector()) {
      if (dynamic_cast<ParaCopyInstr *>(instr)) {
        panic("why paracopy here?");
      }
    }
  }
}

void MCFunction::replaceJumpTarget(MCBlock *inWhich, MCBlock *old,
                                   MCBlock *nnew) {
  for (auto instr : *inWhich) {
    if (auto b = dynamic_cast<IfConInstr *>(instr)) {
      auto target = b->target;
      if (target->label == old->label) {
        b->target = nnew;
      }
    } else if (auto g = dynamic_cast<GotoInstr *>(instr)) {
      auto target = g->target;
      if (target->label == old->label) {
        g->target = nnew;
      }
    }
  }
}

void MCFunction::replacePhiBlock(MCBlock *inWhich, MCBlock *old,
                                 MCBlock *nnew) {
  for (auto instr : *inWhich) {
    if (auto phi = dynamic_cast<PhiInstr *>(instr)) {
      decltype(phi->pairs) pairs{};
      for (auto [val, block] : phi->pairs) {
        if (block->label == old->label) {
          pairs.emplace_back(val, nnew);
        } else {
          pairs.emplace_back(val, block);
        }
      }
      phi->pairs = pairs;
    }
  }
}

void MCFunction::reg_alloc() {
  auto allocator = new MColorAllocator(this);
  allocator->run();
  auto int_list = allocator->get_int_alloc();
  auto float_list = allocator->get_float_alloc();
  std::map<int, int> regId2offset;
  // 生成callee保护 main函数不用
  if (this->label != "main") {
    for (auto reg : int_list) {
      regId2offset.insert({reg.id, stackTop});
      entry->getFirst()->insertAfterMe(new Store64Instr(
          reg, get_sp(), map.gen_const_operand(stackTop * INT_SIZE)));
      stackTop += 2;
    }
    for (auto reg : float_list) {
      regId2offset.insert({reg.id, stackTop});
      entry->getFirst()->insertAfterMe(new StoreFInstr(
          reg, get_sp(), map.gen_const_operand(stackTop * INT_SIZE)));
      stackTop += 1;
    }
  }
  int padding = 4 - (stackBottom + stackTop) % 4;
  ra_stack_arr = stackBottom + stackTop + padding - ra_stack_arr - 2;
  entry->getFirst()->insertAfterMe(new Store64Instr(
      get_ra(), get_sp(), map.gen_const_operand(ra_stack_arr * INT_SIZE)));

  // todo 带ret的block最后一句一定是return？
  for (auto block : exitBlock) {
    if (this->label != "main") {
      for (auto reg : int_list) {
        block->push_back(new Load64Instr(
            reg, get_sp(),
            map.gen_const_operand(regId2offset[reg.id] * INT_SIZE)));
      }
      for (auto reg : float_list) {
        block->push_back(new LoadFInstr(
            reg, get_sp(),
            map.gen_const_operand(regId2offset[reg.id] * INT_SIZE)));
      }
    }
    block->push_back(new Load64Instr(
        get_ra(), get_sp(), map.gen_const_operand(ra_stack_arr * INT_SIZE)));
    block->push_back(new PopStackInstr(this));
  }
  delete allocator;
}

void MCFunction::clear() {
  auto blockList = this->toVector();
  for (auto &block : blockList) {
    block->clear();
  }
  for (auto &block : blockList) {
    delete block;
  }
}

void MCFunction::foreOPT() {
  auto optimizer = new MOptimizer(this);
  optimizer->foreOPT();
  delete optimizer;
}

void MCFunction::afterOPT() {
  auto optimizer = new MOptimizer(this);
  optimizer->afterOPT();
  delete optimizer;
}

} // namespace riscv