//
// Created by MSI-NB on 2023/7/13.
//

#include "MColorAllocator.h"
namespace riscv {
MColorAllocator::MColorAllocator(MCFunction *_function) {
  this->function = _function;
  for (auto block : *function) {
    for (auto instr : *block) {
      for (auto reg : getDef(instr)) {
        if (is_pesudo(reg))
          initial.insert(reg);
      }
      for (auto reg : getUse(instr)) {
        if (is_pesudo(reg))
          initial.insert(reg);
      }
    }
  }
  for (auto reg : getAllocReg()) {
    preColored.insert(reg);
  }
}

void MColorAllocator::addEdge(int from_id, Reg to) {
  if (from_id == to.id)
    return;
  if (!adjList[from_id].count(to)) {
    adjList[from_id].insert(to);
    degree[from_id]++;
  }
}

void MColorAllocator::run() {
  auto temp = spilledNodes;
  while (true) {
    info("allocator running");
    function->active_analyse();
    calculateSpillPriority();
    clear();
    build();
    makeWorkList();
    do {
      if (!simplifyWorkList.empty()) {
        simplify();
      } else if (!workListMoves.empty()) {
        coalesce();
      } else if (!freezeWorkList.empty()) {
        freeze();
      } else if (!spillWorkList.empty()) {
        selectSpill();
      }
    } while (!(simplifyWorkList.empty() && workListMoves.empty() &&
               freezeWorkList.empty() && spillWorkList.empty()));
    assignColors();
    if (!spilledNodes.empty()) {
      rewriteProgram();
    } else {
      break;
    }
    run_time++;
  }
  finish();
}

// todo 优化
void MColorAllocator::calculateSpillPriority() {
  spillPriority.clear();
  spillPriority.resize(function->map.reg_num, 0.);
  for (auto block : *function) {
    for (auto instr : *block) {
      for (auto reg : getDef(instr)) {
        spillPriority[reg.id] += pow(SpillParam, block->loopDepth);
      }
      for (auto reg : getUse(instr)) {
        spillPriority[reg.id] += pow(SpillParam, block->loopDepth);
      }
    }
  }
}

void MColorAllocator::clear() {
  simplifyWorkList.clear();
  freezeWorkList.clear();
  spillWorkList.clear();
  spilledNodes.clear();
  coalescedNodes.clear();
  coloredNodes.clear();
  selectStack.clear();
  selectSet.clear();
  coalescesMoves.clear();
  constrainedMoves.clear();
  frozenMoves.clear();
  workListMoves.clear();
  activeMoves.clear();
  adjList.clear();
  adjList.resize(getAllRegCnt());
  degree.clear();
  degree.resize(getAllRegCnt());
  alias.clear();
  alias.resize(getAllRegCnt());
  moveList.clear();
  moveList.resize(getAllRegCnt());
  colors.clear();
  colors.resize(getAllRegCnt());
  for (int i = 0; i < PREG_NUM; ++i) {
    colors[i] = get_reg_by_id(i);
    degree[i] = INT32_MAX;
    alias[i] = Reg{};
    moveList[i].clear();
  }
  for (int i = PREG_NUM; i < getAllRegCnt(); ++i) {
    colors[i] = Reg{};
    degree[i] = 0;
    alias[i] = Reg{};
    moveList[i].clear();
  }
}

void MColorAllocator::build() {
  auto vRegType = function->map.vRegType;
  for (int i = 0; i < vRegType.size(); ++i) {
    // float vReg与所有通用冲突
    if (vRegType[i] == Ir::Float) {
      for (int j = 0; j < FLOAT_REG_BEGIN; ++j) {
        addEdge(i + PREG_NUM, get_reg_by_id(j));
      }
    }
    // int vReg与所有浮点寄存器冲突
    if (vRegType[i] == Ir::Int) {
      for (int j = FLOAT_REG_BEGIN; j < PREG_NUM; ++j) {
        addEdge(i + PREG_NUM, get_reg_by_id(j));
      }
    }
    auto spRegs = getSPReg();
    for (auto reg : spRegs) {
      addEdge(i + PREG_NUM, reg);
    }
  }
  for (auto block : *function) {
    unordered_set<Reg> live{};
    for (auto reg : block->activeOut) {
      live.insert(reg);
    }
    auto instrList = block->toVector();

    for (int i = instrList.size() - 1; i >= 0; i--) {
      auto &instr = instrList[i];
      bool isFloat = instr->op1.type == Ir::Float;
      Instanceof(move_instr, MoveInstr *, instr) {
        if (!move_instr->op1.isConst && !move_instr->op2.isConst &&
            move_instr->op1 != move_instr->op2) {
          auto spRegs = getSPReg();
          if (std::find(spRegs.begin(), spRegs.end(), move_instr->op1) ==
                  spRegs.end() &&
              std::find(spRegs.begin(), spRegs.end(), move_instr->op2) ==
                  spRegs.end()) {
            auto mov = std::make_pair(instr->op2, instr->op1);
            for (auto reg : getUse(instr)) {
              live.erase(reg);
              moveList[reg.id].insert(mov);
            }
            for (auto reg : getDef(instr)) {
              moveList[reg.id].insert(mov);
            }
            addWorkListMove(mov);
          }
        }
      }
      //            }
      for (auto reg : getDef(instr)) {
        live.insert(reg);
      }
      //            for (auto reg : getSPReg()) {
      //                live.erase(reg);
      //            }
      // debug 输出
      //            instr->gen_asm(std::cout);
      //            for (auto reg : live) {
      //                std::cout << reg << ", ";
      //            }
      //
      //            std::cout << "def:";
      //            for (auto reg : getDef(instr)) {
      //                std::cout << reg << ", ";
      //            }
      //            std::cout << "\n";

      for (auto reg : getDef(instr)) {
        for (auto reg1 : live) {
          if (is_pesudo(reg1) && reg1.type == reg.type) {
            addEdge(reg1.id, reg);
          }
          if (is_pesudo(reg) && reg1.type == reg.type) {
            addEdge(reg.id, reg1);
          }
        }
      }
      // 所有跨过程调用活跃的变量与临时寄存器冲突 caller saved
      Instanceof(call_inst, CallInstr *, instr) {
        for (auto &reg : live) {
          if (is_pesudo(reg)) {
            for (auto &reg1 : getTempReg()) {
              addEdge(reg.id, reg1);
            }
          }
        }
      }

      for (auto reg : getDef(instr)) {
        live.erase(reg);
      }
      for (auto reg : getUse(instr)) {
        live.insert(reg);
      }
    }
  }
}

void MColorAllocator::makeWorkList() {
  for (auto reg : initial) {
    if (degree[reg.id] >= PREG_ALLOC_NUM) {
      spillWorkList.insert(reg);
    } else if (moveRelated(reg)) {
      freezeWorkList.insert(reg);
    } else {
      simplifyWorkList.push_back(reg);
    }
  }
  initial.clear();
}

void MColorAllocator::simplify() {
  auto &reg = simplifyWorkList.back();
  simplifyWorkList.pop_back();
  selectStack.push_back(reg);
  selectSet.insert(reg);
  for (auto reg1 : adjacent(reg)) {
    if (selectSet.count(reg) == 0 && coalescedNodes.count(reg) == 0)
      decrementDegree(reg1);
  }
}

void MColorAllocator::decrementDegree(Reg &reg) {
  auto d = degree[reg.id];
  degree[reg.id] = d - 1;
  if (d == PREG_ALLOC_NUM) {
    auto t = adjacent(reg);
    t.insert(reg);
    enableMoves(t);
    spillWorkList.erase(reg);
    if (moveRelated(reg)) {
      freezeWorkList.insert(reg);
    } else {
      simplifyWorkList.push_back(reg);
    }
  }
}

unordered_set<Reg> MColorAllocator::adjacent(Reg &reg) {
  auto ret = unordered_set<Reg>{adjList[reg.id]};
  for (auto e : selectSet) {
    ret.erase(e);
  }
  for (auto e : coalescedNodes) {
    ret.erase(e);
  }
  return ret;
}

void MColorAllocator::enableMoves(unordered_set<Reg> &nodes) {
  for (auto reg : nodes) {
    for (auto mov : nodeMoves(reg)) {
      if (activeMoves.count(mov)) {
        activeMoves.erase(mov);
        addWorkListMove(mov);
        //                workListMoves.insert(reg1);
      }
    }
  }
}

void MColorAllocator::coalesce() {
  //    auto m = *workListMoves.begin();
  auto m = workListMoves.front();
  workListMoves.erase(workListMoves.begin());

  auto x = getAlias(m.second);
  auto y = getAlias(m.first);
  Reg u, v;
  if (preColored.count(y)) {
    u = y;
    v = x;
  } else {
    u = x;
    v = y;
  }
  //    workListMoves.erase(m);
  auto adjV = adjacent(v);
  // todo 可优化
  bool cond;
  if (!preColored.count(u)) {
    auto temp = adjacent(u);
    for (auto reg : adjacent(v)) {
      temp.insert(reg);
    }
    cond = conservative(temp);
  }

  if (u == v) {
    coalescesMoves.insert(m);
    addWorkList(u);
  } else if (preColored.count(v) || adjList[u.id].count(v) ||
             adjList[v.id].count(u)) {
    constrainedMoves.insert(m);
    addWorkList(u);
    addWorkList(v);
  } else if ((preColored.count(u) &&
              std::all_of(adjV.begin(), adjV.end(),
                          [&](const auto &t) { return OK(t, u); })) ||
             (!preColored.count(u) && cond)) {
    coalescesMoves.insert(m);
    combine(u, v);
    addWorkList(u);
  } else {
    activeMoves.insert(m);
  }
}

unordered_set<Reg> MColorAllocator::set_union(unordered_set<Reg> &a,
                                              unordered_set<Reg> &b) {
  unordered_set<Reg> ret = {b};
  for (auto reg : a) {
    ret.insert(reg);
  }
  return ret;
}

bool MColorAllocator::OK(Reg t, Reg r) {
  return degree[t.id] < PREG_ALLOC_NUM || preColored.count(t) ||
         adjList[t.id].count(r) || adjList[r.id].count(t);
}

bool MColorAllocator::conservative(unordered_set<Reg> &nodes) {
  int k = 0;
  for (auto reg : nodes) {
    if (degree[reg.id] >= PREG_ALLOC_NUM) {
      k += 1;
    }
  }
  return k < PREG_ALLOC_NUM;
}

void MColorAllocator::combine(Reg u, Reg v) {
  if (freezeWorkList.count(v)) {
    freezeWorkList.erase(v);
  } else {
    spillWorkList.erase(v);
  }
  coalescedNodes.insert(v);
  alias[v.id] = u;
  for (auto move : moveList[v.id]) {
    moveList[u.id].insert(move);
  }
  unordered_set<Reg> temp_set{};
  temp_set.insert(v);
  enableMoves(temp_set);
  for (auto reg : adjacent(v)) {
    if (is_pesudo(reg)) {
      addEdge(reg.id, u);
    }
    if (is_pesudo(u)) {
      addEdge(u.id, reg);
    }
    decrementDegree(reg);
  }
  if (degree[u.id] >= PREG_ALLOC_NUM && freezeWorkList.count(u)) {
    freezeWorkList.erase(u);
    spillWorkList.insert(u);
  }
}

void MColorAllocator::freeze() {
  auto u = *freezeWorkList.begin();
  freezeWorkList.erase(u);
  simplifyWorkList.push_back(u);
  freezeMoves(u);
}

void MColorAllocator::freezeMoves(const Reg &u) {
  for (auto m : nodeMoves(u)) {
    auto x = m.second;
    auto y = m.first;
    Reg v;
    if (getAlias(y) == getAlias(u)) {
      v = getAlias(x);
    } else {
      v = getAlias(y);
    }
    activeMoves.erase(m);
    frozenMoves.insert(m);
    if (nodeMoves(v).empty() && degree[v.id] < PREG_ALLOC_NUM) {
      freezeWorkList.erase(v);
      simplifyWorkList.push_back(v);
    }
  }
}

void MColorAllocator::selectSpill() {
  vector<Reg> spillList, creatSpillList;
  for (auto reg : spillWorkList) {
    if (!spillCreated.count(reg)) {
      spillList.push_back(reg);
    } else {
      creatSpillList.push_back(reg);
    }
    //        std::cout<<reg.id<<" ";
  }
  //    std::cout<<"\n";

  std::sort(spillList.begin(), spillList.end(),
            [&](const auto &l, const auto &r) {
              return spillPriority[l.id] / degree[l.id] <
                     spillPriority[r.id] / degree[r.id];
            });
  std::sort(creatSpillList.begin(), creatSpillList.end(),
            [&](const auto &l, const auto &r) {
              return spillPriority[l.id] / degree[l.id] <
                     spillPriority[r.id] / degree[r.id];
            });
  Reg regToSpill;
  if (!spillList.empty()) {
    regToSpill = *spillList.begin();
  } else {
    regToSpill = *creatSpillList.begin();
  }
  if (!is_pesudo(regToSpill)) {
    panic("???");
  }
  spillWorkList.erase(regToSpill);
  simplifyWorkList.push_back(regToSpill);
  freezeMoves(regToSpill);
}

void MColorAllocator::assignColors() {
  while (!selectStack.empty()) {
    auto reg = selectStack.back();
    selectStack.pop_back();
    unordered_set<Reg> okColors{};
    for (auto reg_item : getAllocReg()) {
      okColors.insert(reg_item);
    }
    for (auto &reg1 : adjList[reg.id]) {
      auto regAlias = getAlias(reg1);
      if (coloredNodes.count(regAlias) || preColored.count(regAlias)) {
        okColors.erase(colors[regAlias.id]);
      }
    }
    if (okColors.empty()) {
      spilledNodes.insert(reg);
    } else {
      bool suss = false;
      auto caller_reg = getTempReg();
      for (auto creg : caller_reg) {
        if (okColors.count(creg)) {
          colors[reg.id] = creg;
          suss = true;
          break;
        }
      }
      if (!suss) {
        auto c = *okColors.begin();
        colors[reg.id] = c;
      }
      coloredNodes.insert(reg);
    }
  }
  for (auto &reg : coalescedNodes) {
    colors[reg.id] = colors[getAlias(reg).id];
  }
}

void MColorAllocator::rewriteProgram() {
  unordered_set<Reg> newTemp{};
  for (auto &r : spilledNodes) {
    if (!is_pesudo(r))
      panic("bad !!!");
    auto vReg = r;
    auto slotOffset = (function->stackTop) * INT_SIZE;
    if (r.is_64) {
      function->stackTop += 2;
      spilledVarCnt += 2;
    } else {
      spilledVarCnt++;
      function->stackTop++;
    }
    for (auto block : *function) {
      for (auto instr : block->toVector()) {
        auto use = getUse(instr);
        auto def = getDef(instr);
        auto usedHere = std::count(use.begin(), use.end(), r);
        auto defHere = std::count(def.begin(), def.end(), r);
        if (usedHere || defHere) {
          Instanceof(move_inst, MoveInstr *, instr) {
            move_inst->is_spilled = true;
          }
          Reg tReg = allocSplitVReg(vReg.type);
          newTemp.insert(tReg);
          spillCreated.insert(tReg);
          if (usedHere) {
            Reg offset = function->map.gen_const_operand(slotOffset);
            if (vReg.type == Ir::Int) {
              if (r.is_64) {
                instr->insertBeforeMe(new Load64Instr(tReg, get_sp(), offset));
              } else {
                instr->insertBeforeMe(new LoadInstr(tReg, get_sp(), offset));
              }
            } else {
              instr->insertBeforeMe(new LoadFInstr(tReg, get_sp(), offset));
            }
          }
          if (defHere) {
            Reg offset = function->map.gen_const_operand(slotOffset);
            if (vReg.type == Ir::Int) {
              if (r.is_64) {
                instr->insertAfterMe(new Store64Instr(tReg, get_sp(), offset));
              } else {
                instr->insertAfterMe(new StoreInstr(tReg, get_sp(), offset));
              }
            } else {
              instr->insertAfterMe(new StoreFInstr(tReg, get_sp(), offset));
            }
          }
          allocReg(instr, vReg, tReg);
        }
      }
    }
  }
  spilledNodes.clear();
  initial.clear();
  initial.insert(coloredNodes.begin(), coloredNodes.end());
  initial.insert(coalescedNodes.begin(), coalescedNodes.end());
  initial.insert(newTemp.begin(), newTemp.end());
  coloredNodes.clear();
  coalescedNodes.clear();
}

void MColorAllocator::finish() {
  for (auto bb : *function) {
    for (auto instr : bb->toVector()) {
      for (auto reg : getDef(instr)) {
        if (is_pesudo(reg)) {
          allocReg(instr, reg, colors[reg.id]);
        }
      }
      for (auto reg : getUse(instr)) {
        if (is_pesudo(reg)) {
          allocReg(instr, reg, colors[reg.id]);
        }
      }
      Instanceof(moveInstr, MoveInstr *, instr) {
        if (moveInstr->op1 == moveInstr->op2) {
          moveInstr->removeMe();
          delete moveInstr;
        }
      }
    }
  }
}

Reg MColorAllocator::allocSplitVReg(Ir::ScalaType _type) {
  return function->map.gen_var_operand(_type == Ir::Float);
}

Reg MColorAllocator::getAlias(const Reg &reg) {
  if (coalescedNodes.count(reg)) {
    return getAlias(alias[reg.id]);
  } else {
    return reg;
  }
}

bool MColorAllocator::moveRelated(const Reg &reg) {
  return !nodeMoves(reg).empty();
}

unordered_set<MMove> MColorAllocator::nodeMoves(const Reg &reg) {
  auto ret = unordered_set<MMove>{};
  for (auto move : moveList[reg.id]) {
    if (activeMoves.count(move) ||
        std::find(workListMoves.begin(), workListMoves.end(), move) !=
            workListMoves.end()) {
      ret.insert(move);
    }
  }
  return ret;
}

int MColorAllocator::getAllRegCnt() { return function->map.reg_num; }

vector<Reg> MColorAllocator::getDef(MRInstr *instr) {
  vector<Reg> ret;
  Instanceof(call_inst, CallInstr *, instr) { ret = call_inst->get_def(); }
  else Instanceof(fm_inst, FMInstr *, instr) {
    ret = fm_inst->get_def();
  }
  else {
    if (!instr->op1.isConst) {
      ret.push_back(instr->op1);
    }
  }
  return ret;
}

vector<Reg> MColorAllocator::getUse(MRInstr *instr) {
  vector<Reg> ret;
  Instanceof(call_inst, CallInstr *, instr) { ret = call_inst->get_use(); }
  else Instanceof(fm_inst, FMInstr *, instr) {
    ret = fm_inst->get_use();
  }
  else {
    if (!instr->op2.isConst) {
      ret.push_back(instr->op2);
    }
    if (!instr->op3.isConst) {
      ret.push_back(instr->op3);
    }
  }
  return ret;
}

void MColorAllocator::allocReg(MRInstr *instr, Reg &vreg, Reg &preg) {
  if (!instr->op1.isConst && instr->op1 == vreg) {
    instr->op1 = preg;
  }
  if (!instr->op2.isConst && instr->op2 == vreg) {
    instr->op2 = preg;
  }
  if (!instr->op3.isConst && instr->op3 == vreg) {
    instr->op3 = preg;
  }
  Instanceof(fm_inst, FMInstr *, instr) {
    if (!fm_inst->op4.isConst && fm_inst->op4 == vreg) {
      fm_inst->op4 = preg;
    }
  }
  if (!is_pesudo(preg) && preg.regAttribute == RegAttribute::CalleeSaved) {
    if (preg.type == Ir::Int) {
      int_alloced.insert(preg);
    } else if (preg.type == Ir::Float) {
      float_alloced.insert(preg);
    }
  }
}

void MColorAllocator::addWorkList(const Reg &u) {
  if (!preColored.count(u) && is_pesudo(u) && !moveRelated(u) &&
      degree[u.id] < PREG_ALLOC_NUM) {
    freezeWorkList.erase(u);
    simplifyWorkList.push_back(u);
  }
}

void MColorAllocator::addWorkListMove(MMove move) {
  if (std::find(workListMoves.begin(), workListMoves.end(), move) ==
      workListMoves.end()) {
    workListMoves.push_back(move);
  }
}

} // namespace riscv