//
// Created by 11067 on 2023/5/18.
//

#include "arch.h"
#include <cassert>
#include <set>

namespace riscv {

static Reg regs[]{
    Reg{0, "zero", RegAttribute::NA},
    Reg{1, "ra", RegAttribute::CallerSaved},
    Reg{2, "sp", RegAttribute::CalleeSaved},
    Reg{3, "gp", RegAttribute::NA},
    Reg{4, "tp", RegAttribute::NA},
    Reg{5, "t0", RegAttribute::CallerSaved},
    Reg{6, "t1", RegAttribute::CallerSaved},
    Reg{7, "t2", RegAttribute::CallerSaved},
    Reg{8, "s0" /*also fp*/, RegAttribute::CalleeSaved},
    Reg{9, "s1", RegAttribute::CalleeSaved},
    Reg{10, "a0", RegAttribute::CallerSaved},
    Reg{11, "a1", RegAttribute::CallerSaved},
    Reg{12, "a2", RegAttribute::CallerSaved},
    Reg{13, "a3", RegAttribute::CallerSaved},
    Reg{14, "a4", RegAttribute::CallerSaved},
    Reg{15, "a5", RegAttribute::CallerSaved},
    Reg{16, "a6", RegAttribute::CallerSaved},
    Reg{17, "a7", RegAttribute::CallerSaved},
    Reg{18, "s2", RegAttribute::CalleeSaved},
    Reg{19, "s3", RegAttribute::CalleeSaved},
    Reg{20, "s4", RegAttribute::CalleeSaved},
    Reg{21, "s5", RegAttribute::CalleeSaved},
    Reg{22, "s6", RegAttribute::CalleeSaved},
    Reg{23, "s7", RegAttribute::CalleeSaved},
    Reg{24, "s8", RegAttribute::CalleeSaved},
    Reg{25, "s9", RegAttribute::CalleeSaved},
    Reg{26, "s10", RegAttribute::CalleeSaved},
    Reg{27, "s11", RegAttribute::CalleeSaved},
    Reg{28, "t3", RegAttribute::CallerSaved},
    Reg{29, "t4", RegAttribute::CallerSaved},
    Reg{30, "t5", RegAttribute::CallerSaved},
    Reg{31, "t6", RegAttribute::CallerSaved},
    Reg{32, "ft0", RegAttribute::CallerSaved},
    Reg{33, "ft1", RegAttribute::CallerSaved},
    Reg{34, "ft2", RegAttribute::CallerSaved},
    Reg{35, "ft3", RegAttribute::CallerSaved},
    Reg{36, "ft4", RegAttribute::CallerSaved},
    Reg{37, "ft5", RegAttribute::CallerSaved},
    Reg{38, "ft6", RegAttribute::CallerSaved},
    Reg{39, "ft7", RegAttribute::CallerSaved},
    Reg{40, "fs0", RegAttribute::CalleeSaved},
    Reg{41, "fs1", RegAttribute::CalleeSaved},
    Reg{42, "fa0", RegAttribute::CallerSaved},
    Reg{43, "fa1", RegAttribute::CallerSaved},
    Reg{44, "fa2", RegAttribute::CallerSaved},
    Reg{45, "fa3", RegAttribute::CallerSaved},
    Reg{46, "fa4", RegAttribute::CallerSaved},
    Reg{47, "fa5", RegAttribute::CallerSaved},
    Reg{48, "fa6", RegAttribute::CallerSaved},
    Reg{49, "fa7", RegAttribute::CallerSaved},
    Reg{50, "fs2", RegAttribute::CalleeSaved},
    Reg{51, "fs3", RegAttribute::CalleeSaved},
    Reg{52, "fs4", RegAttribute::CalleeSaved},
    Reg{53, "fs5", RegAttribute::CalleeSaved},
    Reg{54, "fs6", RegAttribute::CalleeSaved},
    Reg{55, "fs7", RegAttribute::CalleeSaved},
    Reg{56, "fs8", RegAttribute::CalleeSaved},
    Reg{57, "fs9", RegAttribute::CalleeSaved},
    Reg{58, "fs10", RegAttribute::CalleeSaved},
    Reg{59, "fs11", RegAttribute::CalleeSaved},
    Reg{60, "ft8", RegAttribute::CallerSaved},
    Reg{61, "ft9", RegAttribute::CallerSaved},
    Reg{62, "ft10", RegAttribute::CallerSaved},
    Reg{63, "ft11", RegAttribute::CallerSaved},
    // Reg{-1, "pc", RegAttribute::NA,},// RISCV有除跳转外修改PC的指令
};

Reg::Reg(int id, bool isFloat) {
    this->id = id;
    this->isConst = false;
    this->type = isFloat ? Ir::ScalaType::Float : Ir::ScalaType::Int;
    this->abiName = ("vir" + std::to_string(id)).c_str();
    this->regAttribute = RegAttribute::VIR;
}

Reg::Reg(int num) {
    this->id = -1;
    this->isConst = true;
    this->int_num = num;
    this->regAttribute = RegAttribute::NA;
    this->type = Ir::ScalaType::Int;
}

Reg &get_reg_by_id(int i) {
    // assert i in [0,31]
    return regs[i];
}

Reg &get_freg_by_id(int i) {
    return regs[32 + i];
}

Reg &get_sp() {
    return regs[2];
}

Reg &get_ra() {
    return regs[1];
}

Reg &get_temp() {
    return regs[8];
}

Reg &get_zero() {
    return regs[0];
}

Reg &get_ret(bool isFloat) {
    if (isFloat) {
        return get_freg_by_id(10);
    } else {
        return get_reg_by_id(10);
    }
}

bool isImm12(int num){
    return !(num>=1<<11||num<-(1<<11));
}

std::ostream &operator<<(std::ostream &os, const Reg &reg) {
    int id = reg.id;
    //    if (0 <= id && id < 32) {
    //        os << std::string("x") + std::to_string(id);
    //    } else if (32 <= id && id < 64) {
    //        os << std::string("f") + std::to_string(id - 32);
    //    }
    if (0 <= id && id < 64) {
        os << reg.abiName;
    } else if (64 <= id) {
        os << std::string("vir") + std::to_string(id - 64);
    } else {
        //        assert(false);
        os << std::string{"error!!"};
    }
    return os;
}

bool is_pesudo(Reg reg) {
    return reg.id >= PREG_NUM;
}

std::vector<Reg> getSPReg() {
    std::vector<Reg> ret{};
    ret.push_back(regs[0]);
    ret.push_back(regs[1]);
    ret.push_back(regs[2]);
    ret.push_back(regs[3]);
    ret.push_back(regs[4]);
    ret.push_back(regs[8]);//职业temp
    return ret;
}

std::vector<Reg> getTempReg() {
    std::vector<Reg> ret{};
    for (int i = 5; i < 8; ++i) {
        ret.push_back(regs[i]);
    }
    for (int i = 10; i < 18; ++i) {
        ret.push_back(regs[i]);
    }
    for (int i = 28; i < 40; ++i) {
        ret.push_back(regs[i]);
    }
    for (int i = 42; i < 50; ++i) {
        ret.push_back(regs[i]);
    }
    for (int i = 60; i < 64; ++i) {
        ret.push_back(regs[i]);
    }
    return ret;
}

std::vector<Reg> getAllocReg() {
    std::vector<Reg> ret{};
    std::set<int> spId = {0, 1, 2, 3, 4, 8};
    for (auto reg : regs) {
        if (!spId.count(reg.id)) {
            ret.push_back(reg);
        }
    }
    return ret;
}
bool Reg::operator==(const Reg &rhs) const {
    return id == rhs.id;
}

bool Reg::operator!=(const Reg &rhs) const {
    return !(rhs == *this);
}

bool Reg::operator<(const Reg &rhs) const {
    return id < rhs.id;
}

bool Reg::operator>(const Reg &rhs) const {
    return id > rhs.id;
}

} // namespace riscv