//
// Created by shliba on 2023/5/21.
//
#include "iostream"
#include "antlr4-runtime.h"

#include "ast/SysYLexer.h"
#include "ast/SysYParser.h"
#include "ast/ast_visitor.h"

#include "utils/logger.h"
#include "optimization/opt.h"
#include "backend/arm/ACodeGen.h"
#include "backend/arm/pass/ColorAllocator.h"
#include "backend/arm/pass/BakaAllocator.h"
#include "backend/arm/pass/AfterCodegenPass.h"
#include "backend/riscv/MCCompileUnit.h"
#include "optimization/ir.h"

// 使用WIN上MSVC的链接器，开栈
#pragma comment(linker, "/STACK:1073741824")
using namespace std;
using namespace antlr4;

// xxx.exe -o <output_path> <input_path> [-O2] [--ir] [--llvm]
// -O2 开启优化
// --ir 输出中间代码
// --llvm 输出llvm
// --arm 输出 arm
// --riscv 输出riscv

const std::string DefaultInputStream = "input.txt";
const std::string DefaultOutputStream = "output.txt";
const std::string O2Pattern = "-O1";
const std::string IRPattern = "--ir";
const std::string LLVMPattern = "--llvm";
const std::string ArmPattern = "--arm";
const std::string RISCPattern = "--riscv";


istream &get_input(int argc, char *argv[]) {
    static ifstream _ifs;
    for (int i = 1; i < argc; i++) {
        auto sv = std::string_view(argv[i]);
        if (sv == "-o") { // 接下来是输出的文件名,所以跳过
            i += 1;
        } else if (sv.front() == '-') {
            continue;
        } else {
            _ifs.open(sv.data());
            atriAssert(_ifs.is_open());
            return _ifs;
        }
    }
    _ifs.open(DefaultInputStream);
    return _ifs;
}

ostream &get_output(int argc, char *argv[]) {
    static ofstream _ofs;
    for (int i = 0; i < argc; i++) {
        auto sv = std::string_view(argv[i]);
        if (sv == "-o") {
            auto outputPath = std::string_view(argv[i + 1]);
            _ofs.open(outputPath.data());
            atriAssert(_ofs.is_open());
            return _ofs;
        }
    }
    _ofs.open(DefaultOutputStream);
    return _ofs;
}

bool has_option(int argc, char *argv[], std::string_view option) {
    for (int i = 0; i < argc; i++) {
        auto sv = std::string_view(argv[i]);
        if (sv == option) {
            return true;
        }
    }
    return false;
}

int main(int argc, char *argv[]) {
    //    std::cout << "hello, world!" << std::endl;
    //    info("main start");
    //    warn("warn is colorful");
    //    error("also error");
    //    info("1",1,4,5,1,"4");
    // get input and output path

    auto &is = get_input(argc, argv);
    auto &os = get_output(argc, argv);
    //    std::string buf;
    //    while (!is.eof()) {
    //        is >> buf;
    //        std::cout << buf;
    //    }
    ANTLRInputStream input{is};
    SysYLexer lexer{&input};
    CommonTokenStream tokenStream{&lexer};
    SysYParser parser{&tokenStream};

    if (has_option(argc, argv, ArmPattern)) {
        Ir::CompileUnit::in_arm = true;
    }

    // transform antlr4 ast to our ast
    // todo

    auto root = parser.compUnit();

    auto visitor = new AstVisitor;
    auto cu = std::any_cast<Ir::CompileUnit *>(visitor->visitCompUnit(root));
    //info("1",1,4,5,1,"4");

;



//    cu->reset_num();
//    info(*cu);
//    os << *cu;

    //  std::cout << root->toStringTree(&parser);

    // IR generator
    // todo

    // get compile option and output
    for (auto &func : cu->get_functions()) {
        auto cfg = new opt::CFG(func.get());
        cfg->build();
        func->cfg = cfg;
    }
    opt::convert_pass(cu);
    opt::tail_recur(cu);
    opt::useless_func_clean(cu);
    opt::global_var_pass(cu);
    opt::mem2reg(cu);
    //opt::run_opt(cu);
    auto optSwitch{true};

    if (has_option(argc, argv, O2Pattern)) {
        opt::run_opt(cu);
        optSwitch = true;
        // todo
    }
    cu->reset_num();
    //info(*cu);





//        info(*cu);
        info("generating riscv assemble...");
        auto RiscvComp = riscv::MCCompileUnit(*cu);

//        static ofstream _ofs;
//        _ofs.open(DefaultOutputStream);
//        RiscvComp.gen_asm(_ofs);
        RiscvComp.foreOPT();
        RiscvComp.reg_alloc();
        RiscvComp.afterOPT();
        RiscvComp.gen_asm(os);
        RiscvComp.clear();
        return 0;



    return 0;
}