//
// Created by shliba on 2023/5/21.
//
#ifndef COMPILER_IR_H
#define COMPILER_IR_H
#include <memory>
#include <ostream>
#include "list"
#include "unordered_set"
#include "unordered_map"
#include "string"
#include "vector"
#include "../utils/utils.h"
#include "cassert"
#include <iostream>
#include <iomanip>
#include <cstring>
#include "../utils/logger.h"

namespace opt{
class CFG;
}

namespace Ir {
using std::shared_ptr;
using std::unordered_set;
using std::unordered_map;
enum ScalaType {
    Void,
    Bool,
    Int,
    Float,
};

class Type{
    int id = -1;
    ScalaType type;
public:
    virtual ScalaType get_type() const;
    int get_id() const;
    void set_id(int id);
    Type(int id, ScalaType type);
    virtual void set_type(ScalaType type);

};

class Function;
class Loop;

class MemObject: public Type{
    //


public:

    MemObject(ScalaType type, int id);

};

union RegConst {
    int int_num;
    float float_num = 0;
};

class  Reg: public Type{


    bool constant = false;    //是否为常数
    bool param = false;    //是否为函数参数
    bool global = false;       //是否为全局变量
    std::string global_name; //只有全局有用
    RegConst num;
public:
    friend std::ostream &operator<<(std::ostream &os, const Reg &reg);
    bool operator==(const Reg &rhs) const;
    bool operator!=(const Reg &rhs) const;
    bool is_constant() const;
    const std::string &get_global_name() const;
    void set_global_name(const std::string &globalName);
    void set_constant(bool constant);

    bool is_param() const;

    void set_param(bool param);

    bool is_global() const;

    void set_global(bool global);

    RegConst &get_num();

    void set_num(const RegConst &num);


    Reg(ScalaType type, int id);

};

/**
 * 向量类型，注意和std的vector区分
 * dims是数组维数，代表数组每一维的大小
 */
class Vector: public MemObject{

    std::string name = "";
    bool constant = false;
    bool param = false;
    std::vector<int> dims;

public:
    friend std::ostream &operator<<(std::ostream &os, const Vector &vector);

public:
    Vector(ScalaType type, std::string name, bool constant);
    bool is_constant() const;
    Vector(ScalaType type, int id);
    bool is_param() const;
    void set_param(bool param);
    void set_dims(const std::vector<int> &dims);
    const std::string &get_name() const;
    std::vector<int> &get_dims();
};
class BasicBlock;

/**
 * 中端指令类，通过Instanceof判断指令类型
 * 虚方法getuses获得所有指令使用的寄存器
 */
class Instruction {
    BasicBlock * belongto_bb;
public:
    virtual unordered_set<shared_ptr<Type>> get_uses()const = 0;
    virtual void gen_ir(std::ostream &os) const;
    BasicBlock *get_belongto_bb() const;
    void set_belongto_bb(BasicBlock *belongtoBb);
    virtual void add_use_def() = 0;
    virtual void remove_use_def() = 0;
    virtual void change_use_name(int src_id,shared_ptr<Reg> new_val) {
        info(123);
    };
    virtual void change_use_name(int src_id,shared_ptr<MemObject> new_val){};
    virtual void change_use_name(int src_id,shared_ptr<Vector> new_val){};
    virtual void change_use_name(shared_ptr<Reg> src,shared_ptr<Reg> new_val) {};
    virtual void change_use_name(shared_ptr<MemObject> src,shared_ptr<MemObject> new_val) {};
    virtual void change_use_name(shared_ptr<Vector> src,shared_ptr<Vector> new_val) {};
    friend std::ostream &operator<<(std::ostream &os, const Instruction &instruction);
    virtual Instruction* clone(std::unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock*, Ir::BasicBlock*>&) = 0;
};

class DefInstr: public Instruction {
public:
    virtual shared_ptr<Type> def()const = 0;
};

enum BinaryType {
    ADD, SUB, SDIV, MUL, UDIV, UREM, SREM, SHL, AND, OR, XOR, ASHR, LSHR,
    FNEG, FSUB, FADD, FREM, FDIV, NOT,
    EQ, NE, SGT, SGE, SLE, SLT
};
bool isCompare(BinaryType binaryType);

class BinaryInstr :public DefInstr {
    shared_ptr<Reg> dst;
    shared_ptr<Reg> src1;
    shared_ptr<Reg> src2;
    BinaryType binaryType;

public:
    void setBinaryType(BinaryType binaryType);


    BinaryInstr(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2, BinaryType binaryType);

    shared_ptr<Type> def() const override;

    void set_dst(const shared_ptr<Reg> &dst);

    unordered_set<shared_ptr<Type>> get_uses() const override;
    void gen_ir(std::ostream &os) const override;

    shared_ptr<Reg> &get_dst() ;
    shared_ptr<Reg> &get_src_1() ;
    shared_ptr<Reg> &get_src_2() ;
    BinaryType get_binary_type() const;

    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(shared_ptr<Reg> src,shared_ptr<Reg> new_val) override;
    void set_src_1(const shared_ptr<Reg> &src1);
    void set_src_2(const shared_ptr<Reg> &src2);
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};

/**
 * void_flag判断是否该函数为void类型
 */
class CallInstr: public DefInstr {
    shared_ptr<Ir::Reg> dst;
    Function* function;
    std::vector<shared_ptr<Type>> params;
    bool void_flag = false;

public:
    unordered_set<shared_ptr<Type>> get_uses() const override;

    void gen_ir(std::ostream &os) const override;
    shared_ptr<Type> def() const override;

    CallInstr(const shared_ptr<Ir::Reg> &dst, Function *function, const std::vector<shared_ptr<Type>>&);
    CallInstr(Function *function, const std::vector<shared_ptr<Type>>&);
    shared_ptr<Ir::Reg> &get_dst();
    Function *get_function() const;
    std::vector<shared_ptr<Type>> &get_params();
    bool is_void_flag() const;
    void set_void_flag(bool voidFlag);
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void set_dst(const shared_ptr<Ir::Reg> &dst);
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<MemObject> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    unordered_set<shared_ptr<Type>> used_global; // 限制 mem read read global 的 motion 范围
};

/**
 * bb2 为null时是无条件跳转否则根据reg的值判断是否跳转
 */
class BranchInstr :public Instruction {
    shared_ptr<Reg> reg;
    BasicBlock *bb2 = nullptr;
    BasicBlock *bb1;

public:
    void set_bb1(BasicBlock *bb1);
    void set_bb2(BasicBlock *bb2);
    const shared_ptr<Reg> &get_reg() const;
    BasicBlock *get_bb_1() const;
    BasicBlock *get_bb_2() const;
    explicit BranchInstr(BasicBlock *bb1);
    BranchInstr(const shared_ptr<Reg> &reg, BasicBlock *bb1, BasicBlock *bb2);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void set_reg(const shared_ptr<Reg> &reg);
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void gen_ir(std::ostream &os) const override;
};

/**
 * 1. AllocaInstr 分配一个标量类型空间则addr为memObject基类
 * 2. AllocaInstr 分配一个Vector则addr为子类Vector
 */
class AllocaInstr :public DefInstr {
    //TODO array
    shared_ptr<MemObject> addr;
public:
    void gen_ir(std::ostream &os) const override;
    AllocaInstr(ScalaType type, int id);
    explicit AllocaInstr(const shared_ptr<MemObject> &addr);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    const shared_ptr<MemObject> &get_addr() const;
    shared_ptr<Type> def() const override;
    void add_use_def() override;
    void remove_use_def() override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};



/** Store类型有两类
 * 1.Store将val存入addr中
 * 2.StoreGlobal将val存入全局reg中
 */

class StoreInstr :public Instruction {
    shared_ptr<Reg> val;
    shared_ptr<MemObject> addr;
public:
    const shared_ptr<Reg> &get_val() const;
    const shared_ptr<MemObject> &get_addr() const;
    StoreInstr(const shared_ptr<Reg> &val, const shared_ptr<MemObject> &addr);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void set_val(const shared_ptr<Reg> &val);
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<MemObject> new_val) override;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};

class StoreGlobal :public Instruction {
    shared_ptr<Reg> val;
    shared_ptr<Reg> addr;
public:
    StoreGlobal(const shared_ptr<Reg> &val, const shared_ptr<Reg> &addr);
    const shared_ptr<Reg> &get_val() const;
    const shared_ptr<Reg> &get_addr() const;
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void add_use_def() override;
    void remove_use_def() override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void gen_ir(std::ostream &os) const override;
};


/**
 * load 指令有四类
 * 1.Load 从MemObject中取出标量类型
 * 2.LoadArray 从Vector中取出MemObject
 * 3.LoadVector 从Vector中取出Vector, 只可能在函数参数的时候使用
 * 4.LoadGlobal 从全局reg中取出标量
 */

class LoadVector:public DefInstr {
    shared_ptr<Vector> dst;
    shared_ptr<Vector> addr;
    std::vector<shared_ptr<Reg>> indexs;
public:
    const shared_ptr<Vector> &get_dst() const;
    const shared_ptr<Vector> &get_addr() const;
    std::vector<shared_ptr<Reg>> &get_indexs();
    unordered_set<shared_ptr<Type>> get_uses() const override;
    LoadVector(const shared_ptr<Vector> &dst, const shared_ptr<Vector> &addr, const std::vector<shared_ptr<Reg>>&);
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    shared_ptr<Type> def() const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void gen_ir(std::ostream &os) const override;
};


//class StoreArray:public Instruction {
//    shared_ptr<Reg> src;
//    shared_ptr<Vector> addr;
//    std::vector<shared_ptr<Reg>> indexs;
//
//public:
//    const shared_ptr<Reg> &get_src() const;
//    const shared_ptr<Vector> &get_addr() const;
//    std::vector<shared_ptr<Reg>> &get_indexs();
//    StoreArray(const shared_ptr<Reg> &src, const shared_ptr<Vector> &addr, const std::vector<shared_ptr<Reg>>&);
//    unordered_set<shared_ptr<Type>> get_uses() const override;
//
//};



class LoadArray:public DefInstr { // equals gep
    shared_ptr<MemObject> dst;
    shared_ptr<Vector> addr;
    std::vector<shared_ptr<Reg>> indexs;
public:

    LoadArray(const shared_ptr<MemObject> &dst, const shared_ptr<Vector> &addr, const std::vector<shared_ptr<Reg>>&);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    const shared_ptr<MemObject> &get_dst() const;
    const shared_ptr<Vector> &get_addr() const;
    std::vector<shared_ptr<Reg>> &get_indexs();
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    void gen_ir(std::ostream &os) const override;
    shared_ptr<Type> def() const override;
};


class Ret: public Instruction {
    bool void_flag = false;
    shared_ptr<Reg> src;
public:
    Ret();
    explicit Ret(const shared_ptr<Reg> &src);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    bool is_void_flag() const;
    const shared_ptr<Reg> &get_src() const;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;

    void gen_ir(std::ostream &os) const override;
};

class Convert: public DefInstr {
    shared_ptr<Reg> dst;
    shared_ptr<Reg> src;

public:
    const shared_ptr<Reg> &get_dst() const;
    const shared_ptr<Reg> &get_src() const;
    Convert(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    shared_ptr<Type> def() const override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};


class LoadInstr :public DefInstr {
    shared_ptr<Reg> dst;
    shared_ptr<MemObject> addr;

public:
    void gen_ir(std::ostream &os) const override;
    LoadInstr(const shared_ptr<Reg> &dst, const shared_ptr<MemObject> &addr);
    shared_ptr<Type> def() const override;
    unordered_set<shared_ptr<Type>> get_uses() const override;
    const shared_ptr<Reg> &get_dst() const;
    const shared_ptr<MemObject> &get_addr() const;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<MemObject> new_val) override;
};

class LoadGlobal :public DefInstr {
    shared_ptr<Reg> dst;
    shared_ptr<Reg> addr;

public:
    LoadGlobal(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &addr);
    unordered_set<shared_ptr<Type>> get_uses() const override;
    shared_ptr<Type> def() const override;
    const shared_ptr<Reg> &get_dst() const;
    const shared_ptr<Reg> &get_addr() const;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    void add_use_def() override;
    void remove_use_def() override;
};

class PhiInstr: public DefInstr {

    shared_ptr<Reg> dst;
    unordered_map<BasicBlock*, shared_ptr<Reg>> bb2value;


public:
    PhiInstr(const shared_ptr<Reg> &dst,const unordered_map<BasicBlock*, shared_ptr<Reg>> &bb2value);
    const shared_ptr<Reg> &get_dst() const;
    unordered_map<BasicBlock *, shared_ptr<Reg>> &get_bb_2_value() ;
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    explicit PhiInstr(const shared_ptr<Reg> &dst);
    void gen_ir(std::ostream &os) const override;
    shared_ptr<Type> def() const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};

// array ssa

// array ssa phi
class ArrayPhi:public DefInstr { // reg 代表global,vector表示数组
    shared_ptr<Type> dst;
    unordered_map<BasicBlock*, shared_ptr<Type>> bb2value;
    shared_ptr<Type> scope;    // 只能是 reg 或者 vector
public:
    shared_ptr<Type> def() const override;
    ArrayPhi(const shared_ptr<Type>& dst, const shared_ptr<Type>& scope);
    const shared_ptr<Type> &get_dst() const;
    const shared_ptr<Type> &get_scope() const;
    unordered_map<BasicBlock *, shared_ptr<Type>> &get_bb_2_value();
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
    unordered_set<shared_ptr<Type>> read_before_write; // 是 read 的dst, reg 或者 vector,该use 用于限制 memRead 的 motion,可能导致memRead motion 到 array phi后面
};

// memWrite
class MemWrite:public Instruction { // store , store global
    shared_ptr<Type> addr;
    shared_ptr<Reg> val;  // reg and memobject
    shared_ptr<Type> scope; // 只能是 reg 或者 vector

    shared_ptr<Type> new_scope;
public:
    MemWrite(const shared_ptr<Type> &addr, const shared_ptr<Reg> &val, const shared_ptr<Type> &scope, const shared_ptr<Type> new_scope,bool call_write);
    const shared_ptr<Type> &get_addr() const;
    const shared_ptr<Reg> &get_val() const;
    const shared_ptr<Type> &get_scope() const;
    unordered_set<shared_ptr<Type>> get_uses() const override;
    unordered_set<shared_ptr<Type>> get_pure_uses() const ;
    void add_use_def() override;
    bool call_write; // 表明是否是call用的,如果是call用的那就只是一个占位,在destruction的时候可以不管
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    void change_use_name(int src_id,shared_ptr<MemObject> new_val) override;
    unordered_set<shared_ptr<Type>> read_before_write; // 是 read 的dst, reg 或者 vector
    const shared_ptr<Type> &get_new_scope() const;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};

// memRead
class MemRead:public Instruction {
    shared_ptr<Type> dst; // reg 或者 vector
    shared_ptr<Type> scope; // 只能是 reg 或者 vector
    shared_ptr<Type> addr; // reg and memobject


public:
    bool call_read;
    MemRead(const shared_ptr<Type> &dst, const shared_ptr<Type> &scope, const shared_ptr<Type> &addr,bool call_read);
    const shared_ptr<Type> &get_dst() const;
    const shared_ptr<Type> &get_scope() const;
    const shared_ptr<Type> &get_addr() const;
    unordered_set<shared_ptr<Type>> get_uses() const override;
    void add_use_def() override;
    void remove_use_def() override;
    void change_use_name(int src_id,shared_ptr<Reg> new_val) override;
    void change_use_name(int src_id,shared_ptr<Vector> new_val) override;
    void change_use_name(int src_id,shared_ptr<MemObject> new_val) override;
    void gen_ir(std::ostream &os) const override;
    Instruction *clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) override;
};


class BasicBlock {
    int id;
    std::list<std::unique_ptr<Instruction>> instrs;
    Function* belongto_func;
    Ir::Loop * belong_to_loop = nullptr;
    int depth = 0;
    bool terminal = false;
public:
    int get_id() const;
    void set_id(int id);
    friend std::ostream &operator<<(std::ostream &os, const BasicBlock &block);
    BasicBlock(int id);
    void push_back(std::unique_ptr<Instruction>& instr);
    void push_front(Instruction *instr);
    void push_front(std::unique_ptr<Instruction> instr);
    void insert_after_phi(Instruction* instr);
    void insert_before_br(Instruction* instr);
    std::list<std::unique_ptr<Instruction>> &get_instrs() ;
    Function *get_belongto_func() const;
    void set_belongto_func(Function *belongtoFunc);
    Loop *get_belong_to_loop() const;
    void set_belong_to_loop(Loop *belongToLoop);
    int get_depth() const;
    void set_depth(int depth);
    bool is_terminal() const;
    void set_terminal(bool terminal);
    std::unordered_set<BasicBlock *> prev;
    std::unordered_set<BasicBlock *> succ;
    unordered_set<BasicBlock*> dom_by; // any e in the domby doms this
    unordered_set<BasicBlock*> dom; // any e in the dom is domed by this
    unordered_set<BasicBlock*> idom; // any e in the dom is idomed by this
    BasicBlock* idom_by = nullptr; // immediate dom
    int dom_level = 0;
    unordered_set<BasicBlock*> df;
    int get_loop_level();
};

class Function {
    std::string name;
    ScalaType type;
    bool recur = false;
    bool lib = false;
    std::list<std::unique_ptr<BasicBlock>> bbs;
    std::list<std::shared_ptr<Type>> params;
    std::list<std::unique_ptr<Loop>> loops;
    Function() = default;
    int reg_max_num = 0;
    int bb_max_num = 0;
    BasicBlock * first_bb = nullptr;

public:
    std::shared_ptr<MemObject> ret_mem;
    bool is_lib() const;
    void set_lib(bool lib);
    int get_bb_max_num() const;
    void set_bb_max_num(int bbMaxNum);
    void set_reg_max_num(int regMaxNum);
    int get_reg_max_num() const;
    friend std::ostream &operator<<(std::ostream &os, const Function &function);
    bool pure = false;
    bool array_ssa_pure = false;
    bool visited = false;
    std::list<std::shared_ptr<Type>> &get_params();
    const std::string &get_name() const;
    ScalaType get_type() const;
    std::list<std::unique_ptr<BasicBlock>> &get_bbs();
    bool is_recur() const;
    bool is_void() const;
    void set_recur(bool recur);
    Function(std::string name, ScalaType type);
    std::list<std::unique_ptr<Loop>> &get_loops() ;
    std::unordered_map<int, Instruction *> def_place; // id 2 use place
    std::unordered_map<int, std::unordered_set<Instruction *>> use_place; // id 2 use place
    opt::CFG* cfg;
    bool is_param(Type * type);
    BasicBlock *get_first_bb() const;
    void set_first_bb(BasicBlock *firstBb);
};

class Loop {
    BasicBlock * header = nullptr;
    BasicBlock* exiting_bb;
    int depth = 0;

private:
    Function * belong_to_func = nullptr;
    Loop *parent = nullptr;
    BasicBlock * preheader = nullptr;

    std::unordered_set<BasicBlock *> all_bbs;


public:
    Loop(BasicBlock *header, BasicBlock *exitingBb, int depth, Function *belongToFunc, Loop *parent);
    BasicBlock *get_header() const;
    BasicBlock *get_exiting_bb() const;
    int get_depth() const;
    Function *get_belong_to_func() const;
    BasicBlock *get_preheader() const;
    Loop *get_parent() const;
    unordered_set<BasicBlock *> &get_all_bbs() ;
    void set_parent(Loop *parent);
    void set_depth(int depth);
    void set_all_bbs(const unordered_set<BasicBlock *> &allBbs);
    unordered_set<Loop *> children;
};
/**
 *global_arrays 是全局数组、以及可以被转化成全局数组的临时数组的值，有int和float两种类型，只有一维的概念
 *global_consts 是全局变量的初始值，包含全局变量和常量（包括局部常量）
 */
class CompileUnit {

    std::list<std::unique_ptr<Function>> functions;
    std::unordered_map<std::string, std::vector<int>> global_int_arrays;
    std::unordered_map<std::string, std::vector<float>> global_float_arrays;
//    static std::unordered_map<std::string, int> global_consts;
//    static std::unordered_map<std::string, double> global_floats;
    std::list<shared_ptr<Type>> globals;

public:
    static bool in_arm;
    static bool in_array_ssa;
public:
    friend std::ostream &operator<<(std::ostream &os, const CompileUnit &unit);
    std::list<std::unique_ptr<Function>> &get_functions();
    unordered_map<std::string, std::vector<int>> &get_global_int_arrays();
    unordered_map<std::string, std::vector<float>> &get_global_float_arrays();
    std::list<shared_ptr<Type>> &get_globals();
    std::string print_array(Vector* vector, int deep, int& num) const;
    void reset_num();
};


class Factory {


public:
    static int global_num;
    static int reg_num ;
    static int bb_num;

public:
    static void reset(Function* func);
    //获取全局数组变量
    static std::shared_ptr<Vector> get_vector(ScalaType type, const std::string& name,  bool constant);
    //获取alloca和loadvector生成的局部变量
    static std::shared_ptr<Vector> get_vector(ScalaType type);
    static std::shared_ptr<Reg> get_reg(ScalaType);
    static std::shared_ptr<MemObject> get_mem(ScalaType type);
    static std::shared_ptr<Reg> get_const_reg(ScalaType, RegConst num);
    static std::shared_ptr<Reg> get_global_reg(ScalaType, RegConst num, std::string& name);
    static Function*  get_func(std::string&name, ScalaType, CompileUnit * cu);
    static void build_alloca(ScalaType type, BasicBlock &bb);
    static void build_alloca(const shared_ptr<Vector> &addr, BasicBlock &bb);
    static std::unique_ptr<Instruction> get_alloca(ScalaType type);
    static std::unique_ptr<PhiInstr> get_phi(const shared_ptr<Reg> &dst);
    static void build_store(const shared_ptr<Reg> &val, const shared_ptr<MemObject> &addr, BasicBlock& );
    static void build_store_global(const shared_ptr<Reg> &val, const shared_ptr<Reg> &addr, BasicBlock& );

    static void build_binary(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2,
                             BinaryType binaryType, BasicBlock& bb);

    void dumplicated_code(Ir::CompileUnit* compUnit);

    static std::unique_ptr<BinaryInstr> get_binary(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2,BinaryType binaryType);

    static void build_load(const shared_ptr<Reg> &dst, const shared_ptr<MemObject> &addr, BasicBlock&bb);

    static void build_load_global(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &addr,
                           BasicBlock&bb);


    static void build_load_array(const shared_ptr<Reg> &dst, const shared_ptr<Vector> &addr,
                                 const std::vector<shared_ptr<Reg>>&, BasicBlock& bb);

    static void build_store_array(const shared_ptr<Reg> &src, const shared_ptr<Vector> &addr,
                                  const std::vector<shared_ptr<Reg>>& index, BasicBlock&bb);

    static void build_load_vector(const shared_ptr<Vector> &dst, const shared_ptr<Vector> &addr,
                                  const std::vector<shared_ptr<Reg>>&, BasicBlock& bb);

    static std::unique_ptr<Instruction> get_convert(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src);

    static std::shared_ptr<Ir::Reg> get_zero_reg(ScalaType type);

    static std::shared_ptr<Ir::Type> copy_def(const shared_ptr<Type>& def);

    static void build_call(const shared_ptr<Reg> &dst, Function *function,
                           const std::vector<shared_ptr<Type>>&prams, BasicBlock&);

    static void build_call(Function *function, const std::vector<shared_ptr<Type>>&params, BasicBlock&);

    static void build_branch(const shared_ptr<Reg> &reg, BasicBlock *bb1, BasicBlock *bb2, BasicBlock&);
    static void build_branch(BasicBlock *bb1, BasicBlock&);
    static BasicBlock* get_new_bb(Loop *, Function*);

    static std::unique_ptr<BasicBlock> get_noadd_new_bb(Loop * loop, Function* func);

    static Loop* get_loop(BasicBlock *header, BasicBlock *exitingBb,int depth, Function *belongToFunc, Loop*);

    static void build_ret(const shared_ptr<Reg> &src, BasicBlock& bb);

    static void build_ret(BasicBlock& bb);
};




} // namespace Ir
#endif