//
// Created by 11067 on 2023/7/9.
//

#include "Bitset.h"
#include "utils.h"
#include "logger.h"
#include <algorithm>

namespace atri {
    Bitset::Bitset(const size_t &bitLength) : bitLength(bitLength) {
        field = new FieldType[bitLength / bitWidth() + 1];
        const auto wordLength = bitLength / bitWidth() + 1;
        for (size_t i = 0; i < wordLength; i++) {
            field[i] = 0ull;
        }
    }

    Bitset::~Bitset() {
        delete[]field;
    }

    Bitset::Bitset(const Bitset &rhs) {
        bitLength = rhs.bitLength;
        size_t wordLength = bitLength / bitWidth() + 1;
        field = new FieldType[wordLength];
        for (size_t i = 0; i < wordLength; i++) {
            field[i] = 0;
        }
        std::copy(rhs.field, rhs.field + wordLength, field);
    }

    Bitset::Bitset(Bitset &&rhs) noexcept {
        bitLength = rhs.bitLength;
        field = rhs.field;
        rhs.field = nullptr;
        rhs.bitLength = 0;
    }

    Bitset &Bitset::operator=(Bitset &&rhs) noexcept {
        if (this != &rhs) {
            delete[]field;
            bitLength = rhs.bitLength;
            field = rhs.field;
            rhs.field = nullptr;
            rhs.bitLength = 0;
        }
        return *this;
    }

    Bitset &Bitset::operator=(const Bitset &rhs) {
        if (this != &rhs) {
            delete[]field;
            bitLength = rhs.bitLength;
            size_t wordLength = bitLength / bitWidth() + 1;
            field = new FieldType[wordLength];
            for (size_t i = 0; i < wordLength; i++) {
                field[i] = 0;
            }
            std::copy(rhs.field, rhs.field + wordLength, field);
        }
        return *this;
    }

    Bitset &Bitset::operator&=(const Bitset &rhs) {
        atriAssert(bitLength == rhs.bitLength);
        const auto wordLength = bitLength / bitWidth() + 1;
        for (size_t i = 0; i < wordLength; i++) {
            field[i] &= rhs.field[i];
        }
        return *this;
    }

    Bitset &Bitset::operator|=(const Bitset &rhs) {
        atriAssert(bitLength == rhs.bitLength);
        const auto wordLength = bitLength / bitWidth() + 1;
        for (size_t i = 0; i < wordLength; i++) {
            field[i] |= rhs.field[i];
        }
        return *this;
    }

    Bitset Bitset::operator~() {
        auto ret = Bitset(bitLength);
        const auto wordLength = bitLength / bitWidth() + 1;
        for (size_t i = 0; i < wordLength; i++) {
            ret.field[i] = ~field[i];
        }
        return ret;
    }

    bool Bitset::operator[](size_t idx) const {
        FieldType entry = field[idx / bitWidth()];
        size_t entryOffset = idx % bitWidth();
        return (entry & (1ull << entryOffset)) != 0;
    }

    Bitset::reference Bitset::operator[](const size_t idx) {
        return {idx, field};
    }

    size_t Bitset::getBitLength() const {
        return bitLength;
    }

    Bitset &Bitset::andNot(const Bitset &rhs) {
        atriAssert(bitLength == rhs.bitLength);
        const auto wordLength = bitLength / bitWidth() + 1;
        for (size_t i = 0; i < wordLength; i++) {
            field[i] &= ~rhs.field[i];
        }
        return *this;
    }

    std::vector<int> Bitset::bitView() const {
        const auto wordLength = bitLength / bitWidth() + 1;
        std::vector<int> ret;
        for (size_t i = 0; i < wordLength; ++i) {
            FieldType entry = field[i / bitWidth()];
            size_t entryOffset = i % bitWidth();
            if ((entry & (1ull << entryOffset)) != 0) {
                ret.push_back(i);
            }
        }
        return ret;
    }

    Bitset::reference::reference(size_t pos, FieldType *field) : pos(pos), field(field) {}

    Bitset::reference &Bitset::reference::operator=(const bool rhs) {
        FieldType &entry = field[pos / bitWidth()];
        size_t entryOffset = pos % bitWidth();
        FieldType bit = 1ull << entryOffset;
        if (rhs) {
            entry |= bit;
        } else {
            entry &= ~bit;
        }
        return *this;
    }

    Bitset::reference::operator bool() {
        FieldType &entry = field[pos / bitWidth()];
        size_t entryOffset = pos % bitWidth();
        return (entry & (1ull << entryOffset)) != 0;
    }
} // atri