//
// Created by 11067 on 2023/5/18.
//

#ifndef ATRI_BACKEND_LINKEDLIST_H
#define ATRI_BACKEND_LINKEDLIST_H

#include <iterator>
#include <vector>
namespace atri {
    template<typename T> class LinkedNode;
    /**
     * @brief 链表容器
     * 使用了CRTP，例如basicblock包含instructions,那么bb就用CRTP
     * 方法继承这个模板类
     * ownership：容器不会自动管理节点的内存，节点需手动释放。
     * @tparam T 派生类
     */
    template<typename T> class LinkedContainer{
    private:
        int size = 0;
        LinkedNode<T> *first{nullptr};
    public:
        T *getFirst() const {
            return static_cast<T*>(first);
        }

        T *getLast() const {
            return static_cast<T*>(last);
        }
        /**
         * @brief 在迭代时修改容器会有bug，该函数复制容器snapshot以解决此问题
         * @return 复制的vector（浅复制）
         */
        std::vector<T*> toVector() {
            std::vector<T*> ret;
            for (auto elem: *this) {
                ret.push_back(elem);
            }
            return ret;
        }

    private:
        LinkedNode<T> *last{nullptr};
        friend LinkedNode<T>;
    public:
        void setFirst(LinkedNode<T> *first1) {
            first = first1;
        }
        void setLast(LinkedNode<T> *last1) {
            last = last1;
        }
        int getSize() {
            return size;
        }
        void addLast(LinkedNode<T> *node) {
            if (last != nullptr) {
                last->next = node;
                node->prev = last;
                last = node;
            } else {
                last = node;
                first = node;
            }
            node->next = nullptr;
            size++;
            node->setOwner(this);
        }
        void increaseSize() {
            size++;
        }
        void remove(LinkedNode<T> *node) {
            if (first == last && first == node) {
                first = last = nullptr;
            } else if (node == first) {
                first = node->next;
                node->next->prev = nullptr;
            } else if (node == last) {
                last = node->prev;
                node->prev->next = nullptr;
            } else {
                if (node->prev != nullptr) {
                    node->prev->next = node->next;
                }
                if (node->next != nullptr) {
                    node->next->prev = node->prev;
                }
            }
            node->prev = node->next = nullptr;
            size--;
        }
        class iterator{
            LinkedNode<T>* pointer;
        public:
            explicit iterator(LinkedNode<T> * pn) {
                pointer = pn;
            }
            iterator& operator++() {
                pointer = pointer->next;
                return *this;
            }

            iterator operator++(int) {
                iterator ret = *this;
                ++(*this);
                return ret;
            }

            bool operator==(const iterator &rhs) const {
                return pointer == rhs.pointer;
            }

            bool operator!=(const iterator &rhs) const {
                return pointer != rhs.pointer;
            }
            T* operator*() {
                return static_cast<T*>(this->pointer);
            }
        };
        iterator begin() {
            return iterator(first);
        }
        iterator end() {
            return iterator(nullptr);
        }

    };

    /**
     * 链表节点
     * @tparam T 派生类
     */
    template<typename T> class LinkedNode{
    private:
        LinkedNode<T> *prev = nullptr;
        LinkedNode<T> *next = nullptr;
        LinkedContainer<T> *owner = nullptr;
        friend LinkedContainer<T>;
    public:
        LinkedContainer<T> *getOwner() {
            return owner;
        }

        void setOwner(LinkedContainer<T> *owner1) {
            owner = owner1;
        }
        void insertBeforeMe(LinkedNode<T> *other) {
            if (prev != nullptr) {
                prev->next = other;
            } else {
                getOwner()->setFirst(other);
            }
            other->prev = prev;
            prev = other;
            other->next = this;
            this->getOwner()->increaseSize();
            other->setOwner(this->getOwner());
        }

        void insertAfterMe(LinkedNode<T> *other) {
            if (next != nullptr) {
                next->prev = other;
            } else {
                getOwner()->setLast(other);
            }
            other->next = next;
            next = other;
            other->prev = this;
            this->getOwner()->increaseSize();
            other->setOwner(this->getOwner());
        }
        /**
         * @brief  删除自己
         * 不会检查owner是否处于有效状态。
         */
        void removeMe() {
            getOwner()->remove(this);
        }

        T *getNext() {
            return static_cast<T*>(next);
        }

        T *getPrev() {
            return static_cast<T*>(prev);
        }
    };






} // atri

#endif //ATRI_BACKEND_LINKEDLIST_H
